/* cpu / network monitor program */
#include <ctype.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#define UDPPORT 7075

static const char *internal_intf = "eth0";
static const char *external_intf = "eth1";

/* borrowed from vmstat.c and modified */
typedef unsigned long long jiff;

static int debug_flag = 0;

static int
debug_printf(const char *format, ...)
{
    int rv = 0;
    va_list ap;

    if (debug_flag) {
        va_start(ap, format);
        rv = vprintf(format, ap);
        va_end(ap);
    }
    return rv;
}

static int
debug_fprintf(FILE *stream, const char *format, ...)
{
    int rv = 0;
    va_list ap;

    if (debug_flag) {
        va_start(ap, format);
        rv = vfprintf(stream, format, ap);
        va_end(ap);
    }
    return rv;
}

static void
crash(const char *filename) {
    perror(filename);
    exit(EXIT_FAILURE);
}

static void
getstat(jiff *cuse, jiff *cice, jiff *csys, jiff *cide, jiff *ciow,
        jiff *cxxx, jiff *cyyy, jiff *czzz)
{
    static int fd;
    int rv;
    char *b;
    char buf[BUFSIZ];

    memset(buf, 0, sizeof(buf));
    if (fd) {
        lseek(fd, 0L, SEEK_SET);
    } else {
        fd = open("/proc/stat", O_RDONLY, 0);
        if (fd == -1)
            crash("/proc/stat");
    }
    read(fd, buf, sizeof(buf) - 1);
    *ciow = 0;  /* not separated out until the 2.5.41 kernel */
    *cxxx = 0;  /* not separated out until the 2.6.0-test4 kernel */
    *cyyy = 0;  /* not separated out until the 2.6.0-test4 kernel */
    *czzz = 0;  /* not separated out until the 2.6.11 kernel */

    b = strstr(buf, "cpu ");
    if (b) {
        rv = sscanf(b, "cpu  %Lu %Lu %Lu %Lu %Lu %Lu %Lu %Lu",
            cuse, cice, csys, cide, ciow, cxxx, cyyy, czzz);
        debug_printf("sscanf returns: %d\n", rv);
    } else {
        debug_fprintf(stderr, "``cpu '' is not found in: %s\n", buf);
    }
    debug_printf("getstat: use:%Lu ice:%Lu sys:%Lu ide:%Lu iow:%Lu"
            " xxx:%Lu yyy:%Lu zzz:%Lu\n",
            *cuse, *cice, *csys, *cide, *ciow, *cxxx, *cyyy, *czzz);
}

static int
get_cpuusage(void)
{
    int idle_percent, iow_percent;
    static int tog = 1;
    static int debt = 0;
    static jiff cuse[2], cice[2], csys[2], cide[2], ciow[2],
                cxxx[2], cyyy[2], czzz[2];
    static jiff duse, dsys, didl, diow, dstl, div, divo2;

    tog = !tog;
    getstat(cuse + tog, cice + tog, csys + tog, cide + tog, ciow + tog,
            cxxx + tog, cyyy + tog, czzz + tog);
    duse = cuse[tog] - cuse[!tog] + cice[tog] - cice[!tog];
    dsys = csys[tog] - csys[!tog] + cxxx[tog] - cxxx[!tog]
            + cyyy[tog] - cyyy[!tog];
    didl = cide[tog] - cide[!tog];
    diow = ciow[tog] - ciow[!tog];
    dstl = czzz[tog] - czzz[!tog];
    if (debt) {
        didl = (int)didl + debt;
        debt = 0;
    }
    if ((int)didl < 0) {
        debt = (int)didl;
        didl = 0;
    }
    div = duse + dsys + didl + diow + dstl;
    if (div == 0) {
        debug_fprintf(stderr, "call too frequently\n");
        return 0;
    }
    divo2 = div / 2UL;
    idle_percent = (100 * didl + divo2) / div;
    iow_percent = (100 * diow + divo2) / div;
    idle_percent += iow_percent;

    debug_printf("diff: use:%Lu sys:%Lu idl:%Lu iow:%Lu stl:%Lu debt:%d -> div%Lu\n",
            duse, dsys, didl, diow, dstl, debt, div);
    debug_printf("use:%u sys:%u idl:%u iow:%u\n",
        (unsigned)((100 * duse + divo2) / div),
        (unsigned)((100 * dsys + divo2) / div),
        (unsigned)((100 * didl + divo2) / div),
        (unsigned)((100 * diow + divo2) / div));
    debug_printf("idle_percent %d\n", idle_percent);

    if (idle_percent > 100)
        idle_percent = 100;
    if (idle_percent < 0)
        idle_percent = 0;
    return 100 - idle_percent;
}

/* borrowed from interface.c and modified */

struct user_net_device_stats {
    unsigned long long rx_packets;  /* total packets received       */
    unsigned long long tx_packets;  /* total packets transmitted    */
    unsigned long long rx_bytes;    /* total bytes received         */
    unsigned long long tx_bytes;    /* total bytes transmitted      */
    unsigned long rx_errors;    /* bad packets received         */
    unsigned long tx_errors;    /* packet transmit problems     */
    unsigned long rx_dropped;   /* no space in linux buffers    */
    unsigned long tx_dropped;   /* no space available in linux  */
    unsigned long rx_multicast; /* multicast packets received   */
    unsigned long rx_compressed;
    unsigned long tx_compressed;
    unsigned long collisions;

    /* detailed rx_errors: */
    unsigned long rx_length_errors;
    unsigned long rx_over_errors;   /* receiver ring buff overflow  */
    unsigned long rx_crc_errors;    /* recved pkt with crc error    */
    unsigned long rx_frame_errors;  /* recv'd frame alignment error */
    unsigned long rx_fifo_errors;   /* recv'r fifo overrun          */
    unsigned long rx_missed_errors; /* receiver missed packet     */
    /* detailed tx_errors */
    unsigned long tx_aborted_errors;
    unsigned long tx_carrier_errors;
    unsigned long tx_fifo_errors;
    unsigned long tx_heartbeat_errors;
    unsigned long tx_window_errors;
};

static int procnetdev_vsn = 1;

char
*get_name(char *name, char *p)
{
    while (isspace(*p))
    p++;
    while (*p) {
    if (isspace(*p))
        break;
    if (*p == ':') {    /* could be an alias */
        char *dot = p++;
        while (*p && isdigit(*p)) p++;
        if (*p == ':') {
            /* Yes it is, backup and copy it. */
            p = dot;
            *name++ = *p++;
            while (*p && isdigit(*p)) {
                *name++ = *p++;
            }
        } else {
            /* No, it isn't */
            p = dot;
        }
        p++;
        break;
    }
    *name++ = *p++;
    }
    *name++ = '\0';
    return p;
}

static int
procnetdev_version(char *buf)
{
    if (strstr(buf, "compressed"))
        return 3;
    if (strstr(buf, "bytes"))
        return 2;
    return 1;
}

static int
get_dev_fields(char *bp, struct user_net_device_stats *stats)
{
    switch (procnetdev_vsn) {
    case 3:
        sscanf(bp,
        "%Lu %Lu %lu %lu %lu %lu %lu %lu %Lu %Lu %lu %lu %lu %lu %lu %lu",
               &stats->rx_bytes,
               &stats->rx_packets,
               &stats->rx_errors,
               &stats->rx_dropped,
               &stats->rx_fifo_errors,
               &stats->rx_frame_errors,
               &stats->rx_compressed,
               &stats->rx_multicast,

               &stats->tx_bytes,
               &stats->tx_packets,
               &stats->tx_errors,
               &stats->tx_dropped,
               &stats->tx_fifo_errors,
               &stats->collisions,
               &stats->tx_carrier_errors,
               &stats->tx_compressed);
        break;
    case 2:
        sscanf(bp, "%Lu %Lu %lu %lu %lu %lu %Lu %Lu %lu %lu %lu %lu %lu",
               &stats->rx_bytes,
               &stats->rx_packets,
               &stats->rx_errors,
               &stats->rx_dropped,
               &stats->rx_fifo_errors,
               &stats->rx_frame_errors,

               &stats->tx_bytes,
               &stats->tx_packets,
               &stats->tx_errors,
               &stats->tx_dropped,
               &stats->tx_fifo_errors,
               &stats->collisions,
               &stats->tx_carrier_errors);
        stats->rx_multicast = 0;
        break;
    case 1:
        sscanf(bp, "%Lu %lu %lu %lu %lu %Lu %lu %lu %lu %lu %lu",
               &stats->rx_packets,
               &stats->rx_errors,
               &stats->rx_dropped,
               &stats->rx_fifo_errors,
               &stats->rx_frame_errors,

               &stats->tx_packets,
               &stats->tx_errors,
               &stats->tx_dropped,
               &stats->tx_fifo_errors,
               &stats->collisions,
               &stats->tx_carrier_errors);
        stats->rx_bytes = 0;
        stats->tx_bytes = 0;
        stats->rx_multicast = 0;
        break;
    }
    return 0;
}

static void
get_netbytes(unsigned long long *int_rx_bytes,
             unsigned long long *int_tx_bytes,
             unsigned long long *ext_rx_bytes,
             unsigned long long *ext_tx_bytes)
{
    int n_read_intf = 0;
    FILE *fh;
    static unsigned long long last_int_rx_bytes = 0, last_int_tx_bytes = 0;
    static unsigned long long last_ext_rx_bytes = 0, last_ext_tx_bytes = 0;
    char buf[BUFSIZ];

    fh = fopen("/proc/net/dev", "r");
    if (!fh) {
        fprintf(stderr, "cannot open %s\n", "/proc/net/dev");
        return;
    }
    fgets(buf, sizeof(buf), fh);
    fgets(buf, sizeof(buf), fh);
    procnetdev_vsn = procnetdev_version(buf);
    while (fgets(buf, sizeof(buf), fh)) {
        struct user_net_device_stats stats;
        char *s, name[BUFSIZ];
        s = get_name(name, buf);
        if(strcmp(name, internal_intf) == 0) {
            get_dev_fields(s, &stats);
            /* printf("%llu %llu\n", stats.rx_bytes, stats.tx_bytes); */
            if (last_int_rx_bytes != 0)
                *int_rx_bytes = stats.rx_bytes - last_int_rx_bytes;
            if (last_int_tx_bytes != 0)
                *int_tx_bytes = stats.tx_bytes - last_int_tx_bytes;
            last_int_rx_bytes = stats.rx_bytes;
            last_int_tx_bytes = stats.tx_bytes;
            n_read_intf++;
        } else if(strcmp(name, external_intf) == 0) {
            get_dev_fields(s, &stats);
            /* printf("%llu %llu\n", stats.rx_bytes, stats.tx_bytes); */
            if (last_ext_rx_bytes != 0)
                *ext_rx_bytes = stats.rx_bytes - last_ext_rx_bytes;
            if (last_ext_tx_bytes != 0)
                *ext_tx_bytes = stats.tx_bytes - last_ext_tx_bytes;
            last_ext_rx_bytes = stats.rx_bytes;
            last_ext_tx_bytes = stats.tx_bytes;
            n_read_intf++;
        }
        if (n_read_intf >= 2)
            break;
    }
    fclose(fh);
    return;
}

static char *
build_send_str(void)
{
    static char buf[BUFSIZ];
    unsigned long long int_rx_bytes = 0, int_tx_bytes = 0;
    unsigned long long ext_rx_bytes = 0, ext_tx_bytes = 0;

    get_netbytes(&int_rx_bytes, &int_tx_bytes, &ext_rx_bytes, &ext_tx_bytes);
    debug_printf("internal\n");
    debug_printf("  rx: %llu\n", int_rx_bytes);
    debug_printf("  tx: %llu\n", int_tx_bytes);
    debug_printf("external\n");
    debug_printf("  rx: %llu\n", ext_rx_bytes);
    debug_printf("  tx: %llu\n", ext_tx_bytes);
    sprintf(buf, "%d\n%llu\n%llu\n%llu\n%llu\n", get_cpuusage(), int_rx_bytes, int_tx_bytes, ext_rx_bytes, ext_tx_bytes);

    return buf;
}

void
usage(char *argv[])
{
fprintf(stderr,
            "Usage: %s [OPTION]...\n\n"
            "OPTION:\n"
            "      [-d]\n"
            "      [-e <external network interface>(default: eth1)]\n"
            "      [-i <internal network interface>(default: eth0)]\n", argv[0]);
    fprintf(stderr, "\nATTENTION: DONT'T SET ALIAS NAME AS NETWORK INTERFACE.\n");
    exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[])
{
    int sd, rv;
    struct sockaddr_in in;
    struct sockaddr_in from;
    socklen_t fromlen;
    char buf[BUFSIZ];

    argv++;
    for(; *argv != NULL; argv++) {
        if (strcmp(*argv, "-d") == 0) {
            debug_flag = 1;
            debug_fprintf(stderr, "Entering debug mode\n");
        } else if (strcmp(*argv, "-e") == 0) {
            external_intf = *(argv++);
        } else if (strcmp(*argv, "-i") == 0) {
            internal_intf = *(argv++);
        } else if (strcmp(*argv, "-h") == 0 || strcmp(*argv, "--help") == 0) {
            usage(argv);
        } else {
            usage(argv);
        }
    }

    memset(&in, 0, sizeof(in));
    in.sin_family = AF_INET;
    in.sin_port = htons(UDPPORT);
    in.sin_addr.s_addr = INADDR_ANY;

    sd = socket(PF_INET, SOCK_DGRAM, 0);
    if (sd < 0) {
        perror("socket");
        exit(1);
    }
    rv = bind(sd, (struct sockaddr*)&in, sizeof(in));
    if (rv < 0) {
        perror("bind");
        exit(1);
    }
    build_send_str();
    while (1) {
        char *cp;
        fromlen = sizeof(from);
        recvfrom(sd, buf, sizeof(buf), 0, (struct sockaddr*)&from, &fromlen);
        cp = build_send_str();
        sendto(sd, cp, strlen(cp), 0, (struct sockaddr*)&from, fromlen);
    }
    return 0;
}
