#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "common.h"
#include "dtsadmin.h"
#include "dtsadmin_codes.h"
#include "dtsadmin_dtsdio.h"
#include "dtsadmin_result_handler.h"
#include "dtsadmin_server_conf_handler.h"
#include "dtsadmin_nodeinfo_handler.h"
#include "dtsd_common.h"
#include "dtsd_fifocom.h"
#include "dtsd_optparser.h"
#include "dtsd_worker.h"
#include "dtsd_util.h"

struct com_method cmethod;

static
void print_usage()
{
	fprintf(stderr,
			"Usage: dtsadmin [options]\n"
			"\n"
			"options:\n"
			"  -A <host:srv> -i <iface> [-M <netmask>] -f <on/off> -r <ri> [-s <ri opt>] [<initial node set>]\n"
			"     add the server configuration to dtsd.\n"
			"\n"
			"     <host>\n"
			"        outer IP address/hostname of the cluster\n"
			"     <srv>\n"
			"        service port number.\n"
			"     -i <iface>\n"
			"        interface name which used for outer communication.\n"
			"     -M <netmask>\n"
			"        netmask which set to <iface>.(possible to omit)\n"
			"     -f <on/off>\n"
			"        fakemac opt(on: enable, off: disable).\n"
			"     -r <ri>\n"
			"        request inspector name.\n"
			"     -s <ri opt>\n"
			"        request inspector's option.(possible to omit)\n"
			"     <initial node set>\n"
			"        initial server node set of this server configuration.(possible to omit)\n"
			"        the format is the following.\n"
			"\n"
			"        <Heartbeat HostName>[/<VirtualNode ID>,....]:...\n"
			"  -a <host:srv> <node spec>\n"
			"     add the server node or virtual node to server configuration.\n"
			"\n"
			"     <node spec>\n"
			"        node to add the server configuration.\n"
			"        the format is the following.\n"
			"\n"
			"        <Heartbeat HostName>[/<VirtualNode ID>,....]\n"
			"  -d <host:srv> <node spec>\n"
			"     delete the server node or virtual node from server configuration.\n"
			"  -L [-c] [-n] [-v]\n"
			"     print list the current server configuration to stdout.\n"
			"\n"
			"     -c\n"
			"        print the hash value of virtual node with the list.\n"
			"     -n\n"
			"        print the node which currently stopped with the list.\n"
			"     -v\n"
			"        verbose output.\n"
			"  -C\n"
			"     clear all current server configuration.\n"
			"  -S\n"
			"     print all current server configuration to stdout.\n"
			"  -R\n"
			"     recovery server configuration from stdin.\n"
			"  -h | --help:\n"
			"     show this help and exit.\n"
			"\n");
}

/* execute send request to dtsd and receive result from it. */
static
int exec_send_recv(const struct options * opt, struct results_head * rhead)
{
	int ret = 0, i = 0, result_cnt = 0;
	char ** result_strv = NULL;

	/* send request to dtsd. */
	/* if opt->cmd is OPT_ADD_CONF, append the -m and -p option. */
	if (cmethod.send_request(opt) != 0) {
		eprintf(HERE "failed to send request.\n");
		return -1;
	}
	debug_printf(HERE "send request success.\n");

	/* correct the all result. */
	if (cmethod.receive_result(&result_strv, &result_cnt) != 0) {
		eprintf(HERE "failed to receive all results.\n");
		return -1;
	}
	debug_printf(HERE "receive result success.\n");

	/* convert to tailq. */
	if (strarr2resultq((const char **)result_strv, result_cnt, rhead) != 0) {
		eprintf(HERE "failed to add all results.\n");
		ret = -1;
		goto free4fail;
	}
	debug_printf(HERE "convert string array to result queue success.\n");

	/* judge the all result if "OK" */
	if (judge_resultq(rhead) != 0) {
		eprintf(HERE "some dtsd failed to exec request.\n");
		ret = -1;
		goto free4fail;
	}
	debug_printf(HERE "all dtsd's result is ok.\n");

	ret = 0;
	goto free4success;

free4fail:
free4success:
	for (i = 0; i < result_cnt; i++) {
		free(result_strv[i]);
		result_strv[i] = NULL;
	}
	free(result_strv);

	return ret;
}

/* get the current tried proxy port from dtsd. */
static
int get_proxy_port()
{
	int ret = -1, tmp_port = -1;
	char cmd_str[MAX_OPT_LEN] = "";
	struct options opt;
	struct results_head rhead = TAILQ_HEAD_INITIALIZER(rhead);
	struct processed_results * result = NULL;

	memset(&opt, 0, sizeof(struct options));

	sprintf(cmd_str, "%s", OPT_GET_PROXY_PORT);

	if (str2opt(cmd_str, &opt, INNER_OPT_PARSE) < 0) {
		eprintf(HERE "str2opt failed. str: %s\n", cmd_str);
		return -1;
	}

	if (exec_send_recv(&opt, &rhead) != 0) {
		eprintf(HERE "failed to get current max proxy port.\n");
		ret = -1;
		goto free4fail;
	}

	for (result = TAILQ_FIRST(&rhead); result != NULL; result = TAILQ_NEXT(result, entry)) {
		if (result->msg == NULL && strlen(result->msg) == 0) {
			eprintf(HERE "returned proxy port is invalid format.\n");
			continue;
		}

		/* if no proxy port bound at dtsd,         */
		/* then dtsd returns the 0 as proxy_port.  */
		tmp_port = atoi(result->msg);
		if (tmp_port >= ret) {
			ret = tmp_port;
		}
	}

	goto free4success;

free4fail:
free4success:
	free_resultq(&rhead);
	return ret;
}

/* decide the proxy port using opt->host_srv */
dstatic
int decide_proxy_port(const struct options * opt)
{
	int srv_port, proxy_port = 0;
	char * tp = NULL;

	/* get port from host_srv. host_srv is following format. */
	/*                                                       */
	/*   192.168.1.1:80                                      */
	tp = strstr(opt->host_srv, TOKEN_COLON);
	if (tp == NULL) {
		eprintf(HERE "failed to split. str:\n", opt->host_srv);
		return -1;
	}

	srv_port = atoi(tp + 1);

	if (srv_port > MAX_VAL_PROXY_PORT) {
		eprintf(HERE "invalid srv_port: %d\n", srv_port);
		return -1;
	}

	proxy_port = srv_port + MIN_VAL_PROXY_PORT;

	while (proxy_port > MAX_VAL_PROXY_PORT) {
		proxy_port = MIN_VAL_PROXY_PORT + (proxy_port % MAX_VAL_PROXY_PORT);
	}

	return proxy_port;
}

/* send bind/unbind socket request to dtsd.                                     */
/* bind/unbind repeats until it reached the MAX_TRY_BIND_PROXY or bind succeed. */
/* if succeed, set the proxy port to opt->proxy_port.                           */
static
int bind_sock_request(struct options * opt)
{
	int ret, i, proxy_port = 0;
	char bind_cmd_str[MAX_OPT_LEN] = "";
	char unbind_cmd_str[MAX_OPT_LEN] = "";
	struct options bind_opt;
	struct options unbind_opt;
	struct results_head bind_rhead = TAILQ_HEAD_INITIALIZER(bind_rhead);
	struct results_head unbind_rhead = TAILQ_HEAD_INITIALIZER(unbind_rhead);

	memset(&bind_opt, 0, sizeof(struct options));
	memset(&unbind_opt, 0, sizeof(struct options));

	if ((proxy_port = get_proxy_port()) <= 0) {
		debug_printf(HERE "proxy_port is <= 0. then decide by dtsadmin.\n");
		if ((proxy_port = decide_proxy_port(opt)) < 0) {
			eprintf(HERE "failed to decide proxy port.\n");
			return -1;
		}
	}
	debug_printf(HERE "proxy port decided: %d\n", proxy_port);

	for (i = 0; i < MAX_TRY_BIND_PROXY; i++) {
		/* remove the all results if exist. */
		free_resultq(&bind_rhead);
		free_resultq(&unbind_rhead);

		sprintf(bind_cmd_str, "%s %s %d", OPT_BIND_SOCK, OPT_PROXY_PORT, proxy_port);
		sprintf(unbind_cmd_str, "%s %s %d", OPT_UNBIND_SOCK, OPT_PROXY_PORT, proxy_port);

		if (str2opt(bind_cmd_str, &bind_opt, INNER_OPT_PARSE) < 0) {
			eprintf(HERE "str2opt failed. str: %s\n", bind_cmd_str);
			return -1;
		}

		if (str2opt(unbind_cmd_str, &unbind_opt, INNER_OPT_PARSE) < 0) {
			eprintf(HERE "str2opt failed. str: %s\n", unbind_cmd_str);
		}

		if (exec_send_recv(&bind_opt, &bind_rhead) == 0) {
			opt->proxy_port = proxy_port;
			ret = 0;
			debug_printf(HERE "dtsd bound proxy port: %d\n", proxy_port);
			goto free4success;
		}

		if (exec_send_recv(&unbind_opt, &unbind_rhead) == 0) {
			free_resultq(&unbind_rhead);
		} else {
			eprintf(HERE "failed to unbind socket request. proxy port: %d\n", proxy_port);
		}

		proxy_port++;
		if (proxy_port > MAX_VAL_PROXY_PORT) {
			proxy_port = MIN_VAL_PROXY_PORT;
		}
	}
	ret = -1;
	goto free4fail;

free4fail:
free4success:
	free_resultq(&bind_rhead);
	free_resultq(&unbind_rhead);
	return ret;
}

/* send set macaddr request to dtsd.                                            */
/* if succeed, set the maccaddr to opt->macaddr.                                */
static
int set_macaddr_request(struct options * opt)
{
	int ret = 0;
	char macaddr_cmd_str[MAX_OPT_LEN] = "";
	struct options macaddr_opt;
	struct processed_results * result = NULL;
	struct results_head rhead = TAILQ_HEAD_INITIALIZER(rhead);

	memset(&macaddr_opt, 0, sizeof(struct options));

	sprintf(macaddr_cmd_str, "%s %s %s", OPT_SET_MACADDR, OPT_IFACE, opt->iface);

	if (str2opt(macaddr_cmd_str, &macaddr_opt, INNER_OPT_PARSE) < 0) {
		eprintf(HERE "str2opt failed. str: %s\n", macaddr_cmd_str);
		return -1;
	}
	if (exec_send_recv(&macaddr_opt, &rhead) != 0) {
		eprintf(HERE "failed to set macaddr request.\n");
		return -1;
	}

	result = TAILQ_FIRST(&rhead);
	if (result == NULL) {
		eprintf(HERE "result is not exist.\n");
		ret = -1;
		goto free4fail;
	}

	/* get the macaddr and set it to opt->macaddr. */
	strncpy(opt->macaddr, result->msg, MAX_LEN_MACADDR);
	if (opt->macaddr[strlen(opt->macaddr) - 1] == '\n') {
		opt->macaddr[strlen(opt->macaddr) - 1] = '\0';
	}

	ret = 0;
	goto free4success;

free4fail:
free4success:
	free_resultq(&rhead);
	return ret;
}

/* prepare for send option. for example, bind/unbind socket. */
static
int prepare_for_send(struct options * opt)
{
	switch(opt->cmd) {
		case ADD_CONF:
			if (set_macaddr_request(opt) != 0) {
				eprintf(HERE "failed to set macaddr request.\n");
				return -1;
			}
			debug_printf(HERE "set_macaddr_request success.\n");
			if (bind_sock_request(opt) != 0) {
				eprintf(HERE "failed to bind socket request.\n");
				return -1;
			}
			debug_printf(HERE "bind_sock_request success.\n");
			break;
		case ADD_NODE:
		case DEL_NODE:
		case LIST:
		case RESTORE:
		case CLEAR:
		case SAVE:
		case NONE:
		default:
			break;
	}
	return 0;
}

static
int set_com_method()
{
	struct stat st;

	memset(&st, 0, sizeof(struct stat));

	if (stat(DTSD_FIFO_IN, &st) != 0 || S_ISFIFO(st.st_mode) != 1) {
		cmethod.init = dtsadmin_init;
		cmethod.send_request = dtsadmin_send_request;
		cmethod.receive_result = dtsadmin_receive_result;
		cmethod.fin = dtsadmin_finalize;
		debug_printf(HERE "communication method: heartbeat.\n");
	} else {
		cmethod.init = dtsadmin_init_fifocom;
		cmethod.send_request = dtsadmin_send_request_fifocom;
		cmethod.receive_result = dtsadmin_receive_result_fifocom;
		cmethod.fin = dtsadmin_fin_fifocom;
		debug_printf(HERE "communication method: fifo.\n");
	}

	return 0;
}

dstatic
void ** make2Darray(int x, int y, int unit_size)
{
	void ** res;
	int i;

	res = (void **) malloc ((sizeof(void *)) * x);
	if (res == NULL) {
		eprintf(HERE "No memory available.\n");
		return NULL;
	}

	for (i = 0; i < x; i++) {
		res[i] = (char *) malloc (unit_size * y);
		if (res[i] == NULL) {
			eprintf(HERE "No memory available.\n");
			goto free4fail;
		}
		memset(res[i], 0, sizeof(char) * y);
	}
	return res;

free4fail:
	for (i = 0; i < x; i++) {
		free(res[i]);
		res[i] = NULL;
	}
	free(res);
	res = NULL;

	return NULL;
}

dstatic
int sconf2AddNodeCmdStrv(const struct server_confs * sconf, char ** cmd_strv[], int * cnt)
{
	int i = 0, ret = 0;
	char ** res = NULL;
	struct node_info_head nihead = TAILQ_HEAD_INITIALIZER(nihead);
	struct node_info * ni = NULL;

	/* node info not exist. this is not err. */
	if (strlen(sconf->nodes) == 0) {
		return 0;
	}

	if ((res = (char **) make2Darray(MAX_LEN_LIST_STRV,
					 MAX_LEN_LIST_STRV_VAL,
					 sizeof(char))) == NULL) {
		eprintf(HERE "No memory available.\n");
		return -1;
	}

	if (str2ninfoq(sconf->nodes, &nihead) != 0) {
		eprintf(HERE "str2ninfoq failed. str: %s\n", sconf->nodes);
		ret = -1;
		goto free4fail;
	}
	debug_printf(HERE "str2ninfoq success.%s\n");

	*cnt = 0;
	char cur_node_name[MAX_LEN_HOST] = "";
	char vnode_str[MAX_LEN_NODES] = "";
	vnode_idset cur_vnode_idset = 0;
	for (ni = TAILQ_FIRST(&nihead); ni != NULL; ni = TAILQ_NEXT(ni, entry)) {
		// the node has no vnode id.
		if (strcmp(ni->id, "0") == 0) {
			snprintf((char *) res[*cnt], MAX_OPT_LEN,
						"%s %s:%s %s",
						OPT_ADD_NODE,
						sconf->srv_addr, sconf->srv_port,
						ni->hb_node_name);
			(*cnt)++;
			continue;
		}
		// set the node name.
		if (strlen(cur_node_name) == 0) {
			strncpy(cur_node_name, ni->hb_node_name, MAX_LEN_HOST);
		}
		if (strncmp(cur_node_name, ni->hb_node_name, MAX_LEN_HOST) == 0) {
			// current node setting. add the vnode id.
			append_IDset(atoi(ni->id), cur_vnode_idset);

		} else {
			// new node setting.
			if (serialize_idset(cur_vnode_idset, vnode_str) != 0) {
				eprintf("failed to serialize the idset.\n");
				goto free4fail;
			}

			snprintf((char *) res[*cnt], MAX_OPT_LEN,
						"%s %s:%s %s%s%s",
						OPT_ADD_NODE,
						sconf->srv_addr, sconf->srv_port,
						cur_node_name, TOKEN_SLASH, vnode_str);
			(*cnt)++;

			memset(cur_node_name, 0, sizeof(char) * MAX_LEN_HOST);
			memset(vnode_str, 0, sizeof(char) * MAX_LEN_NODES);
			cur_vnode_idset = 0;

			strncpy(cur_node_name, ni->hb_node_name, MAX_LEN_HOST);
			append_IDset(atoi(ni->id), cur_vnode_idset);
		}
		if (TAILQ_NEXT(ni, entry) == NULL) {
			if (serialize_idset(cur_vnode_idset, vnode_str) != 0) {
				eprintf("failed to serialize the idset.\n");
				goto free4fail;
			}
			snprintf((char *) res[*cnt], MAX_OPT_LEN,
						"%s %s:%s %s%s%s",
						OPT_ADD_NODE,
						sconf->srv_addr, sconf->srv_port,
						cur_node_name, TOKEN_SLASH, vnode_str);
			(*cnt)++;
		}
	}
	ret = 0;
	goto free4success;
free4fail:
	for (i = 0; i < MAX_LEN_LIST_STRV; i++) {
		free(res[i]);
		res[i] = NULL;
	}
	free(res);
	res = NULL;
	*cmd_strv = NULL;

	free_ninfoq(&nihead);
	return ret;
free4success:
	for (i = *cnt; i < MAX_LEN_LIST_STRV; i++) {
		free(res[i]);
		res[i] = NULL;
	}
	*cmd_strv = res;
	free_ninfoq(&nihead);
	return ret;
}

static
int sconf2AddConfCmdStr(struct server_confs * sconf, char ** addConfCmdStr)
{
	char * res = NULL;

	if (sconf == NULL) {
		eprintf(HERE "server conf is null.\n");
		return -1;
	}

	res = (char *) malloc ((sizeof(char)) * MAX_LEN_LIST_STRV_VAL);
	if (res == NULL) {
		eprintf(HERE "malloc failed.\n");
		return -1;
	}
	memset(res, 0, sizeof(char) * MAX_LEN_LIST_STRV_VAL);

	snprintf(res, MAX_LEN_LIST_STRV_VAL,
			"%s %s:%s %s %s %s %s %s %s %s %s %s \"%s\"",
			OPT_ADD_CONF, sconf->srv_addr, sconf->srv_port,
			OPT_IFACE, sconf->iface,
			OPT_NETMASK, sconf->netmask,
			OPT_FAKEMAC, sconf->fakemac,
			OPT_RI, sconf->ri,
			OPT_RI_OPT, sconf->ri_opt);

	*addConfCmdStr = res;

	return 0;
}

/* print the results for "-S". */
static
int print_setting_cmds(struct processed_results * result)
{
	int ret = 0;
	char ** cmd_strv = NULL;
	struct server_confs_head shead = TAILQ_HEAD_INITIALIZER(shead);
	struct server_confs * sconf = NULL;

	/* no server conf exist. */
	if (strlen(result->msg) == 0) {
		return 0;
	}

	if (str2sconfq(result->msg, &shead) != 0) {
		eprintf(HERE "str2sconfq failed. str: %s\n", result->msg);
		return -1;
	}
	debug_printf(HERE "str2sconfq success.\n");

	for (sconf = TAILQ_FIRST(&shead); sconf != NULL; sconf = TAILQ_NEXT(sconf, entry)) {
		int i = 0, cnt = 0; 
		char * addConfCmdStr;
		if (sconf2AddConfCmdStr(sconf, &addConfCmdStr) != 0) {
			eprintf(HERE "sconf2SconfStr failed.\n");
			ret = -1;
			goto free4fail;
		}
		if (sconf2AddNodeCmdStrv(sconf, &cmd_strv, &cnt) != 0) {
			eprintf(HERE "sconf2AddNodeCmdStrv failed.\n");
			ret = -1;
			goto free4fail;
		}
		fprintf(stdout, "%s\n", addConfCmdStr);
		for (i = 0; i < cnt; i++) {
			fprintf(stdout, "%s\n", cmd_strv[i]);
			free(cmd_strv[i]);
			cmd_strv[i] = NULL;
		}
		if (cmd_strv != NULL) {
			free(cmd_strv);
		}
		cmd_strv = NULL;
		free(addConfCmdStr);
		addConfCmdStr = NULL;
	}

	ret = 0;
	goto free4success;

free4fail:
free4success:
	free_sconfq(&shead);
	return ret;
}

dstatic
int sconf2NodeInfoStrv(const struct server_confs * sconf,
			int nodeWithHash,
			char ** list_strv[],
			int * cnt)
{
	int i = 0, ret = 0;
	char ** res = NULL;
	struct node_info_head nihead = TAILQ_HEAD_INITIALIZER(nihead);
	struct node_info * ni = NULL;

	/* node info not exist. this is not err. */
	if (strlen(sconf->nodes) == 0) {
		return 0;
	}
	if ((res = (char **) make2Darray(MAX_LEN_LIST_STRV,
					 MAX_LEN_LIST_STRV_VAL,
					 sizeof(char))) == NULL) {
		eprintf(HERE "malloc failed.\n");
		return -1;
	}
	if (str2ninfoq(sconf->nodes, &nihead) != 0) {
		eprintf(HERE "str2ninfoq failed. str: %s\n", sconf->nodes);
		ret = -1;
		goto free4fail;
	}
	*cnt = 0;
	if (nodeWithHash == 1) {
		// sorted node info by kernel module.
		for (ni= TAILQ_FIRST(&nihead); ni != NULL; ni = TAILQ_NEXT(ni, entry)) {
			if (ni->id == NULL || strlen(ni->id) == 0) {
				snprintf(res[*cnt], MAX_LEN_LIST_STRV_VAL,
							"  -> %s",
							 ni->hb_node_name);
			} else {
				snprintf(res[*cnt], MAX_LEN_LIST_STRV_VAL,
							"  -> %s %s %s",
							ni->hb_node_name,
							ni->id,
							ni->id_hash);
			}
			(*cnt)++;
		}
	} else {
		int id = 0, id_cnt = 0;
		char id_str[MAX_LEN_VNODE_ID];
		struct node_info_vidset_head nivhead = TAILQ_HEAD_INITIALIZER(nivhead);
		struct node_info_vidset * niv;

		memset(&nivhead, 0, sizeof(struct node_info_vidset_head));
		if (ninfoq2ninfo_vidsetq(&nihead, &nivhead)) {
			eprintf(HERE "ninfoq2ninfo_vidsetq failed.");
			ret = -1;
			goto free4fail;
		}
		for (niv = TAILQ_FIRST(&nivhead); niv != NULL; niv = TAILQ_NEXT(niv, entry)) {
			snprintf(res[*cnt], MAX_LEN_LIST_STRV_VAL,
						"  -> %s",
						niv->hb_node_name);
			for_each_idset(id, niv->idset) {
				memset(id_str, 0, MAX_LEN_VNODE_ID);
				snprintf(id_str, MAX_LEN_VNODE_ID, "%d", id);
				if (id == 0) {
					// no id exist.
					break;
				}
				if (id_cnt == 0) {
					strcat(res[*cnt], " ");
				} else {
					strcat(res[*cnt], ",");
				}
				strcat(res[*cnt], id_str);
				id_cnt++;
			}
			id_cnt = 0;
			(*cnt)++;
		}
		free_ninfo_vidsetq(&nivhead);
	}

	ret = 0;
	goto free4success;
free4fail:
	for (i = 0; i < MAX_LEN_LIST_STRV; i++) {
		free(res[i]);
		res[i] = NULL;
	}
	free(res);
	res = NULL;
	*list_strv = NULL;
	free_ninfoq(&nihead);
	return ret;
free4success:
	for (i = *cnt; i < MAX_LEN_LIST_STRV; i++) {
		free(res[i]);
		res[i] = NULL;
	}
	*list_strv = res;

	free_ninfoq(&nihead);
	return ret;
}

static
int sconf2SconfStr(struct server_confs * sconf, char ** addConfCmdStr, int verbose)
{
	char * res = NULL;

	if (sconf == NULL) {
		eprintf(HERE "server conf is null.\n");
		return -1;
	}

	res = (char *) malloc ((sizeof(char)) * MAX_LEN_LIST_STRV_VAL);
	if (res == NULL) {
		eprintf(HERE "malloc failed.\n");
		return -1;
	}
	memset(res, 0, sizeof(char) * MAX_LEN_LIST_STRV_VAL);

	if (verbose == 1) {
		snprintf(res, MAX_LEN_LIST_STRV_VAL,
				"%s:%s %s %s %s %s %s \"%s\" %s",
					sconf->srv_addr,
					sconf->srv_port,
					sconf->iface,
					sconf->netmask,
					sconf->fakemac,
					sconf->macaddr,
					sconf->ri,
					sconf->ri_opt,
					sconf->proxy_port);
	} else {
		snprintf(res, MAX_LEN_LIST_STRV_VAL,
				"%s:%s %s %s",
					sconf->srv_addr,
					sconf->srv_port,
					sconf->ri,
					sconf->proxy_port);
	}

	*addConfCmdStr = res;

	return 0;
}

/* print the results for "-L". */
static
int print_setting_list(struct processed_results * result, int nodeWithHash, int verbose)
{
	int ret = 0;
	char ** node_info_strv = NULL;
	struct server_confs_head shead = TAILQ_HEAD_INITIALIZER(shead);
	struct server_confs * sconf = NULL;

	/* no server conf exist. */
	if (strlen(result->msg) == 0) {
		return 0;
	}

	if (str2sconfq(result->msg, &shead) != 0) {
		eprintf(HERE "str2sconfq failed. str: %s\n", result->msg);
		return -1;
	}

	/* print the header */
	if (verbose == 1) {
		if (nodeWithHash == 1) {
			fprintf(stdout, "%s\n", SETTING_HEADER_VERBOSE_HASH);
		} else {
			fprintf(stdout, "%s\n", SETTING_HEADER_VERBOSE_NOHASH);
		}
	} else if (nodeWithHash == 1) {
		fprintf(stdout, "%s\n", SETTING_HEADER_HASH);
	} else {
		fprintf(stdout, "%s\n", SETTING_HEADER_NOHASH);
	}

	for (sconf = TAILQ_FIRST(&shead); sconf != NULL; sconf = TAILQ_NEXT(sconf, entry)) {
		int i = 0, cnt = 0;
		char * sconfStr;
		/* get the string of server setting. */
		if ((sconf2SconfStr(sconf, &sconfStr, verbose)) != 0) {
			eprintf(HERE "sconf2SconfStr failed.\n");
			ret = -1;
			goto free4fail;
		}
		/* get the node information. */
		if (sconf2NodeInfoStrv(sconf, nodeWithHash, &node_info_strv, &cnt) != 0) {
			eprintf(HERE "sconf2NodeInfoStrv failed.\n");
			ret = -1;
			goto free4fail;
		}
		/* print the server setting and node information. */
		fprintf(stdout,"%s\n", sconfStr);
		for (i = 0; i < cnt; i++) {
			fprintf(stdout, "%s\n", node_info_strv[i]);
			free(node_info_strv[i]);
			node_info_strv[i] = NULL;
		}
		free(node_info_strv);
		node_info_strv = NULL;
		free(sconfStr);
		sconfStr = NULL;
	}

	ret = 0;

free4fail:
	free_sconfq(&shead);
	return ret;
}

static
int print_result(const struct options * opt, struct results_head * rhead)
{
	struct processed_results * result = NULL;

	if (opt == NULL) {
		return -1;
	}

	result = TAILQ_FIRST(rhead);
	if (result == NULL) {
		eprintf(HERE "result is null.\n");
		return -1;
	}

	switch(opt->cmd) {
		case LIST:
			print_setting_list(result, opt->chash, opt->verbose);
			break;
		case SAVE:
			print_setting_cmds(result);
			break;
		/* for debug. */
#if DEBUG
		case SHOWDATE:
			printf("msg: %s\n", result->msg);
			break;
#endif
		case ADD_CONF:
		case ADD_NODE:
		case DEL_NODE:
		case RESTORE:
		case CLEAR:
		case NONE:
		default:
			break;
	}
	return 0;
}

static
int do_process(struct options * opt)
{
	struct results_head rhead = TAILQ_HEAD_INITIALIZER(rhead);

	/* prepare to send request. */
	if (prepare_for_send(opt) != 0) {
		eprintf(HERE "prepare_for_send failed.\n");
		return -1;
	}
	debug_printf(HERE "prepare_for_send success.\n");

	/* execute send and receive. */
	if (exec_send_recv(opt, &rhead) != 0) {
		eprintf(HERE "exec_send_recv failed.\n");
		return -1;
	}
	debug_printf(HERE "exec_send_recv success.\n");

	/* print the result if needed. */
	if (print_result(opt, &rhead) != 0) {
		eprintf(HERE "print_result failed.\n");
		return -1;
	}
	debug_printf(HERE "print_result success.\n");

	free_resultq(&rhead);

	debug_printf(HERE "do_process success.\n");

	return 0;
}

/* do process with stdin input.(used by "-R" option.) */
static
int do_process_stdin()
{
	int i = 0, cmd_cnt = 0;
	char clear_cmd_str[MAX_OPT_LEN] = "";
	char buf[MAX_LINE_BUFSIZ] = "";
	char cmds[MAX_LEN_CMD_STRV][MAX_LEN_CMD_STRV_VAL];
	struct options opt;
	struct options clear_opt;
	struct results_head clear_rhead = TAILQ_HEAD_INITIALIZER(clear_rhead);

	memset(&opt, 0, sizeof(struct options));
	memset(&clear_opt, 0, sizeof(struct options));

	fprintf(stdout, "preparing for the restore...\n");

	sprintf(clear_cmd_str, "%s", OPT_CLEAR);
	if (str2opt(clear_cmd_str, &clear_opt, INNER_OPT_PARSE) < 0) {
		eprintf(HERE "str2opt failed. str: %s\n", clear_cmd_str);
		return -1;
	}
	if (exec_send_recv(&clear_opt, &clear_rhead) == 0) {
		debug_printf(HERE "failed to get current max proxy port.\n");
	}

	fprintf(stdout, "preparing done.\n");
	fprintf(stdout, "Ctrl-D to exit input.\n");

	/* read the commands from stdin. */
	while (1) {
		if (fgets(buf, BUFSIZ, stdin) == NULL) {
			debug_printf(HERE "input from stdin end.\n");
			break;
		}
		if (cmd_cnt > MAX_LEN_CMD_STRV) {
			eprintf("over the max command num. allowed the %d cmd.\n", MAX_LEN_CMD_STRV);
			return -1;
		}
		/* trim the line feed code if exist. */
		if (buf[strlen(buf) - 1] == '\n') {
			buf[strlen(buf) - 1] = '\0';
		}
		if (strlen(buf) == 0) {
			continue;
		} else {
			strcpy(cmds[cmd_cnt], buf);
			cmd_cnt++;
		}
	}

	fprintf(stdout, "restoring the configuration...\n");

	/* execute commands. */
	for (i = 0; i < cmd_cnt; i++) {
		if (str2opt(cmds[i], &opt, OUTER_OPT_PARSE) != 0 ||
			opt.is_inner_cmd == INNER_CMD) {
			fprintf(stderr, "failed to cmd: %s\n", cmds[i]);
			fprintf(stderr, "failed to restore.\n");
			return -1;
		}
		if (do_process(&opt) != 0) {
			fprintf(stderr, "failed to cmd: %s\n", cmds[i]);
			fprintf(stderr, "failed to restore.\n");
			return -1;
		}
	}

	fprintf(stdout, "restoring done.\n");

	return 0;
}

/* if dtsadmin still runnning, returns 1 else returns 0. */
static
int is_already_running()
{
	struct stat st;

	memset(&st, 0, sizeof(struct stat));

	if (stat(DTSADMIN_PID_FILE, &st) == 0) {
		return 1;
	}
	return 0;
}

static
int initialize()
{
	FILE *fp = NULL;

	fp= fopen(DTSADMIN_PID_FILE, "w");
	if (fp == NULL) {
		perror("fopen");
		return -1;
	}

	fprintf(fp, "%d", getpid());

	if (fclose(fp)) {
		perror("fclose");
		return -1;
	}

	if (set_com_method() != 0) {
		eprintf(HERE "set_com_method failed.\n");
		return -1;
	}

	/* initialize for communication. */
	if (cmethod.init(MODULE_PREFIX, 0) != 0) {
		eprintf(HERE "initialize communication method failed.\n");
		return -1;
	}

	return 0;
}

static
int finalize()
{
	struct stat st;

	memset(&st, 0, sizeof(struct stat));

	if (stat(DTSADMIN_PID_FILE, &st) != 0) {
		eprintf(HERE "pid file not exist: %s.\n", DTSADMIN_PID_FILE);
		return -1;
	}
	if (unlink(DTSADMIN_PID_FILE) != 0) {
		perror("unlink");
		return -1;
	}

	/* finalize for communication to exit. */
	if (cmethod.fin != NULL) {
		if (cmethod.fin(MODULE_PREFIX, 0) != 0) {
			eprintf(HERE "cmethod.fin failed.\n");
			return -1;
		}
	}
	debug_printf(HERE "cmethod.fin success.\n");

	return 0;
}

static
void sig_handler(int signum)
{
	debug_printf(HERE "signal received: %d.\n", signum);
	finalize();
	exit(DTSADMIN_RCV_SIGNAL);
}

static
int setup_signals()
{
	if (signal(SIGINT, sig_handler) == SIG_ERR) {
		perror("signal");
		return -1;
	}
	if (signal(SIGTERM, sig_handler) == SIG_ERR) {
		perror("signal");
		return -1;
	}
	return 0;
}

int main(int argc, char * argv[])
{
	int ret = DTSADMIN_OK;
	struct options opt;

	memset(&opt, 0, sizeof(struct options));
	memset(&cmethod, 0, sizeof(struct com_method));

	if (argc < 2) {
		print_usage();
		return 0;
	} else if (strncmp("--help", *(argv + 1), strlen("--help")) == 0 ||
		strncmp("-h", *(argv + 1), strlen("-h")) == 0) {
		print_usage();
		return 0;
	}

	if (is_already_running() == 1) {
		ret = DTSADMIN_ALRDY_RUNNING;
		eprintf(HERE "dtsadmin already running.\n");
		/* don't finalize! because pid file will be removed! */
		return ret;
	}

	if (setup_signals() != 0) {
		ret = DTSADMIN_INTRNL_ERROR;
		eprintf(HERE "setup_signal failed.\n");
		goto FINALIZE;
	}

	if (initialize() != 0) {
		ret = DTSADMIN_INIT_FAILED;
		eprintf(HERE "initialize failed.\n");
		goto FINALIZE;
	}

	if (optv2opt(argv + 1, &opt, OUTER_OPT_PARSE) != 0||
		opt.is_inner_cmd == INNER_CMD) {
		print_usage();
		ret = DTSADMIN_INVLD_ARGS;
		eprintf(HERE "optv2opt failed.\n");
		goto FINALIZE;
	}
	debug_printf(HERE "optv2opt success.\n");

	debug_flag = opt.debug;

	if (opt.cmd == RESTORE) {
		if (do_process_stdin() != 0) {
			ret = DTSADMIN_INTRNL_ERROR;
			eprintf(HERE "do_process_stdin failed.\n");
			goto FINALIZE;
		}
	} else {
		if (do_process(&opt) != 0) {
			ret = DTSADMIN_INTRNL_ERROR;
			eprintf(HERE "do_process failed.\n");
			goto FINALIZE;
		}
	}

	ret = DTSADMIN_OK;
	goto FINALIZE;

FINALIZE:
	finalize();

	return ret;
}

