#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <pthread.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <signal.h>

#include "common.h"
#include "dtsd_command_exec.h"
#include "dtsd_common.h"
#include "dtsd_util.h"
#include "dtsd_ri_plugins.h"
#include "dtsd_plugin_loader.h"
#include "dtsd_mediate.h"
#include "dtsd_optparser.h"
#include "dtsd_worker.h"

extern int dtsd_worker(struct dtsd_config *);


static void usage()
{
	struct ri_intf_container * cont;
	struct ri_intf * intf = NULL;
	fprintf(stderr,
			"Distribute TCP Splicing Cluster Administrator Daemon.\n"
			"Usage dtsd [options] <iface>:\n"
			"		<iface>\n"
			"			specify an interface on internal network.\n"
			"		--daemonize\n"
			"			Run as daemon.\n"
			"		-f|--force\n"
			"			Force to run dtsd.\n"
			"		-h|--help\n"
			"			Show this help.\n");
	for_each_ri_intf_set(intf, cont){
		if(intf != NULL) {
			fprintf(stderr,
					"Available Request Inspector\n");
		}
		fprintf(stderr,
				"	%s\n	Usage: %s\n", intf->name, intf->usage());
	}
}

static void usage_and_exit(int ret) {
	usage();
	exit(ret);
}

static void daemonize()
{
	pid_t pid, sid;

	if(chdir("/") != 0) {
		perror("chdir");
		exit(1);
	}

	if((pid = fork()) == -1){
		perror("fork");
		exit(1);
	}

	if(pid != 0) {
		exit(0);
	}

	if((sid = setsid()) < 0) {
		perror("setsid");
		exit(1);
	}

	FILE * file;
	if((file = fopen(DTSD_PID_FILE, "w")) == NULL) {
		perror("fopen");
		exit(1);
	};

	fprintf(file, "%d", getpid());
	fclose(file);

	umask(0);

	fclose(stdin);
	fclose(stdout);
	fclose(stderr);
}

static int dtsctl_flush()
{
        char flush_cmd[WORK_BUF_SIZE] = "";

        snprintf(flush_cmd, WORK_BUF_SIZE, "%s/dtsctl flush", DTSCTL_PATH);
        if (system(flush_cmd) != 0) {
		return -1;
	}
	return 0;
}

static int dtsd_parse_opt(int argc, char * argv[], struct dtsd_config * conf)
{
	int ret = 0;
	assert(argc != 0);
	conf->progname = *argv;
	conf->interface = NULL;
	conf->debug_flag = 0;
	conf->force_flag = 0;
	conf->daemonize_flag = 0;
	conf->help_flag = 0;
	conf->ready_flag = 0;
	conf->server_conf = NULL;
	conf->ipaddr = NULL;
	conf->mynode = NULL;
	argv++;
	for(; *argv != NULL; argv++) {
		if(strcmp(*argv, "-d") == 0 || strcmp(*argv, "--debug") == 0) {
			conf->debug_flag = 1;
		} else if(strcmp(*argv, "-f") == 0 || strcmp(*argv, "--force") == 0) {
			conf->force_flag = 1;
		} else if(strcmp(*argv, "--daemonize") == 0) {
			conf->daemonize_flag = 1;
		} else if(strcmp(*argv, "-h") == 0|| strcmp(*argv, "--help") == 0) {
			conf->help_flag = 1;
		} else if(conf->interface == NULL) {
			conf->interface = *argv;
			ret = 1;
		} else {
			ret = 0;
		}
	}
	return ret;
}

static int dtsd_init(struct dtsd_config * conf)
{
	int fd;
	struct ifreq ifr;
	char * addr;

	if((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}

	ifr.ifr_addr.sa_family = AF_INET;
	strncpy(ifr.ifr_name, conf->interface, IFNAMSIZ-1);

	if(ioctl(fd, SIOCGIFADDR, &ifr) < 0) {
		perror("ioctl");
		exit(1);
	}
	close(fd);

	addr = (char *) malloc((sizeof(char)) * MAX_LEN_SRV_ADDR);
	if (addr == NULL) {
		fprintf(stderr, HERE "failed to allocate.");
	}

	conf->naddr  = (((struct sockaddr_in *)&ifr.ifr_addr))->sin_addr.s_addr;
	strcpy(addr, inet_ntoa((((struct sockaddr_in *)&ifr.ifr_addr))->sin_addr));
	conf->ipaddr = addr;
	debug_fprintf(stderr, "internal ipaddr = %s\n", conf->ipaddr);

	return 0;
}

void thread_finalize()
{
	/* remove the fifos for mediator thread. */
	remove_fifo();
}

void thread_master(struct dtsd_config * conf) {
	int err;
	void * ret;
	pthread_t mediator_thread;

	if((err = pthread_create(&mediator_thread, NULL, dtsd_mediate, (void*)conf)) != 0) {
		fprintf(stderr, HERE "failed to create a thread.");
		perror(HERE "pthread_create");
		exit(1);
	}

	if(dtsd_worker(conf) != 0) {
		exit(1);
	}

	if ((err = pthread_join(mediator_thread, &ret)) != 0) {
		fprintf(stderr, HERE "failed to join a thread.");
		perror(HERE "pthread_create");
		exit(1);
	}

	thread_finalize();
}

static
void sig_handler(int signum)
{
	if (signum == SIGINT || signum == SIGTERM) {
		dtsctl_flush();
		thread_finalize();
		exit(1);
	}
}

static
int setup_signals()
{
	if (signal(SIGINT, sig_handler) == SIG_ERR) {
		perror("signal");
		return -1;
	}
	if (signal(SIGTERM, sig_handler) == SIG_ERR) {
		perror("signal");
		return -1;
	}
	return 0;
}

int main(int argc, char *argv[])
{
	struct dtsd_config conf;

	memset(&conf, 0, sizeof(struct dtsd_config));

	if (load_plugins() <= 0) {
		fprintf(stderr, "No plugin was loaded.\n");
		return -1;
	}

	/* parse options */
	if(dtsd_parse_opt(argc, argv, &conf) == 0) {
		if(conf.interface == NULL) {
			fprintf(stderr, "Internal inteface not specified.\n");
		} else {
			fprintf(stderr, "Invalid arguments.\n");
		}
		usage_and_exit(1);
	}

	/* check options */
	debug_flag = conf.debug_flag;

	if(conf.help_flag == 1) {
		usage_and_exit(0);
	}

	if(conf.daemonize_flag == 1) {
		daemonize();
	}

	if (setup_signals() != 0) {
		eprintf(HERE "setup_signal failed.\n");
		return -1;
	}

	/* initialize */
	if(dtsd_init(&conf)) {
		return -1;
	}

	/* flush the consistent hashing. */
	if (dtsctl_flush() != 0) {
		return -1;
	}

	/* create threads */
	thread_master(&conf);

	return 0;
}
