/* 
 * dtsd: DTS node monitoring daemon
 * by Hajime Fujita <hfujita@cc.u-tokyo.ac.jp>
 *
 * based on api_test, Copyright (C) 2000 Alan Robertson <alanr@unix.sh>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdint.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>

#include <heartbeat.h>
#include <ha_msg.h>
#include <hb_api.h>

#include "common.h"
#include "dtsd_common.h"
#include "dtsd_util.h"
#include "dtsd_optparser.h"
#include "dtsd_worker.h"
#include "dtsd_server_conf.h"
#include "dtsd_command_exec.h"
#include "dtsadmin_dtsdio.h"
#include "dtsd_ri_plugins.h"

// heartbeat message type
#define DTSD_SERVER_CONF_RES "dtsd_server_conf_res"
#define DTSD_SERVER_CONF_REQ "dtsd_server_conf_req"
#define DTSD_HBC_SYNC        "dtsd_hbc_sync"
#define DTSD_READY           "dtsd_ready"
#define DTSD_CMD_REQUEST     "dtsd_cmd_request"
#define DTSD_CMD_RESULT      "dtsd_cmd_result"

// heartbeat message key
#define DTSD_MESSAGE         "dtsd_message"

struct hbc_node_container
{
	struct hbc_node_container * next;
	struct hbc_node * hbc_node;
};

static struct {struct hbc_node_container * next; struct hbc_node_container * prev;}
	HBC_NODE_CONTANER = {.next = NULL, .prev = NULL};

static struct hbc_node mynode;

int set_mynode_state(struct dtsd_config * conf, int status)
{
	conf->mynode->ready = status;
	return 0;
}

int get_mynode_state(const struct dtsd_config * conf)
{
	return conf->mynode->ready;
}

#define for_each_hbc_node(cont) \
	for(cont = HBC_NODE_CONTANER.next; cont != NULL; cont = cont->next)

void dtsd_add_node(struct dtsd_config * dconf, struct hbc_node * node)
{
	struct server_conf * i;
	struct dtsd_node_info * j;
	for_each_server_conf(i, dconf->server_conf) {
		for_each_node_info(j, i) {
			if(strcmp(j->hb_node_name, node->name) == 0) {
				exec_NodeInfo(j, i, NULL, NULL); // discards command output
				if(i->ri_intf != NULL)
					i->ri_intf->add_node(i->ri, node->addr);
			}
		}
	}
}

void dtsd_remove_node(struct dtsd_config * conf, struct hbc_node * node)
{
	struct server_conf * i;
	debug_printf(HERE "remove virtual nodes of %s\n", node->name);
	for_each_server_conf(i, conf->server_conf) {
		if(node->ready == 1) {
			if(disable_hbc_node(i, node->addr, NULL, NULL) != 0) { // discards command output
				eprintf(HERE "failed to remove virtual nodes of %s\n", node->name);
			} else {
				debug_printf(HERE "succeeded to remove virtual nodes of %s\n", node->name);
			}
			/* i->ri is null when ri_intf->initialize failed. */
			if(i->ri_intf != NULL && i->ri != NULL)
				i->ri_intf->remove_node(i->ri, node->addr);
		}
	}
}

static ll_cluster_t *hb;

struct hbc_node * find_hbc_node(const char * name)
{
	struct hbc_node_container * next;

	if(name == NULL) {
		eprintf(HERE "heartbeat host name not specified.\n");
		return NULL;
	}
	for(next = HBC_NODE_CONTANER.next; next != NULL; next = next->next) {
		if(strcmp(next->hbc_node->name, name) == 0) {
			return next->hbc_node;
		}
	}
	debug_printf(HERE "no heartbeat host named %s\n", name);
	return NULL;
}

dstatic struct hbc_node * find_ready_hbc_node()
{
	struct hbc_node_container * next;
	for(next = HBC_NODE_CONTANER.next; next != NULL; next = next->next) {
		if(next->hbc_node->ready) {
			return next->hbc_node;
		}
	}
	debug_printf(HERE "no ready dtsd named\n");
	return NULL;
}

dstatic struct hbc_node * add_hbc_node(const struct hbc_node new_node)
{
	struct hbc_node * hbc_node;
	struct hbc_node_container * cont;
	hbc_node = (struct hbc_node *) malloc (sizeof(struct hbc_node));
	if(hbc_node == NULL) {
		eprintf(HERE "No Memory Available\n");
		return NULL;
	}

	cont = (struct hbc_node_container *) malloc (sizeof(struct hbc_node_container));
	if(cont == NULL) {
		eprintf(HERE "No Memory Available\n");
		free(hbc_node);
		return NULL;
	}

	cont->hbc_node = hbc_node;
	if(HBC_NODE_CONTANER.next == NULL) {
		HBC_NODE_CONTANER.next = cont;
		cont->next = NULL;
	} else {
		cont->next = HBC_NODE_CONTANER.next;
		HBC_NODE_CONTANER.next = cont;
	}
	hbc_node->name = strdup(new_node.name);;
	hbc_node->addr = new_node.addr;
	hbc_node->ready = new_node.ready;
	return hbc_node;
}

dstatic int update_hbc_node(const struct hbc_node new_node)
{
	struct hbc_node * node;
	if((node = find_hbc_node(new_node.name)) == NULL) {
		if(add_hbc_node(new_node) == NULL) {
			return -1;
		}
	} else {
		node->addr = new_node.addr;
		node->ready = new_node.ready;
	}
	return 0;
}

struct hbc_node * find_or_add_hbc_node(const char * name)
{
	struct hbc_node * node;
	if((node = find_hbc_node(name)) == NULL) {
		if (name == NULL || strlen(name) == 0) {
			eprintf(HERE "name is NULL or length 0.\n");
			return NULL;
		}
		struct hbc_node new_node;
		new_node.name = strdup(name);
		new_node.addr = -1;
		new_node.ready = -1;
		return add_hbc_node(new_node);
	}
	return node;
}

dstatic int count_hbc_node() {
	struct hbc_node_container * cont;
	int count = 0;
	for_each_hbc_node(cont) {
		if (cont->hbc_node != NULL && cont->hbc_node->ready != -1)
			count++;
	}
	return count;
}

dstatic int count_ready_hbc_node() {
	struct hbc_node_container * cont;
	int count = 0;
	for_each_hbc_node(cont) {
		if (cont->hbc_node != NULL &&
			cont->hbc_node->ready == 1)
			count++;
	}
	return count;
}

struct result_queue_container {
	struct result_queue_container * next;
	struct result_queue_container * prev;
	char * result;
};

struct result_queue_container RESULT_QUEUE =
{
	.next = &RESULT_QUEUE,
	.prev = &RESULT_QUEUE,
	.result = NULL
};

dstatic int is_emptyQueue();
dstatic int enqueue(const char * result);
dstatic char * dequeue();
dstatic void clear_queue();
dstatic int count_queue();

dstatic int is_emptyQueue()
{
	if(RESULT_QUEUE.next == &RESULT_QUEUE) {
		return 1;
	}
	return 0;
}

dstatic int enqueue(const char * result)
{
	struct result_queue_container * cont;
	cont = (struct result_queue_container *) malloc (sizeof(struct result_queue_container));
	if(cont == NULL) {
		eprintf(HERE "malloc failed.\n");
		return -1;
	}
	if(result == NULL) {
		eprintf(HERE "invalid arugments.\n");
		return -1;
	}

	cont->result = strdup(result);
	if(RESULT_QUEUE.prev == &RESULT_QUEUE) {
		RESULT_QUEUE.next = cont;
		RESULT_QUEUE.prev = cont;
		cont->next = &RESULT_QUEUE;
		cont->prev = &RESULT_QUEUE;
	} else {
		RESULT_QUEUE.prev->next = cont;
		RESULT_QUEUE.prev = cont;
		cont->next = &RESULT_QUEUE;
	}
	return 0;
}

dstatic char * dequeue()
{
	if(RESULT_QUEUE.next == &RESULT_QUEUE) {
		debug_printf(HERE "result queue is empty\n");
		return NULL;
	} else {
		struct result_queue_container * cont;
		char * res;
		cont = RESULT_QUEUE.next;
		cont->next->prev = &RESULT_QUEUE;
		RESULT_QUEUE.next = cont->next;
		res = cont->result;
		free(cont);
		return res;
	}
}

dstatic void clear_queue()
{
	struct result_queue_container * cont;
	while(RESULT_QUEUE.next != &RESULT_QUEUE) {
		cont = RESULT_QUEUE.next;
		RESULT_QUEUE.next = RESULT_QUEUE.next->next;
		debug_printf(HERE "free(%s)\n", cont->result);
		free(cont->result);
		free(cont);
		debug_printf(HERE "go to next loop\n");
	}
	RESULT_QUEUE.next = &RESULT_QUEUE;
	RESULT_QUEUE.prev = &RESULT_QUEUE;
}

dstatic int count_queue()
{
	int num = 0;
	struct result_queue_container * cont;
	for(cont = RESULT_QUEUE.next; cont != &RESULT_QUEUE; cont = cont->next) {
		num++;
	}
	return num;
}

dstatic void print_queue()
{
	int num = 0;
	struct result_queue_container * cont;
	for(cont = RESULT_QUEUE.next; cont != &RESULT_QUEUE; cont = cont->next) {
		num++;
		fprintf(stderr, "cont[%d] = %s\n", num, cont->result);
	}
}

pthread_mutex_t TIMED_QUEUE_LOCK = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t QUEUE_TIMER_LOCK = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t TIMER_LOCK = PTHREAD_MUTEX_INITIALIZER;
int wait_count = 0;

struct timespec timer =
{ .tv_sec = -1, .tv_nsec = -1 };
struct queue_timer_arguments {
	struct timespec timeout;
	struct timespec stride;
	int count;
};

dstatic void * queue_timer(void * args)
{
	void * ret = NULL;
	struct queue_timer_arguments * arg = (struct queue_timer_arguments *)args;
	struct timespec timeout, stride;
	if(args == NULL) {
		pthread_exit(ret);
	}
	timeout = arg->timeout;
	stride = arg->stride;

	if(pthread_mutex_trylock(&QUEUE_TIMER_LOCK) != 0) {
		fprintf(stderr, HERE "dtsd is busy\n");
		pthread_exit(ret);
	}
	for(;;) {
		pthread_mutex_lock(&TIMED_QUEUE_LOCK);
		if(timer.tv_nsec < 0 || timer.tv_sec < 0) {
			eprintf(HERE "dtsd response is expired.\n");
			pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
			break;
		}
		if(wait_count <= 0) {
			pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
			break;
		}
		timer.tv_sec -= stride.tv_sec;
		timer.tv_nsec -= stride.tv_nsec;
		if(timer.tv_nsec < 0) {
			timer.tv_sec--;
			timer.tv_nsec += 999999999;
		}
		pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
		nanosleep(&stride, NULL);
	}

	timer.tv_sec = -1;
	timer.tv_nsec = -1;
	pthread_mutex_unlock(&QUEUE_TIMER_LOCK);
	pthread_exit(ret);
	return ret;
}

dstatic int timed_is_emptyQueue();
dstatic int timed_enqueue(const char * result);
dstatic char * timed_dequeue();
dstatic void timed_clear_queue();
dstatic int timed_count_queue();

dstatic int timed_enqueue(const char * res)
{
	int ret = -1;
	pthread_mutex_lock(&TIMED_QUEUE_LOCK);
	if((timer.tv_sec >= 0 && timer.tv_nsec >= 0) && wait_count > 0) {
		wait_count--;
		enqueue(res);
		ret = 0;
	}
	pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
	return ret;
}

dstatic char * timed_dequeue()
{
	char * res = NULL;
	pthread_mutex_lock(&TIMED_QUEUE_LOCK);
	res = dequeue();
	pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
	return res;
}

dstatic void timed_clear_queue()
{
	pthread_mutex_lock(&TIMED_QUEUE_LOCK);
	clear_queue();
	pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
}

dstatic int timed_is_emptyQueue()
{
	int ret = 0;
	pthread_mutex_lock(&TIMED_QUEUE_LOCK);
	ret = is_emptyQueue();
	pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
	return ret;
}

dstatic int timed_count_queue()
{
	int ret = 0;
	pthread_mutex_lock(&TIMED_QUEUE_LOCK);
	ret = count_queue();
	pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
	return ret;
}

static void flush_hb_msg()
{
	struct ha_msg * msg;
	if(hb != NULL) {
		while((msg = hb->llc_ops->readmsg(hb, 0)) != NULL) {
			debug_printf(HERE "discard not-registered message.\n");
			ha_msg_del(msg);
		}
	}
}

dstatic void start_receive_result(pthread_t *th, struct queue_timer_arguments * args)
{
	flush_hb_msg();
	timed_clear_queue();
	pthread_mutex_lock(&TIMED_QUEUE_LOCK);
	wait_count = args->count;
	timer = args->timeout;
	pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
	if(pthread_create(th, NULL, queue_timer, args) != 0) {
		eprintf("failed to create thread.");
		return;
	}
}

dstatic void wait_all_request(pthread_t th)
{
	void * ret;
	pthread_join(th, &ret);
}

static void send_nop();
static void send_hbc(const char * hostname);
static void on_receive_hbc(struct ha_msg *msg_in, void *private);

static void request_server_config(struct hbc_node * node);
static void send_server_config(struct server_conf * conf, const char * node);
static void on_receive_server_config_req(struct ha_msg *msg_in, void *private);
static void on_receive_server_config_res(struct ha_msg *msg_in, void *private);

static void on_node_up(const char * hostname, void * private);
static void on_node_down(struct hbc_node *node, void * private);
static void on_dtsd_up(const char * hostname);

static int send_dtsd_request_to_node(char * mes);
static int send_dtsd_request(char * mes);
static int send_dtsd_result(const char * node, char * res);

static void on_receive_cmd_request(struct ha_msg *msg_in, void *private);
static void on_receive_cmd_result(struct ha_msg *msg_in, void *private);
static int heartbeat_init(const char * progname, int debug_flag);
static void * hbc_mainloop();

static void send_hbc(const char * hostname)
{
	struct ha_msg * message = NULL;
	struct in_addr in = { .s_addr = mynode.addr, };
	char * addr = inet_ntoa(in);
	int ready = mynode.ready;

	debug_printf(HERE "tried to send " DTSD_HBC_SYNC " to %s.\n", hostname);
	message = ha_msg_new(3);
	ha_msg_add(message, F_NODE, mynode.name);
	ha_msg_add(message, F_TYPE, DTSD_HBC_SYNC);
	ha_msg_add(message, "dtsd_internal_ipaddr", addr);
	ha_msg_add_int(message, "dtsd_ready", ready);
	if (hb->llc_ops->sendnodemsg(hb, message, hostname) != HA_OK) {
		eprintf(HERE "%s\n", hb->llc_ops->errmsg(hb));
	}
	debug_printf(HERE "succeeded to send " DTSD_HBC_SYNC " to %s.\n", hostname);
}

int send_hbc_to_cluster(void)
{
	struct ha_msg * message = NULL;
	struct in_addr in = { .s_addr = mynode.addr, };
	char * addr = inet_ntoa(in);
	int ready = mynode.ready;

	message = ha_msg_new(4);
	ha_msg_add(message, F_NODE, mynode.name);
	ha_msg_add(message, F_TYPE, DTSD_HBC_SYNC);
	ha_msg_add(message, "dtsd_internal_ipaddr", addr);
	ha_msg_add_int(message, "dtsd_ready", ready);
	if (hb->llc_ops->sendclustermsg(hb, message) != HA_OK) {
		eprintf(HERE "%s\n", hb->llc_ops->errmsg(hb));
		return -1;
	}
	debug_printf(HERE "succeeded to send " DTSD_HBC_SYNC " to cluster.\n");
	return 0;
}

static void on_receive_hbc(struct ha_msg *msg_in, void *private)
{
	uint32_t addr = 0; // avoid warning
	int ready, old;
	struct hbc_node * node;
	const char * host;
	const char * addr_str;
	host = ha_msg_value(msg_in, F_NODE);
	addr_str = ha_msg_value(msg_in, "dtsd_internal_ipaddr");
	resolve_host(addr_str, &addr);
	ha_msg_value_int(msg_in, "dtsd_ready", &ready);

	debug_printf(HERE "received " DTSD_HBC_SYNC " from %s\n", host);

	node = find_or_add_hbc_node(host);
	node->addr = addr;
	old = node->ready;
	node->ready = ready;

	if(private != NULL && node->ready == 1) {
		debug_printf(HERE "received " DTSD_HBC_SYNC " from readied node %s\n", host);
		if(mynode.ready == 1) {
			struct dtsd_config * conf = (struct dtsd_config *) private;
			if(old != 1) {
				debug_printf(HERE "node named %s is readied now.\n", node->name);
				dtsd_add_node(conf, node);
			}
		} else {
			debug_printf(HERE "received " DTSD_HBC_SYNC " from %s\n", host);
			request_server_config(node);
		}
	/*
	 * if receive HBC_SYNC and node is not ready,
	 * then remove the node.
	 */
	} else if (private != NULL && old == 1 && node->ready != 1) {
		debug_printf(HERE "received " DTSD_HBC_SYNC " from downed node %s\n", host);
		node->ready = old;
		on_node_down(node, private);
		node->ready = ready;
	}
}

static void on_dtsd_up(const char * hostname)
{
	send_hbc(hostname);
}

static void request_server_config(struct hbc_node * node)
{
	struct ha_msg * message = NULL;
	message = ha_msg_new(3);
	ha_msg_add(message, F_NODE, mynode.name);
	ha_msg_add(message, F_TYPE, DTSD_SERVER_CONF_REQ);
	if (hb->llc_ops->sendnodemsg(hb, message, node->name) != HA_OK) {
		eprintf(HERE "%s\n", hb->llc_ops->errmsg(hb));
	} else {
		debug_printf(HERE "succeeded to send Server Configuration request to %s.\n", node->name);
	}
}

static void send_server_config(struct server_conf * conf, const char * node)
{
	struct ha_msg * message = NULL;
	char str[DTSD_MESSAGE_MAX];

	if(ServerConf2str(conf, str, sizeof(str)) != 0) {
		return;
	}

	message = ha_msg_new(3);
	ha_msg_add(message, F_NODE, mynode.name);
	ha_msg_add(message, F_TYPE, DTSD_SERVER_CONF_RES);
	ha_msg_add(message, "dtsd_server_conf", str);

	if (hb->llc_ops->sendnodemsg(hb, message, node) != HA_OK) {
		debug_printf(HERE "%s\n", hb->llc_ops->errmsg(hb));
	} else {
		debug_printf(HERE "succeeded to send Server Configuration to %s.\n", node);
	}
}

static void on_receive_server_config_req(struct ha_msg *msg_in, void *private)
{
	struct dtsd_config * conf = (struct dtsd_config *) private;
	const char * node = ha_msg_value(msg_in, F_NODE);

	debug_printf(HERE "received " DTSD_SERVER_CONF_REQ " from %s\n", node);
	if(mynode.ready == 0) {
		debug_printf(HERE "But my node is not readied yet, Then ignore requeset.\n", node);
		return;
	}
	send_server_config(conf->server_conf, node);
}

static void on_receive_server_config_res(struct ha_msg *msg_in, void *private)
{
	struct dtsd_config * dtsd_conf = (struct dtsd_config *)private;
	struct server_conf * server_conf = NULL; //avoid warning
	const char * dtsd_server_conf = ha_msg_value(msg_in, "dtsd_server_conf");

	debug_printf(HERE "received " DTSD_SERVER_CONF_RES "\n");
	debug_printf(HERE "current setting is \"%s\"\n", dtsd_server_conf);

	if(str2ServerConf(dtsd_server_conf, &server_conf) != 0) {
		eprintf(HERE "failed to parse Server Configuration.\n");
		return;
	}

	if(dtsd_conf->server_conf == NULL) {
		int old = mynode.ready;
		mynode.ready = 1;
		update_hbc_node(mynode);

		debug_printf(HERE "apply Server Confiugration.\n");
		if(apply_ServerConf(dtsd_conf, server_conf) != 0) {
			eprintf(HERE "failed to apply Server Configuration.\n");
			free_server_conf_list(server_conf);
			dtsd_conf->server_conf = NULL;
			mynode.ready = 0;
			update_hbc_node(mynode);
			return;
		}
		debug_printf(HERE "succeed to apply Server Configuration.\n");
		if(old != 1) {
			struct hbc_node_container * cont;
			struct hbc_node * node;
			mynode.ready = 1;
			update_hbc_node(mynode);
			send_hbc_to_cluster();
		}
	}
}

static void on_node_up(const char * hostname, void * private)
{
	struct hbc_node * node;
	private = NULL; // avoid warning;
	node = find_hbc_node(hostname);
	if(node != NULL) {
		node->ready = 0;
	}
	send_hbc(hostname);
}

static void on_node_down(struct hbc_node *node, void * private)
{
	struct dtsd_config * conf = (struct dtsd_config *)private;
	dtsd_remove_node(conf, node);
	node->ready = -1;
}

static int send_dtsd_request_to_node(char * mes)
{
	struct ha_msg * message = NULL;
	struct hbc_node * node;

	node = find_ready_hbc_node();
	if(node == NULL) {
		debug_printf(HERE "no node is ready\n");
		// NO NODE IS READY
		return -1;
	}

	message = ha_msg_new(3);
	ha_msg_add(message, F_NODE, mynode.name);
	ha_msg_add(message, F_TYPE, DTSD_CMD_REQUEST);
	ha_msg_add(message, DTSD_MESSAGE, mes);
	if (hb->llc_ops->sendnodemsg(hb, message, node->name) != HA_OK) {
		eprintf(HERE "%s\n", hb->llc_ops->errmsg(hb));
		return -1;
	}
	debug_printf(HERE "send message \"%s\" to %s\n", mes, node->name);
	return 0;
}

static int send_dtsd_request(char * mes)
{
	struct ha_msg * message = NULL;

	if(mes == NULL || strlen(mes) <= 1) {
		eprintf(HERE "Invalid Arguments\n");
		return -1;
	}
	if (mes[strlen(mes) - 1] == '\n') {
		mes[strlen(mes) - 1] = '\0';
	}

	message = ha_msg_new(3);
	ha_msg_add(message, F_NODE, mynode.name);
	ha_msg_add(message, F_TYPE, DTSD_CMD_REQUEST);
	ha_msg_add(message, DTSD_MESSAGE, mes);

	if (hb->llc_ops->sendclustermsg(hb, message) != HA_OK) {
		eprintf(HERE "%s\n", hb->llc_ops->errmsg(hb));
		return -1;
	}
	debug_printf(HERE "send message \"%s\"\n", mes);

	return 0;
}

static int send_dtsd_result(const char * node, char * res)
{
	struct ha_msg * message = NULL;
	message = ha_msg_new(3);
	ha_msg_add(message, F_NODE, mynode.name);
	ha_msg_add(message, F_TYPE, DTSD_CMD_RESULT);
	ha_msg_add(message, DTSD_MESSAGE, res);
	if (hb->llc_ops->sendnodemsg(hb, message, node) != HA_OK) {
		debug_printf(HERE "%s\n", hb->llc_ops->errmsg(hb));
		return 1;
	}
	return 0;
}

static void on_receive_cmd_request(struct ha_msg *msg_in, void *private)
{
	struct hbc_node * node;
	const char * nodename = ha_msg_value(msg_in, F_NODE);
	const char * cmd = ha_msg_value(msg_in, DTSD_MESSAGE);
	char res[MAX_MSG_SIZE] = ""; // avoid warning
	struct dtsd_config * conf = (struct dtsd_config *) private;

	debug_printf(HERE "Receipt command request from %s.\n", nodename);

	sprintf(res, ">>\nHOST-%s\nRESULT-NG\nERR-internal error occured.\nMSG-\n",
			conf->hostname);

#if 0
	if(mynode.ready == 0) {
		debug_printf(HERE "Receipt command request but not readied yet.\n");
		return;
	}
#endif

	node = find_hbc_node(nodename);
	if(node == NULL) {
		debug_printf(HERE "No node found.\n");
	}

	dtsd_command_exec(conf, cmd, res);
	if(send_dtsd_result(nodename, res) != 0) {
		eprintf(HERE "failed to send command result.\n");
	} else {
		debug_printf(HERE "succeeded sending command result to %s.\n", nodename);
	}
}

static void on_receive_cmd_result(struct ha_msg *msg_in, void *private)
{
	struct hbc_node * node;
	const char * nodename = ha_msg_value(msg_in, F_NODE);
	const char * res = ha_msg_value(msg_in, DTSD_MESSAGE);
	debug_printf(HERE "received command result from %s\n", nodename);
	if(private != NULL) {
		// struct dtsd_config * conf = (struct dtsd_config *) private;
		node = find_hbc_node(nodename);
		if(node == NULL) {
			return;
		}
	}
	if(res == NULL)
		eprintf(HERE "receipt command result without message\n");
	timed_enqueue(res);
}



static int count_hb_join_nodes()
{
	int num = 0, i, ret = -1;
	const char * node, * status;
	int timeout = 3000; /* ms */

	if (hb->llc_ops->init_nodewalk(hb)!=HA_OK) {
		eprintf(HERE "Cannot init node walk\n");
		eprintf(HERE "REASON: %s\n", hb->llc_ops->errmsg(hb));
		return -1;
	}

	while ((node=hb->llc_ops->nextnode(hb)) != NULL) {
		status = hb->llc_ops->client_status(hb, node, "dtsd", timeout);
		debug_printf(HERE "node %s has status %s\n", node, status);
		if(status != NULL && strcmp(status, ONLINESTATUS) == 0) {
			num++;
		}
	}

	if (hb->llc_ops->end_nodewalk(hb) != HA_OK) {
		eprintf(HERE "Cannot end node walk\n");
		eprintf(HERE "REASON: %s\n", hb->llc_ops->errmsg(hb));
		return -1;
	}

	return num;
}

dstatic void clear_globals ()
{
	hb = NULL;
	timed_clear_queue();
}

static void
NodeStatus(const char * name, const char * status, void * private)
{
	debug_printf("Status update: Node %s now has status %s\n", name, status);

	if (strcmp(status, DEADSTATUS) == 0) {
		struct hbc_node *node;

		node = find_hbc_node(name);
		if (node != NULL)
			on_node_down(node, private);
		else
			debug_printf("NodeStatus: find_hbc_node returned NULL\n");

	} else if (strcmp(status, ACTIVESTATUS) == 0) {
		on_node_up(name, private);
	}
}

static void
ClientStatus(const char * name, const char * client,
			 const char * status, void * private)
{
	debug_printf("Status update: Client %s/%s now has status [%s]\n",
			name, client, status);

	if (strcmp(status, JOINSTATUS) == 0) {
		on_dtsd_up(name);
	} else if (strcmp(status, LEAVESTATUS) == 0) {
		struct hbc_node *node;
		node = find_hbc_node(name);
		if (node != NULL)
			on_node_down(node, private);
	} else if (strcmp(status, DEADSTATUS) == 0) {
		struct hbc_node *node;
		node = find_hbc_node(name);
		if (node != NULL)
			on_node_down(node, private);
	}
}

pthread_mutex_t RCV_LOCK = PTHREAD_MUTEX_INITIALIZER;

int dtsd_send_and_exec(const struct options * opt, char ** res[], int * resnum)
{
	char str[MAX_OPT_LEN];
	char ** tmp;
	int count, num, i;
	pthread_t th;
	struct queue_timer_arguments args = {
		.timeout = {.tv_sec = 5, .tv_nsec = 0, },
		.stride = {.tv_sec = 0, .tv_nsec = 100000000, },
		.count = 0,
	};

	pthread_mutex_lock(&RCV_LOCK); // stop receive ha_msg
	{
		switch(opt->cmd) {
		case CLEAR:
		case RESTORE:
#if DEBUG
		case SHOWDATE:
#endif
			count = count_hbc_node();
			break;
#if 0
		case SET_MACADDR:
			count = 1;
			break;
#endif
		default:
			count = count_ready_hbc_node();
			break;
		}

		if(count == 0) {
			* res = NULL;
			* resnum = 0;
			debug_printf(HERE "No node receive the commands.\n");
			pthread_mutex_unlock(&RCV_LOCK);
			return 0;
		}

		args.count = count;
		if(opt2str((struct options *)opt, str, INNER_OPT_PARSE) != 0) {
			eprintf(HERE "failed to convert message\n");
			pthread_mutex_unlock(&RCV_LOCK);
			return -1;
		}
		debug_printf(HERE "succeeded to convert message\n");

		start_receive_result(&th, &args);

		if(send_dtsd_request(str) != 0) {
			eprintf(HERE "failed to send message\n");
			pthread_mutex_unlock(&RCV_LOCK);
			return -1;
		}
	}
	pthread_mutex_unlock(&RCV_LOCK); // start receive ha_msg

	debug_printf(HERE "succeeded to send message\n");

	wait_all_request(th);
	num = timed_count_queue();
	tmp = (char **) malloc ((sizeof(char*)) * (num + mynode.ready));
	for(i = 0; i < num; i++) {
		tmp[i] = timed_dequeue();
	}

	switch(opt->cmd) {
	case SAVE:
	case LIST:
	case SET_MACADDR:
		if (num != 0)
			num = 1;
		break;
	default:
		break;
	}

	* res = tmp;
	* resnum = num;
	return 0;
}

static void * hbc_mainloop()
{
	struct timespec stride = {.tv_sec = 0, .tv_nsec = 1000000, };
	for(;;) {
		pthread_mutex_lock(&RCV_LOCK);
		if(hb->llc_ops->rcvmsg(hb, 0)) { // not blocking
			debug_printf(HERE "received not-registered message.\n");
		}
		pthread_mutex_unlock(&RCV_LOCK);
		nanosleep(&stride, NULL);
	}
	return NULL;
}

static int heartbeat_init(const char * progname, int debug_flag)
{
	int ret;
	const char *clname = "dtsd";

	if (hb != NULL)
		return EINVAL;

	cl_log_set_entity(progname);
	cl_log_enable_stderr(debug_flag);
	cl_log_set_facility(LOG_USER);

	if((hb = ll_cluster_new("heartbeat")) == NULL) {
		eprintf("Cannot allocate the cluster object\n");
		return ENOMEM;
	}

	ret = ENOMEM;
	if (hb->llc_ops->signon(hb, clname)!= HA_OK) {
		eprintf("Cannot sign on with heartbeat\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_delete;
	}

	return 0;

out_delete:
	hb->llc_ops->delete(hb);

	clear_globals();

	return ret;
}

int heartbeat_delete()
{
	hb->llc_ops->delete(hb);
	return 0;
}

int heartbeat_signoff()
{
	hb->llc_ops->signoff(hb, TRUE);
	return heartbeat_delete();
}

int dtsd_worker(struct dtsd_config * conf)
{
	const char * nodename;
	if(heartbeat_init(conf->progname, conf->debug_flag) != 0) {
		eprintf(HERE "failed to initialize Heartbeat API\n");
		return -1;
	}

	debug_printf(HERE "succeeded initialization of Heartbeat API\n");

	if (hb->llc_ops->
		set_nstatus_callback(hb, NodeStatus, (void*)conf) !=HA_OK){
		eprintf("Cannot set node status callback\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
	}

	if (hb->llc_ops->
		set_cstatus_callback(hb, ClientStatus, (void*)conf)!=HA_OK){
		eprintf("Cannot set client status callback\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
	}

	if (hb->llc_ops->
		set_msg_callback(hb, DTSD_HBC_SYNC,
						 on_receive_hbc, (void*)conf)!=HA_OK){
		eprintf("Cannot set " DTSD_HBC_SYNC " callback\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_signoff;
	}

	if (hb->llc_ops->
		set_msg_callback(hb, DTSD_SERVER_CONF_RES,
						 on_receive_server_config_res, (void*)conf)!=HA_OK){
		eprintf("Cannot set " DTSD_SERVER_CONF_RES " callback\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_signoff;
	}

	if (hb->llc_ops->
		set_msg_callback(hb, DTSD_SERVER_CONF_REQ,
						 on_receive_server_config_req, (void*)conf)!=HA_OK){
		eprintf("Cannot set " DTSD_SERVER_CONF_REQ "callback\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_signoff;
	}

	if (hb->llc_ops->
		set_msg_callback(hb, DTSD_CMD_REQUEST,
						 on_receive_cmd_request, (void*)conf)!=HA_OK){
		eprintf("Cannot set " DTSD_CMD_REQUEST " callback\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_signoff;
	}

	if (hb->llc_ops->
		set_msg_callback(hb, DTSD_CMD_RESULT,
						 on_receive_cmd_result, (void*)conf)!=HA_OK){
		eprintf("Cannot set " DTSD_CMD_RESULT " callback\n");
		eprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_signoff;
	}

	debug_printf(HERE "set all callbacks\n");

	nodename = hb->llc_ops->get_mynodeid(hb);
	mynode.name = strdup(nodename);
	mynode.addr = conf->naddr;
	mynode.ready = 0;

	add_hbc_node(mynode);
	conf->mynode = &mynode;
	conf->hostname = strdup(nodename);

	debug_printf(HERE "entering hbc mainloop.\n");

	send_hbc_to_cluster();

	if(hbc_mainloop() != 0) {
		debug_printf(HERE "exit from hbc mainloop.\n");
		return -1;
	}
	
	// NEVER REACHED
	return 0;

out_signoff:
	heartbeat_signoff();
	return 1;
}

int dtsadmin_init(const char * progname, int debug_flag)
{
	const char * nodename;

	if(heartbeat_init(progname, debug_flag) != 0) {
		eprintf(HERE "failed to initialize Heartbeat API\n");
		return -1;
	}

	debug_printf(HERE "dtsadmin: succeeded initialization of Heartbeat API\n");

	nodename = hb->llc_ops->get_mynodeid(hb);
	mynode.name = strdup(nodename);

	return 0;

out_signoff:
	heartbeat_signoff();
	return 1;
}

int DTSADMIN_RES_COUNT = 0;
int dtsadmin_send_request(const struct options * opt)
{
	char str[MAX_OPT_LEN];
	opt2str((struct options *)opt, str, INNER_OPT_PARSE);

	DTSADMIN_RES_COUNT = count_hb_join_nodes() - 1;
	debug_printf(HERE "send request to %d node.\n", DTSADMIN_RES_COUNT);

	flush_hb_msg();
	if(send_dtsd_request(str) != 0) {
		eprintf(HERE "dtsadmin: failed to send request");
	}
	return 0;
}

int dtsadmin_receive_result(char ** result[], int * result_num)
{
	int i = 0;
	char ** results;
	struct timespec timeout = {.tv_sec = 5, .tv_nsec = 0, };
	struct timespec stride = {.tv_sec = 0, .tv_nsec = 10000000, };
	results = (char **) malloc ((sizeof(char *)) * MAX_HBC_NODES);

	debug_printf(HERE "receiving result\n");

	while(i < DTSADMIN_RES_COUNT) {
		struct ha_msg * msg;
		const char * type, * res;

		if((msg = hb->llc_ops->readmsg(hb, 0)) != NULL) {
			type = ha_msg_value(msg, F_TYPE);
			res = ha_msg_value(msg, DTSD_MESSAGE);
			if(type != NULL && res != NULL && strcmp(type, DTSD_CMD_RESULT) == 0) {
				debug_printf(HERE "receive result \"%s\"\n", res);
				results[i] = strdup(res);
				i++;
			}
		}
		ha_msg_del(msg);

		if(i >= DTSADMIN_RES_COUNT)
			break;

		timeout.tv_sec -= stride.tv_sec;
		timeout.tv_nsec -= stride.tv_nsec;
		if(timeout.tv_nsec < 0) {
			timeout.tv_sec--;
			timeout.tv_nsec += 999999999;
		}
		if(timeout.tv_nsec < 0 || timeout.tv_sec < 0) {
			eprintf(HERE "dtsd response is expired.\n");
			pthread_mutex_unlock(&TIMED_QUEUE_LOCK);
			break;
		}
		nanosleep(&stride, NULL);
	}

	debug_printf(HERE "received result\n");

	*result = results;
	*result_num = i;
	return 0;
}

void send_nop()
{
	struct ha_msg * message = NULL;

	debug_printf(HERE "tried to send NOP.\n");
	message = ha_msg_new(3);
	ha_msg_add(message, F_TYPE, "DTSD_NOP");
	if (hb->llc_ops->sendclustermsg(hb, message) != HA_OK) {
		eprintf(HERE "%s\n", hb->llc_ops->errmsg(hb));
	}
	debug_printf(HERE "succeeded to send NOP.\n");
}

int dtsadmin_finalize()
{
	return 0;
}
