#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <assert.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>

#include <ftcs.h>

#include "common.h"

#define DEFAULT_DISPATCHER "rr"
#define HANDOVER_RETRY_MAX 3

#define IPVSADM_ADD_FMT "/sbin/ipvsadm -a -t %s:%d -r %s -m"
#define IPVSADM_DEL_FMT "/sbin/ipvsadm -d -t %s:%d -r %s"

#define MEMBER_FILE_DELIM " \t"

#define MAINLOOP_TIMEOUT_SEC 0
#define MAINLOOP_TIMEOUT_USEC 100000

static struct httpri_dispatcher_mgr *dispatcher_mgrs[] = {
	&rr_dispatcher_mgr,
//	&specweb_dispatcher_mgr,
	NULL,
};

static struct httpri_dispatcher *find_dispatcher(const char *name)
{
	int i;

	for (i = 0; dispatcher_mgrs[i] != NULL; i++) {
		if (strcmp(dispatcher_mgrs[i]->name, name) == 0)
			return dispatcher_mgrs[i]->create(name);
	}

	return NULL;
}

static void fprints(FILE *fp, const char *str, int len) 
{
	int i;

	for (i = 0; i < len; i++) {
		char c = str[i];
		if (!isprint(c))
			c = '.'; 
		fputc(c, fp); 
	}    
}

struct httpri_context {
	char *buffer;
	int buff_data_len;
	int buff_size;
	int buff_cur_head;
	int buff_next_line_head;

	int socket;
};

struct server_param {
	int lsocket;
	uint16_t target_port; /* in NBO */
};

struct thread_param {
	int id;
	pthread_t thread;
	int alive;
	struct ri * ri;
};

static struct httpri_context *ctx_alloc(void)
{
	struct httpri_context *ctx;

	ctx = calloc(1, sizeof(*ctx));
	if (ctx == NULL)
		return NULL;

	ctx->buffer = malloc(HTTPRI_DEFAULT_BUFSIZE);
	if (ctx->buffer == NULL) {
		free(ctx);
		return NULL;
	}

	ctx->buff_size = HTTPRI_DEFAULT_BUFSIZE;

	return ctx;
}

static void ctx_flush(struct httpri_context *ctx)
{
	ctx->buff_data_len = 0;
	ctx->buff_cur_head = 0;
	ctx->buff_next_line_head = 0;
	ctx->socket = -1;
}

static void ctx_free(struct httpri_context *ctx)
{
	free(ctx->buffer);
	free(ctx);
}

static inline int ctx_tailroom(const struct httpri_context *ctx)
{
	return ctx->buff_size - ctx->buff_data_len;
}

static inline int is_crlf(const char *str)
{
	return str[0] == '\r' && str[1] == '\n';
}

static inline int is_sp(char c)
{
	return c == ' ';
}

static inline int is_lws(char c)
{
	return is_sp(c) || c == '\t';
}

static int ctx_expand_buffer(struct httpri_context *ctx)
{
	char *new_buff;

	new_buff = realloc(ctx->buffer, ctx->buff_size*2);
	if (new_buff == NULL)
		return -ENOMEM;

	ctx->buffer = new_buff;
	ctx->buff_size *= 2;

	return 0;
}

static int recv_line(struct httpri_context *ctx, const char **line_p)
{
	ssize_t r;
	int i;
	int line_len;
	int line_cont = 0;

	for (;;) {
		for (i = ctx->buff_cur_head; i+1 < ctx->buff_data_len; i++) {
			if (is_crlf(ctx->buffer+i)) {
				if (i+2 < ctx->buff_data_len && is_sp(ctx->buffer[i+2])) {
					line_cont = 1;
					i += 2;
					continue;
				}
				line_len = i - ctx->buff_next_line_head;
				goto found;
			}
		}
		ctx->buff_cur_head = i;

		r = recv(ctx->socket,
				 ctx->buffer + ctx->buff_data_len, ctx_tailroom(ctx), 0);
		if (r < 0) {
			if (errno == EINTR)
				continue;
			return -errno;
		}
		else if (r == 0) {
			dprintf(2, "read() returned 0\n");
			return 0;
		}
		else {
			dprintf(3, "read () returned: ");
			dprints(3, ctx->buffer + ctx->buff_data_len, r);
			dprintf(3, "\n");

			ctx->buff_data_len += r;
			if (ctx_tailroom(ctx) == 0) {
				int e = ctx_expand_buffer(ctx);
				if (e)
					return e;
			}
		}
	}

found:
	*line_p = ctx->buffer + ctx->buff_next_line_head;
	ctx->buff_next_line_head += line_len + 2;
	ctx->buff_cur_head = ctx->buff_next_line_head;

	return line_len;
}

static int parse_request(const char *line, int line_len,
						 struct http_request *req)
{
	int i;
	struct http_request r;

	r.method = line;
	for (i = 0; i < line_len; i++)
		if (is_sp(line[i]))
			break;

	if (i >= line_len)
		return -EINVAL;

	r.method_len = i;
	r.uri = line + i + 1;

	for (++i; i < line_len; i++)
		if (is_sp(line[i]))
			break;

	if (i >= line_len)
		return -EINVAL;

	r.uri_len = i - r.method_len - 1;
	r.version = line + i + 1;
	r.version_len = line_len - r.method_len - r.uri_len - 2;

	*req = r;
	
	return 0;
}

static char *format_time(const struct tm *tm, char *buff)
{
	sprintf(buff, "%02d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->tm_sec);
	return buff;
}

static void httpri_main(const struct thread_param * param)
{
	struct httpri_context *ctx;
	struct ri * ri = param->ri;
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;
	struct server_param * server_param = uniq_param->server_param;

	ctx = ctx_alloc();
	if (ctx == NULL) {
		fprintf(stderr, "ERROR: Failed to allocate ctx\n");
		exit(1);
	}

	for (;;) {
		struct sockaddr_in peer, dest;
		socklen_t peerlen = sizeof(peer), destlen = sizeof(dest);
		const char *line = NULL;
		int line_len;
		struct http_request req;
		char addr_buff[INET_ADDRSTRLEN];
		struct timeval tv0, tv1;
		int failed = 0;
		int rval = 0;
		fd_set rfds;
		struct timeval tv;

		FD_ZERO(&rfds);
		FD_SET(server_param->lsocket, &rfds);

		tv.tv_sec = MAINLOOP_TIMEOUT_SEC;
		tv.tv_usec = MAINLOOP_TIMEOUT_USEC;

		dprintf(2, "httpri[%d]: waiting for an incoming connection...\n",
			param->id);

		rval = select(server_param->lsocket + 1, &rfds, NULL, NULL, &tv);

		if (rval < 1) {
			if (param->alive == 0) {
				ctx_flush(ctx);
				pthread_exit(NULL);
			}
			continue;
		}

		ctx->socket = accept(server_param->lsocket,
							 (struct sockaddr *) &peer, &peerlen);
		if (ctx->socket < 0) {
			switch (errno) {
			case EINTR:
			case ECONNABORTED:
				continue;
			}
			fprintf(stderr, "httpri[%d]: ERROR: accept() FAILED: %s\n",
					param->id, strerror(errno));
			break;
		}

		dprintf(2, "httpri[%d]: accepted connection from %s:%d\n",
				param->id,
				inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

		if ((line_len = recv_line(ctx, &line)) < 0)
			goto close_cont;

		if (parse_request(line, line_len, &req) < 0)
			goto close_cont;

		dprintf(2, "httpri[%d]: ", param->id);
		dprintf(2, "request received\n");
		dprintf(3, "  method=\"");
		dprints(3, req.method, req.method_len);
		dprintf(3, "\" uri=\"");
		dprints(3, req.uri, req.uri_len);
		dprintf(3, "\" version=\"");
		dprints(3, req.version, req.version_len);
		dprintf(3, "\"\n");

	retry:
		dest.sin_family = AF_INET;
		dest.sin_addr.s_addr = dispatcher->dispatch(ri, &req);
		dest.sin_port = server_param->target_port;

		inet_ntop(AF_INET, &dest.sin_addr, addr_buff, sizeof(addr_buff));

		dprintf(2, "httpri[%d]: trying to migrate the connection to %s\n",
				param->id, addr_buff);

		gettimeofday(&tv0, NULL);

		if (ftcs_request_handover(ctx->socket,
								  (struct sockaddr *) &dest,
								  destlen, ctx->buffer,
								  ctx->buff_data_len, 0) < 0) {
			struct tm tm0, tm1;
			char buf0[10], buf1[10];

			gettimeofday(&tv1, NULL);
			dprintf(0,
					"httpri[%d]: ERROR: request_handover to %s FAILED: %s"
					" (%s-%s)\n",
					param->id, addr_buff, strerror(errno),
					format_time(localtime_r(&tv0.tv_sec, &tm0), buf0),
					format_time(localtime_r(&tv1.tv_sec, &tm1), buf1));

			if (errno == ETIMEDOUT) {
				dprintf(0, "httpri[%d]: %s may be dead. remove it.\n",
						param->id, addr_buff);
				dispatcher->remove_node(ri, dest.sin_addr.s_addr);
				if (failed <= HANDOVER_RETRY_MAX)
					goto retry;
			} else if (errno == EINPROGRESS || errno == ECONNREFUSED) {
				/* There is a chance that
				   handover to another host may succeed */
				failed++;
				if (failed <= HANDOVER_RETRY_MAX)
					goto retry;
			}
			goto close_cont;
		}

		dprintf(2, "httpri[%d]: connection handover to %s succeeded.\n",
				param->id, addr_buff);

	close_cont:
		ctx_flush(ctx);
	}
}

static void *thread_entry(void * p)
{
	httpri_main(p);

	return NULL;
}

static void print_dispatcher_list(void)
{
	int i;

	fprintf(stderr, "List of dispatchers:\n");
	for (i = 0; dispatcher_mgrs[i] != NULL; i++)
		fprintf(stderr, "  %s\n", dispatcher_mgrs[i]->name);
}

static void add_dispatcher_list(char * str)
{
	int i;

	strcat(str, "List of dispatchers: \n");
	for (i = 0; dispatcher_mgrs[i] != NULL; i++) {
		strcat(str, dispatcher_mgrs[i]->name);
		strcat(str, " ");
	}
}

const char usage_str[] = "[options]\n"
	"\n"
	"Options:\n"
	"  -c threads      specifies the number of server threads (default: 1)\n"
	"  -d dispatcher   specifies the dispatching algorithm (default: "
	DEFAULT_DISPATCHER ")\n\n";

const char const * usage(void)
{
	return usage_str;
}

static void ri_free(struct ri * ri)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;

	free(uniq_param->thread_params);
	dispatcher->free_node_info(ri);
	free(uniq_param->dispatcher);
	free(uniq_param->server_param);
	free(ri->uniq_param);
	free(ri);
}

struct ri * initialize (int sock, uint32_t addr, uint16_t port, char * ri_optarg)
{
	int n_threads = 1;
	const char *dispatcher_name = "rr";
	char * tok, * optarg;
	const char delim[] = " 	";
	struct ri * ri = NULL;
	struct httpri_uniq_param * uniq_param = NULL;
	struct httpri_dispatcher * dispatcher = NULL;
	struct server_param * server_param;

	ri = (struct ri*) malloc(sizeof(struct ri));

	optarg = strdup(ri_optarg);
	tok = strtok(optarg, delim);
	do {
		if(tok == NULL || strlen(tok) == 0) {
			break;
		} else if(strcmp(tok, "-c") == 0) {
			char * thdstr = strtok(NULL, delim);
			n_threads = atoi(thdstr);
			if (n_threads < 1) {
				fprintf(stderr, "-c: must be a positive number.\n");
				goto exception;
			}
			break;
		} else if(strcmp(tok, "-d") == 0) {
			dispatcher_name = strtok(NULL, delim);
		}
	} while((tok = strtok(NULL, delim)) != NULL);

	uniq_param = (struct httpri_uniq_param *) malloc (sizeof(struct httpri_uniq_param)*n_threads);
	uniq_param->n_threads = n_threads;

	server_param = (struct server_param*) malloc (sizeof(struct server_param));

	if (listen(sock, DEFAULT_BACKLOG) < 0)
		goto exception;

	server_param->lsocket = sock;
	server_param->target_port = htons(port);
	uniq_param->server_param = server_param;

	dispatcher = find_dispatcher(dispatcher_name);
	if (dispatcher == NULL) {
		fprintf(stderr,
				"%s: specified dispatcher not found\n", dispatcher_name);
		print_dispatcher_list();
		goto exception;
	}
	uniq_param->dispatcher = dispatcher;
	ri->uniq_param = uniq_param;

	if (dispatcher->init(ri, addr)) {
		fprintf(stderr, "dispatcher->init() failed\n");
		goto exception;
	}

	free(optarg);
	return ri;

exception:
	free(ri);
	free(optarg);
	if(uniq_param != NULL)
		free(uniq_param);
	if(dispatcher != NULL)
		free(dispatcher);
	return NULL;
}

int run (struct ri * ri)
{
	int i, err;
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct thread_param * thread_params;
	
	thread_params = (struct thread_param *) malloc(sizeof(struct thread_param) * uniq_param->n_threads);

	for (i = 0; i < uniq_param->n_threads; i++) {
		thread_params[i].id = i;
		thread_params[i].alive = 1;
		thread_params[i].ri = ri;
		err = pthread_create(&thread_params[i].thread,
							 NULL, thread_entry, &thread_params[i]);
		if (err) {
			fprintf(stderr, "pthread_create() failed: %s\n", strerror(err));
			return 1;
		}
	}
	uniq_param->thread_params = thread_params;

	return 0;
}

int stop (struct ri * ri)
{
	int i;
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct thread_param * thread_params = uniq_param->thread_params;

	for (i = 0; i < uniq_param->n_threads; i++)
		thread_params[i].alive = 0;
	return 0;
}

int add_node (struct ri * ri, uint32_t ipaddr)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;

	if (dispatcher->add_node(ri, ipaddr)) {
		fprintf(stderr, "dispatcher->add_node() failed\n");
		return 1;
	}
	return 0;
}

int remove_node (struct ri * ri, uint32_t ipaddr)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;

	dispatcher->remove_node(ri, ipaddr);
	return 0;
}

int finalize (struct ri * ri)
{
	int i;
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct thread_param * thread_params = uniq_param->thread_params;
	struct server_param * server_param = uniq_param->server_param;

	for (i = 0; i < uniq_param->n_threads; i++)
		pthread_join(thread_params[i].thread, NULL);
	close(server_param->lsocket);
	ri_free(ri);

	return 0;
}

const char name[] = "httpri";

struct ri_intf ri_intf = {
	.name = name,
	.initialize = initialize,
	.run = run,
	.stop = stop,
	.add_node = add_node,
	.remove_node = remove_node,
	.finalize = finalize,
	.usage = usage,
};

struct ri_intf *create_ri_intf()
{
	return &ri_intf;
}

