#include <errno.h>
#include <stdlib.h>
#include <pthread.h>

#include "common.h"

struct node {
	struct node *next;
	uint32_t addr;
};

struct node_info {
	struct node *node_list;
	struct node *last_node;
	uint32_t default_target;
};

static int rr_init(struct ri * ri, uint32_t def)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;
	struct node_info * node_info;

	node_info = (struct node_info *)malloc(sizeof(struct node_info));
	if (node_info == NULL)
		return -ENOMEM;
	node_info->default_target = def;
	node_info->node_list = NULL;
	node_info->last_node = NULL;
	dispatcher->node_info = node_info;

	pthread_mutex_init(&uniq_param->node_list_lock, NULL);
	return 0;
}

static uint32_t rr_dispatch(struct ri * ri, const struct http_request *req)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;
	struct node *node;
	struct node_info * node_info = dispatcher->node_info;
	uint32_t addr = node_info->default_target;

	pthread_mutex_lock(&uniq_param->node_list_lock);
	if (node_info->last_node != NULL) {
		node = node_info->last_node->next;
		if (node == NULL)
			node = node_info->node_list;
	}
	else
		node = node_info->node_list;

	node_info->last_node = node;
	if (node)
		addr = node->addr;
	pthread_mutex_unlock(&uniq_param->node_list_lock);

	return addr;
}

static int rr_add_node(struct ri * ri, uint32_t addr)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;
	struct node *node;
	struct node_info * node_info;

	node = malloc(sizeof(*node));
	if (node == NULL)
		return -ENOMEM;

	node->addr = addr;

	pthread_mutex_lock(&uniq_param->node_list_lock);
	node_info = dispatcher->node_info;
	node->next = node_info->node_list;
	node_info->node_list = node;
	pthread_mutex_unlock(&uniq_param->node_list_lock);

	return 0;
}

static void rr_remove_node(struct ri * ri, uint32_t addr)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;
	struct node_info * node_info = dispatcher->node_info;
	struct node *node, *to_del = NULL, **pprev = &node_info->node_list;

	/* fprintf(stderr, "going to remove %08x\n", addr); */

	pthread_mutex_lock(&uniq_param->node_list_lock);
	node = node_info->node_list;
	while (node != NULL) {
		/* fprintf(stderr, "node=%p *pprev=%p node->addr=%08x\n",
		   node, *pprev, node->addr); */
		if (node->addr == addr) {
			if (node == node_info->last_node)
				node_info->last_node = NULL;
			*pprev = node->next;
			to_del = node;
			break;
		}
		pprev = &node->next;
		node = node->next;
	}
	pthread_mutex_unlock(&uniq_param->node_list_lock);

	/* fprintf(stderr, "node_list=%p\n", node_list); */

	if (to_del)
		free(to_del);
}

static void rr_free_node_info(struct ri * ri)
{
	struct httpri_uniq_param * uniq_param = ri->uniq_param;
	struct httpri_dispatcher * dispatcher = uniq_param->dispatcher;
	struct node_info * node_info = dispatcher->node_info;
	struct node * node, * next;

	pthread_mutex_lock(&uniq_param->node_list_lock);
	node = node_info->node_list;
	while (node != NULL) {
		next = node->next;
		free(node);
		node = next;
	}
	pthread_mutex_unlock(&uniq_param->node_list_lock);
}

static struct httpri_dispatcher * rr_create(const char * name)
{
	struct httpri_dispatcher * rr_dispatcher;

	rr_dispatcher = (struct httpri_dispatcher *)malloc(sizeof(struct httpri_dispatcher));
	rr_dispatcher->name     = name;
	rr_dispatcher->init     = rr_init;
	rr_dispatcher->dispatch = rr_dispatch;
	rr_dispatcher->add_node = rr_add_node;
	rr_dispatcher->remove_node = rr_remove_node;
	rr_dispatcher->free_node_info = rr_free_node_info;

	return rr_dispatcher;
}

struct httpri_dispatcher_mgr rr_dispatcher_mgr = {
	.name   = "rr",
	.create = rr_create,
};

