#include <errno.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>

#include "common.h"
#include "md5.h"

#define HASHKEY_LEN 16
#define VNODE_PER_NODE 63

static struct bt_root vnode_tree;
static pthread_rwlock_t vnode_tree_lock;
static uint32_t default_target;

struct hashkey {
	uint8_t hash[HASHKEY_LEN];
};

struct vnode {
	struct bt_node bt_node;
	uint32_t addr;
	int id;
	struct hashkey hash;
};

static inline
int hash_compare(const struct hashkey *l, const struct hashkey *r)
{
	int i;

	for (i = 0; i < HASHKEY_LEN; i++) {
		if (l->hash[i] > r->hash[i])
			return 1;
		else if (l->hash[i] < r->hash[i])
			return -1;
	}

	return 0;
}

static inline void calc_hash(const void *p, size_t len, struct hashkey *key)
{
	md5_state_t state;

	md5_init(&state);
	md5_append(&state, p, len);
	md5_finish(&state, key->hash);
}

static struct vnode *find_vnode(const struct hashkey *key)
{
	struct bt_node *node;

	node = vnode_tree.node;

	while (node) {
		struct vnode *vn = container_of(node, struct vnode, bt_node);
		int cmp = hash_compare(key, &vn->hash);

		if (cmp > 0) {
			if (node->right == NULL)
				return vn;

			node = node->right;
		}
		else if (cmp < 0) {
			if (node->left == NULL) {
				struct bt_node *prev = bt_prev(node);
				if (prev == NULL) {
					prev = bt_last(&vnode_tree);
				}

				return container_of(prev, struct vnode, bt_node);
			}
			node = node->left;
		}
		else
			return vn;
	}

	return NULL;
}

static int add_vnode(struct vnode *vn_new)
{
	struct bt_node **new = &(vnode_tree.node), *parent = NULL;
	int ret = 0;

	while (*new) {
		struct vnode *vn;
		int cmp;

		vn = container_of(*new, struct vnode, bt_node);
		cmp = hash_compare(&vn_new->hash, &vn->hash);

		parent = *new;
		if (cmp < 0)
			new = &((*new)->left);
		else if (cmp > 0)
			new = &((*new)->right);
		else {
			ret = -EEXIST;
			goto out;
		}
	}

	bt_link_node(&vn_new->bt_node, parent, new);
out:
	return ret;
}

static int match(const struct http_request *req, const char *str)
{
	int i;
	int len = strlen(str);

	for (i = 0; i < req->uri_len - len+1; i++) {
		int j;
		for (j = 0; j < len; j++) {
			if (req->uri[i+j] != str[j])
				break;
		}
		if (j == len)
			return 1;
	}

	return 0;
}

static int specweb_init(uint32_t def)
{
	default_target = def;
	pthread_rwlock_init(&vnode_tree_lock, NULL);
	return 0;
}

static uint32_t specweb_dispatch(const struct http_request *req)
{
	struct vnode *vn;
	struct hashkey hash;

	if (!match(req, "/downloads/") || match(req, "dir000000000"))
		return default_target;

	calc_hash(req->uri, req->uri_len, &hash);

	pthread_rwlock_rdlock(&vnode_tree_lock);
	vn = find_vnode(&hash);
	pthread_rwlock_unlock(&vnode_tree_lock);

	if (vn == NULL)
		return default_target;

	return vn->addr;
}

static int specweb_add_node(uint32_t addr)
{
	struct vnode *vn;
	int i;

	vn = malloc(sizeof(*vn) * VNODE_PER_NODE);
	if (vn == NULL)
		return -ENOMEM;

	for (i = 0; i < VNODE_PER_NODE; i++) {
		int e;
		struct {
			uint32_t addr;
			uint8_t  id;
		} __attribute__((packed)) key = { addr, i };
		vn->addr = addr;
		vn->id   = i;
		calc_hash(&key, sizeof(key), &vn->hash);

		pthread_rwlock_wrlock(&vnode_tree_lock);
		e = add_vnode(vn);
		pthread_rwlock_unlock(&vnode_tree_lock);
		if (e)
			dprintf(1, "specweb_add_node: add_vnode failed: %d\n", e);
		vn++;
	}

	return 0;
}

struct httpri_dispatcher specweb_dispatcher = {
	.name     = "specweb",
	.init     = specweb_init,
	.dispatch = specweb_dispatch,
	.add_node = specweb_add_node,
};
