#include <linux/module.h>
#include <linux/inetdevice.h>
#include <net/ip.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include "ftcs.h"
#include "dts_internal.h"

/* TODO: USE HASH INSTEAD OF LINEAR LIST!!! */
static struct list_head configs = LIST_HEAD_INIT(configs);
static DEFINE_MUTEX(configs_mtx);
static DEFINE_RWLOCK(configs_lock);

static struct proc_dir_entry *proc_net_dts_configs;

static struct file_operations vn_seq_fops;

static inline struct ftcs_kconfig *__find_config(__be32 addr, __be16 port)
{
	struct ftcs_kconfig *conf;
	struct ftcs_kconfig *result = NULL;
	list_for_each_entry(conf, &configs, list) {
		if (conf->inter_addr == addr && conf->port == port) {
			result = conf;
			break;
		}
	}

	return result;
}

/*
  TODO: use a netdevice as a key to specify configurations, rather than
  an IP address, since one device may have more than one IP address.
*/
struct ftcs_kconfig *ftcs_find_get_config(__be32 addr, __be16 port)
{
	struct ftcs_kconfig *conf;
	struct ftcs_kconfig *result = NULL;

	read_lock_bh(&configs_lock);
	list_for_each_entry(conf, &configs, list) {
		if (conf->port == port &&
			(conf->inter_addr == addr || conf->intra_addr == addr)) {
#ifdef CONFIG_FTCS_DEBUG
			BUG_ON(atomic_read(&conf->refcnt) < 1);
#endif
			result = conf;
			atomic_inc(&result->refcnt);
			break;
		}
	}
	read_unlock_bh(&configs_lock);

	return result;
}

/* TODO: copy-paste function!!! */
struct ftcs_kconfig *ftcs_find_get_config_from_proxy(__be32 addr,
						__be16 proxy_port)
{
	struct ftcs_kconfig *conf;
	struct ftcs_kconfig *result = NULL;

	read_lock_bh(&configs_lock);
	list_for_each_entry(conf, &configs, list) {
		if (conf->proxy_port == proxy_port &&
			(conf->inter_addr == addr || conf->intra_addr == addr)) {
#ifdef CONFIG_FTCS_DEBUG
			BUG_ON(atomic_read(&conf->refcnt) < 1);
#endif
			result = conf;
			atomic_inc(&result->refcnt);
			break;
		}
	}
	read_unlock_bh(&configs_lock);

	return result;
}

/*
  As register_config is always called from the system call interface,
  we use spin_lock_bh here.
*/
int ftcs_register_config(struct ftcs_kconfig *config)
{
	struct ftcs_kconfig *exists;
	int ret = 0;

	atomic_set(&config->refcnt, 1);

	mutex_lock(&configs_mtx);
	write_lock_bh(&configs_lock);

	exists = __find_config(config->inter_addr, config->port);
	if (exists) {
		ret = -EEXIST;
		goto out_unlock;
	}

	list_add(&config->list, &configs);

out_unlock:
	write_unlock_bh(&configs_lock);
	mutex_unlock(&configs_mtx);

	return ret;
}

/*
  conf->proc_file should be removed before calling this function
*/
void ftcs_unregister_config(struct ftcs_kconfig *conf)
{
	WARN_ON(conf->proc_file != NULL);

	mutex_lock(&configs_mtx);
	write_lock_bh(&configs_lock);

	list_del(&conf->list);

	write_unlock_bh(&configs_lock);
	mutex_unlock(&configs_mtx);

	ftcs_release_config(conf);
}

static inline void __release_config(struct ftcs_kconfig *conf)
{
	struct rb_node *node;

	write_lock_bh(&conf->neighbors_lock);

	while ((node = rb_first(&conf->idtree)) != NULL) {
		struct ftcs_kvnode *vn;
		vn = container_of(node, struct ftcs_kvnode, rbnode);
		rb_erase(node, &conf->idtree);
		kmem_cache_free(ftcs_node_cachep, vn);
	}

	while ((node = rb_first(&conf->pnode_tree)) != NULL) {
		struct ftcs_kpnode *pn;
		pn = container_of(node, struct ftcs_kpnode, rbnode);
		rb_erase(node, &conf->pnode_tree);
		kmem_cache_free(ftcs_node_cachep, pn);
	}

	write_unlock_bh(&conf->neighbors_lock);

	dprintf(0,
		"dts: configuration unloaded: proxy:server=%d:%d renotify=%d\n",
		atomic_read(&conf->accepted_inter),
		atomic_read(&conf->accepted_intra),
		atomic_read(&conf->renotify));

	kfree(conf);
}

void ftcs_release_config(struct ftcs_kconfig *c)
{
	if (atomic_dec_and_test(&c->refcnt))
		__release_config(c);
}

static int copy_struct_from_user(void *dest,
				const void __user *user, unsigned int len,
				size_t required_len)
{
	if (len != required_len)
		return -EINVAL;

	if (copy_from_user(dest, user, len))
		return -EFAULT;

	return 0;
}

static int copy_uconf_from_user(const void __user *user, unsigned int len,
				struct ftcs_config *uconf)
{
	return copy_struct_from_user(uconf, user, len, sizeof(*uconf));
}

static int find_netmask(__be32 addr, __be32 *pmask)
{
	struct net_device *dev;
	struct in_device *in_dev;
	int ret;

	dev = ip_dev_find(&init_net, addr);
	if (dev == NULL)
		return -EADDRNOTAVAIL;

	in_dev = in_dev_get(dev);
	if (in_dev == NULL) {
		ret = -ENODEV; /* TODO: is this right? */
		goto out_dev_put;
	}

	ret = -EADDRNOTAVAIL;
	for_ifa(in_dev) {
		dprintf(1, "ifa_address:%pI4 ifa_local:%pI4 ifa_mask:%pI4\n",
			&ifa->ifa_address, &ifa->ifa_local, &ifa->ifa_mask);
		if (ifa->ifa_address == addr) {
			*pmask = ifa->ifa_mask;
			ret = 0;
			break;
		}
	} endfor_ifa(in_dev);

	in_dev_put(in_dev);

out_dev_put:
	dev_put(dev);

	return ret;
}

int ftcs_sys_remove_node(void __user *user, unsigned int len)
{
	struct ftcs_remove_node_param p;
	struct ftcs_config uconf;
	struct ftcs_kconfig *conf;
	int err;

	err = copy_struct_from_user(&p, user, len, sizeof(p));
	if (err) {
		dprintf(1, "err=%d\n", err);
		return err;
	}

	err = copy_uconf_from_user(p.config, sizeof(uconf), &uconf);
	if (err) {
		dprintf(1, "err=%d\n", err);
		return err;
	}

	conf = ftcs_find_get_config(uconf.inter_addr, uconf.port);
	if (conf == NULL) {
		dprintf(1, "err=%d\n", err);
		return -EINVAL;
	}

	ftcs_remove_node(conf, p.addr, p.id);

	ftcs_release_config(conf);

	return 0;
}


int ftcs_sys_add_config(void __user *user, unsigned int len)
{
	struct ftcs_config uconf;
	struct ftcs_kconfig *conf = NULL;
	__be32 intra_mask = 0;
	int err;
	char buff[12+1];

	err = copy_uconf_from_user(user, len, &uconf);
	if (err)
		return err;

	err = find_netmask(uconf.intra_addr, &intra_mask);
	if (err) {
		dprintf(1, "find_netmask failed: %d\n", err);
		return err;
	}

	conf = kzalloc(sizeof(*conf), GFP_KERNEL);
	if (conf == NULL)
		return -ENOMEM;

	conf->inter_addr = uconf.inter_addr;
	conf->intra_addr = uconf.intra_addr;
	conf->intra_mask = intra_mask;
	conf->port = uconf.port;
	conf->proxy_port = uconf.proxy_port;
	rwlock_init(&conf->neighbors_lock);
	conf->idtree = RB_ROOT;
	conf->pnode_tree = RB_ROOT;
	conf->flags = uconf.flags;

	err = ftcs_register_config(conf);
	if (err) {
		kfree(conf);
		return err;
	}

	sprintf(buff, "%08X%04X",
		ntohl(conf->inter_addr), ntohs(conf->port));
	conf->proc_file = proc_create_data(buff, 0444, proc_net_dts_configs,
					&vn_seq_fops, conf);
	if (conf->proc_file == NULL) {
		dprintf(1, "proc_create_data failed\n");
		err = -ENOMEM;
		ftcs_unregister_config(conf);
	}

	dprintf(2, "conf=%p\n", conf);

	return err;
}

void ftcs_flush_configs(void)
{
	struct ftcs_kconfig *conf;

	mutex_lock(&configs_mtx);

	list_for_each_entry(conf, &configs, list) {
		char buff[12+1];

		sprintf(buff, "%08X%04X",
			ntohl(conf->inter_addr), ntohs(conf->port));
		remove_proc_entry(buff, proc_net_dts_configs);
		conf->proc_file = NULL;
	}

	write_lock_bh(&configs_lock);
	while (!list_empty(&configs)) {
		struct ftcs_kconfig *c;
		c = list_entry(configs.next, struct ftcs_kconfig, list);
		list_del_rcu(configs.next);
		dts_clear_endpoints(c);
		ftcs_release_config(c);
	}
	write_unlock_bh(&configs_lock);

	mutex_unlock(&configs_mtx);

	/* TODO: make sure that all entries have been freed */
}

int ftcs_sys_add_vnodes(void __user *user, unsigned int len)
{
	struct ftcs_add_vnodes_param param;
	struct ftcs_config uconf;
	struct ftcs_kconfig *conf;
	struct ftcs_vnode *vnodes;
	size_t vnodes_size;
	int ret = 0;

	if (len != sizeof(param)) {
		dprintf(1, "len:%u different from expected\n", len);
		return -EINVAL;
	}

	if (copy_from_user(&param, user, len))
		return -EFAULT;

	if (param.nr_vnodes <= 0 || param.nr_vnodes > FTCS_ADD_VNODES_MAX) {
		dprintf(1, "param.nr_vnodes == %d\n", param.nr_vnodes);
		return -EINVAL;
	}

	if (copy_from_user(&uconf, param.config, sizeof(uconf)))
		return -EFAULT;

	vnodes_size = sizeof(struct ftcs_vnode) * param.nr_vnodes;
	vnodes = kmalloc(vnodes_size, GFP_KERNEL);
	if (vnodes == NULL)
		return -ENOBUFS;

	if (copy_from_user(vnodes, param.vnodes, vnodes_size)) {
		ret = -EFAULT;
		goto out_kfree;
	}

	conf = ftcs_find_get_config(uconf.inter_addr, uconf.port);
	if (conf == NULL) {
		ret = -EINVAL;
		goto out_kfree;
	}

	ret = ftcs_add_vnodes(conf, vnodes, param.nr_vnodes);

	ftcs_release_config(conf);

out_kfree:
	kfree(vnodes);
	return ret;
}

void ftcs_get_configs_info(struct ftcs_info *info)
{
	struct ftcs_kconfig *conf;
	unsigned int nr = 0, size = 0;

	read_lock_bh(&configs_lock);
	list_for_each_entry(conf, &configs, list) {
		nr++;
		size += sizeof(struct ftcs_config);
	}
	read_unlock_bh(&configs_lock);

	info->nr_configs = nr;
	info->totalsize = size;
}

int ftcs_sys_get_configs(void __user *uconfs, int *len)
{
	int ret = 0;
	struct ftcs_kconfig *conf;
	char __user *up = uconfs;
	unsigned int size = 0;

	mutex_lock(&configs_mtx);
	read_lock_bh(&configs_lock);

	list_for_each_entry(conf, &configs, list) {
		size += sizeof(struct ftcs_config);
	}

	if (size != *len) {
		ret = -EINVAL;
		*len = size;
		goto out_unlock;
	}

	if (size == 0)
		goto out_unlock;

	list_for_each_entry(conf, &configs, list) {
		struct ftcs_config uc;
		uc.inter_addr = conf->inter_addr;
		uc.intra_addr = conf->intra_addr;
		uc.port = conf->port;
		uc.proxy_port = conf->proxy_port;
		uc.flags = conf->flags;

		if (copy_to_user(up, &uc, sizeof(uc))) {
			ret = -EFAULT;
			goto out_unlock;
		}

		up += sizeof(uc);
	}

out_unlock:
	read_unlock_bh(&configs_lock);
	mutex_unlock(&configs_mtx);

	return ret;
}

int ftcs_sys_clear_stats(void __user *user, unsigned int len)
{
	struct ftcs_config uconf;
	struct ftcs_kconfig *conf;
	int err;

	err = copy_uconf_from_user(user, len, &uconf);
	if (err)
		return err;

	conf = ftcs_find_get_config(uconf.inter_addr, uconf.port);
	if (conf == NULL)
		return -EINVAL;

	atomic_set(&conf->accepted_inter, 0);
	atomic_set(&conf->accepted_intra, 0);
#ifdef CONFIG_FTCS_DEBUG
	atomic_set(&conf->renotify, 0);
#endif
	atomic_set(&conf->sspa_synnotify_recv, 0);
	atomic_set(&conf->sspa_ack_snoop, 0);
	atomic_set(&conf->sspa_syn_accept, 0);
	atomic_set(&conf->sspa_firstsyn_lost, 0);

	ftcs_release_config(conf);

	return 0;
}

int ftcs_sys_get_stats(void __user *user, int *len)
{
	struct ftcs_get_stats_param param;
	struct ftcs_kconfig *conf;
	struct ftcs_stats *s;

#ifdef CONFIG_FTCS_PROFILING
	int i;
	for (i = 0; i < DTS_PROF_TOTAL; i++) {
		int nr = atomic_read(&dts_prof_entries[i].nr_samples);
		unsigned long ave = 0;

		if (nr > 0) {
			ave = atomic_read(&dts_prof_entries[i].sum) / nr;
		}

		printk(KERN_INFO "dts: %s %lu %d\n",
			dts_prof_entries[i].name, ave, nr);
	}
#endif

	if (*len != sizeof(param))
		return -EINVAL;

	if (copy_from_user(&param, user, sizeof(param)))
		return -EFAULT;

	conf = ftcs_find_get_config(param.inter_addr, param.port);
	if (conf == NULL)
		return -EINVAL;

	s = &param.stats;
	s->nr_accepted_by_proxy  = atomic_read(&conf->accepted_inter);
	s->nr_accepted_by_server = atomic_read(&conf->accepted_intra);
	s->sspa_synnotify_recv   = atomic_read(&conf->sspa_synnotify_recv);
	s->sspa_ack_snoop        = atomic_read(&conf->sspa_ack_snoop);
	s->sspa_syn_accept       = atomic_read(&conf->sspa_syn_accept);
	s->sspa_firstsyn_lost    = atomic_read(&conf->sspa_firstsyn_lost);

	ftcs_release_config(conf);

	if (copy_to_user(user, &param, sizeof(param)))
		return -EFAULT;

	return 0;
}

static struct ftcs_kvnode *get_vnode_from_idx(struct ftcs_kconfig *conf, int id)
{
	struct rb_node *rbnode;
	int i = 0;

	dprintf(2, "id=%d\n", id);

	rbnode = rb_first(&conf->idtree);
	while (i < id && rbnode != NULL) {
		rbnode = rb_next(rbnode);
		i++;
	}

	if (i == id && rbnode != NULL)
		return container_of(rbnode, struct ftcs_kvnode, rbnode);

	return NULL;
}

static void *vn_seq_start(struct seq_file *seq, loff_t *pos)
{
	struct ftcs_kconfig *conf = seq->private;

	dprintf(2, "conf=%p *pos=%lld\n", conf, *pos);

	read_lock_bh(&conf->neighbors_lock);

	return *pos ? get_vnode_from_idx(conf, *pos - 1) : SEQ_START_TOKEN;
}

static void *vn_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	struct ftcs_kconfig *conf = seq->private;
	struct ftcs_kvnode *vn = NULL;
	struct rb_node *rbnode;

	dprintf(2, "conf=%p *pos=%lld\n", conf, *pos);

	if (v == SEQ_START_TOKEN) {
		vn = get_vnode_from_idx(conf, 0);
		goto out;
	} else if (v == NULL) {
		dprintf(2, "v == NULL\n");
		goto out;
	}

	vn = (struct ftcs_kvnode *) v;

	rbnode = rb_next(&vn->rbnode);

	if (rbnode == NULL)
		vn = NULL;
	else
		vn = container_of(rbnode, struct ftcs_kvnode, rbnode);
out:
	++*pos;

	return vn;
}

static int vn_seq_show(struct seq_file *seq, void *v)
{
	struct ftcs_kvnode *vn = v;
	int i;

	dprintf(2, "\n");

	if (v == SEQ_START_TOKEN) {
		seq_printf(seq, "node    \tid\tchash\n");
		return 0;
	}

	seq_printf(seq, "%08X\t%d\t", ntohl(vn->addr), vn->id);

	for (i = 0; i < FTCS_CHASH_LEN; i++)
		seq_printf(seq, "%02x", vn->hash[i]);
	seq_putc(seq, '\n');

	return 0;
}

static void vn_seq_stop(struct seq_file *seq, void *v)
{
	struct ftcs_kconfig *conf = seq->private;

	dprintf(2, "conf=%p\n", conf);

	read_unlock_bh(&conf->neighbors_lock);
}

static struct seq_operations vn_seq_ops = {
	.start = vn_seq_start,
	.next  = vn_seq_next,
	.show  = vn_seq_show,
	.stop  = vn_seq_stop,
};

static int vn_seq_open(struct inode *inode, struct file *file)
{
	struct seq_file *seq;
	struct ftcs_kconfig *conf;
	int ret;

	conf = (struct ftcs_kconfig *) PDE(inode)->data;

	ret = seq_open(file, &vn_seq_ops);
	if (ret < 0)
		return ret;

	seq = file->private_data;
	seq->private = conf;

	return ret;
}

static struct file_operations vn_seq_fops = {
	.open    = vn_seq_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release,
	.owner   = THIS_MODULE,
};

int ftcs_config_init(void)
{
	int ret;

	proc_net_dts_configs = proc_mkdir("configs", dts_proc_net_dts);
	if (proc_net_dts_configs == NULL) {
		ret = -ENOMEM;
		return ret;
	}

	return 0;
}

void ftcs_config_exit(void)
{
	remove_proc_entry("configs", dts_proc_net_dts);
}
