/*
  dts_chash.c - implements consistent hashing for Layer-4 load balancing
*/

#include <linux/module.h>
#include <linux/inetdevice.h>
#include <net/ip.h>

#include "ftcs.h"
#include "dts_internal.h"

static int force_orphan_query = 0;
module_param(force_orphan_query, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(force_orphan_query, "Force orphan_query");

static
struct ftcs_kvnode *find_vnode(struct ftcs_kconfig *conf,
			u8 hash[FTCS_CHASH_LEN])
{
	struct rb_node *node;

	node = conf->idtree.rb_node;

	while (node) {
		struct ftcs_kvnode *vn
			= container_of(node, struct ftcs_kvnode, rbnode);
		int cmp = ftcs_chash_compare(hash, vn->hash);

		if (cmp > 0) {
			if (node->rb_right == NULL)
				return vn;

			node = node->rb_right;
		}
		else if (cmp < 0) {
			if (node->rb_left == NULL) {
				struct rb_node *prev = rb_prev(node);
				if (prev == NULL) {
					/* node is the left-most node
					   in the tree */
					prev = rb_last(&conf->idtree);
				}

				return container_of(prev,
						struct ftcs_kvnode, rbnode);
			}
			node = node->rb_left;
		}
		else
			return vn;
	}

	return NULL;
}

/*
  rbtree insertion code: almost comes from Documentation/rbtree.txt
*/
static int add_vnode(struct ftcs_kconfig *conf, struct ftcs_kvnode *vn_new)
{
	struct rb_node **new = &(conf->idtree.rb_node), *parent = NULL;
	int ret = 0;

	while (*new) {
		struct ftcs_kvnode *vn;
		int cmp;

		vn = container_of(*new, struct ftcs_kvnode, rbnode);
		cmp = ftcs_chash_compare(vn_new->hash, vn->hash);

		parent = *new;
		if (cmp < 0)
			new = &((*new)->rb_left);
		else if (cmp > 0)
			new = &((*new)->rb_right);
		else {
			ret = -EEXIST;
			goto out;
		}
	}

	rb_link_node(&vn_new->rbnode, parent, new);
	rb_insert_color(&vn_new->rbnode, &conf->idtree);

out:
	return ret;
}

static int add_to_pnode_set(struct ftcs_kconfig *conf, __be32 addr, uint8_t id)
{
	struct rb_node **new = &(conf->pnode_tree.rb_node), *parent = NULL;
	int ret = 0;
	struct ftcs_kpnode *pn;

	while (*new) {
		int cmp;

		pn = container_of(*new, struct ftcs_kpnode, rbnode);
		cmp = (int)(addr - pn->addr);

		parent = *new;
		if (cmp < 0)
			new = &((*new)->rb_left);
		else if (cmp > 0)
			new = &((*new)->rb_right);
		else {
			uint64_t mask = (1ull << id);

			pn->idset |= mask;
			goto out;
		}
	}

	pn = kmem_cache_alloc(ftcs_node_cachep, GFP_ATOMIC);
	if (pn == NULL)
		return -ENOMEM;

	pn->addr = addr;
	pn->idset = (1ull << id);

	if (addr != conf->intra_addr)
		conf->nr_neighbors++;

	rb_link_node(&pn->rbnode, parent, new);
	rb_insert_color(&pn->rbnode, &conf->pnode_tree);

out:
	return ret;
}

static struct ftcs_kpnode *find_kpnode(struct ftcs_kconfig *conf, __be32 addr)
{
	struct rb_node *node;
	struct ftcs_kpnode *pn_found = NULL;

	node = conf->pnode_tree.rb_node;

	while (node) {
		struct ftcs_kpnode *pn
			= container_of(node, struct ftcs_kpnode, rbnode);
		int cmp = (int)(addr - pn->addr);

		if (cmp > 0)
			node = node->rb_right;
		else if (cmp < 0) {
			node = node->rb_left;
		}
		else {
			pn_found = pn;
			break;
		}
	}

	return pn_found;
}

static
void remove_from_pnode_set(struct ftcs_kconfig *conf, __be32 addr, uint8_t id)
{
	struct ftcs_kpnode *pn_found = find_kpnode(conf, addr);

	if (pn_found) {
		uint64_t mask = (1ull << id);
		if (pn_found->idset & mask)
			pn_found->idset &= ~mask;

		if (pn_found->idset == 0) {
			if (addr != conf->intra_addr)
				conf->nr_neighbors--;

			rb_erase(&pn_found->rbnode, &conf->pnode_tree);
			kmem_cache_free(ftcs_node_cachep, pn_found);
		}
	}
}

static void revert_nodes(struct ftcs_kconfig *conf,
			const struct ftcs_vnode *uvns, int nr_uvns)
{
	int i;

	for (i = 0; i < nr_uvns; i++) {
		struct ftcs_kvnode *vn;
		u8 hash[FTCS_CHASH_LEN];
		struct {
			__be32 addr;
			u8 id;
		} __attribute__((packed)) key = { uvns[i].ipaddr,
						  uvns[i].id };

		remove_from_pnode_set(conf, key.addr, key.id);

		/* Special rule: do not add vnode for id == 0 */
		if (uvns[i].id == 0)
			continue;

		ftcs_calc_chash(&key, sizeof(key), hash);

		vn = find_vnode(conf, hash);

		if (vn == NULL || ftcs_chash_compare(vn->hash, hash) != 0)
			continue;

		rb_erase(&vn->rbnode, &conf->idtree);
		kmem_cache_free(ftcs_node_cachep, vn);
	}
}

static void remove_pnode_entries(struct ftcs_kconfig *conf, __be32 addr)
{
	struct rb_node *rbnode;

	rbnode = rb_first(&conf->pnode_tree);
	while (rbnode != NULL) {
		struct ftcs_kpnode *pn = container_of(rbnode,
						struct ftcs_kpnode, rbnode);
		rbnode = rb_next(rbnode);
		if (pn->addr == addr) {
			rb_erase(&pn->rbnode, &conf->pnode_tree);
			kmem_cache_free(ftcs_node_cachep, pn);
		}
	}

	if (addr != conf->intra_addr)
		conf->nr_neighbors--;
}

static void remove_vnode_entries(struct ftcs_kconfig *conf, __be32 addr, int id)
{
	struct rb_node *rbnode;

	rbnode = rb_first(&conf->idtree);
	while (rbnode != NULL) {
		struct ftcs_kvnode *vn = container_of(rbnode,
						struct ftcs_kvnode, rbnode);
		rbnode = rb_next(rbnode);
		if (vn->addr == addr && (id == 0 || id == vn->id)) {
			rb_erase(&vn->rbnode, &conf->idtree);
			kmem_cache_free(ftcs_node_cachep, vn);
		}
	}
}

int ftcs_neighbor(struct ftcs_kconfig *conf, __be32 addr)
{
	struct ftcs_kpnode *pn;

	read_lock_bh(&conf->neighbors_lock);
	pn = find_kpnode(conf, addr);
	read_unlock_bh(&conf->neighbors_lock);

	return pn != NULL;
}

static __be32 responsible_node(struct ftcs_kconfig *conf,
			__be32 raddr, __be16 rport, u8 hash[FTCS_CHASH_LEN])
{
	struct ftcs_kvnode *vn;
	struct {
		__be32 addr;
		__be16 port;
	} __attribute__((packed)) key = { raddr, rport };
	char hashstr[FTCS_CHASH_LEN*2 + 1];
	__be32 ret = FTCS_IPADDR_INVALID;

	ftcs_calc_chash(&key, sizeof(key), hash);
	dprintf(2, "hash=%s\n", ftcs_chash_to_str(hash, hashstr));

	read_lock_bh(&conf->neighbors_lock);
	vn = find_vnode(conf, hash);
	if (vn != NULL)
		ret = vn->addr;
	read_unlock_bh(&conf->neighbors_lock);

	return ret;
}

__be32 ftcs_responsible_node(struct ftcs_kconfig *conf,
			__be32 raddr, __be16 rport)
{
	u8 hash[FTCS_CHASH_LEN];
	return responsible_node(conf, raddr, rport, hash);
}

int ftcs_responsible_for_packet(struct ftcs_kconfig *conf,
				__be32 raddr, __be16 rport)
{
	__be32 resp_node;
	int ret;
	u8 hash[FTCS_CHASH_LEN];
	char hashstr[FTCS_CHASH_LEN*2 + 1];

	resp_node = responsible_node(conf, raddr, rport, hash);

	/* dprintf(1, "%pI4 is responsible for %pI4:%d\n",
	   &resp_node, &raddr, ntohs(rport)); */

	ret = (resp_node == conf->intra_addr);
	if (ret)
		dprintf(2, "responsible: %pI4:%d:%s\n",
			&raddr, ntohs(rport),
			ftcs_chash_to_str(hash, hashstr));

	return ret;
}

__be32 ftcs_prev_responsible_node(struct ftcs_kconfig *conf,
				__be32 raddr, __be16 rport)
{
	struct ftcs_kvnode *vn, *vn_resp;
	struct {
		__be32 addr;
		__be16 port;
	} __attribute__((packed)) key = { raddr, rport };
	u8 hash[FTCS_CHASH_LEN];
	__be32 ret = FTCS_IPADDR_INVALID;
	struct rb_node *prev;

	ftcs_calc_chash(&key, sizeof(key), hash);

	read_lock_bh(&conf->neighbors_lock);
	vn_resp = find_vnode(conf, hash);
	if (vn_resp == NULL)
		goto out_unlock;

	prev = &vn_resp->rbnode;

	do {
		prev = rb_prev(prev);
		if (prev == NULL)
			vn = NULL;
		else
			vn = container_of(prev, struct ftcs_kvnode, rbnode);
	} while (vn && vn->addr == vn_resp->addr);

	if (vn != NULL) {
		ret = vn->addr;
		goto out_unlock;
	}

	prev = rb_last(&conf->idtree);
	while (prev != &vn_resp->rbnode) {
		vn = container_of(prev, struct ftcs_kvnode, rbnode);
		if (vn->addr != vn_resp->addr) {
			ret = vn->addr;
			goto out_unlock;
		}
		prev = rb_prev(prev);
	}

out_unlock:
	read_unlock_bh(&conf->neighbors_lock);

	return ret;
}

__be32 prev_responsible_node(
	struct ftcs_kconfig *conf, const struct ftcs_kvnode *vn_resp)
{
	struct ftcs_kvnode *vn;
	const struct rb_node *prev;
	__be32 ret = FTCS_IPADDR_INVALID;

	prev = &vn_resp->rbnode;

	do {
		prev = rb_prev(prev);
		if (prev == NULL)
			vn = NULL;
		else
			vn = container_of(prev, struct ftcs_kvnode, rbnode);
	} while (vn && vn->addr == vn_resp->addr);

	if (vn != NULL) {
		ret = vn->addr;
		goto out;
	}

	prev = rb_last(&conf->idtree);
	while (prev != &vn_resp->rbnode) {
		vn = container_of(prev, struct ftcs_kvnode, rbnode);
		if (vn->addr != vn_resp->addr) {
			ret = vn->addr;
			goto out;
		}
		prev = rb_prev(prev);
	}

out:
	return ret;
}

__be32 ftcs_next_responsible_node(struct ftcs_kconfig *conf,
				__be32 raddr, __be16 rport)
{
	struct ftcs_kvnode *vn, *vn_resp;
	struct {
		__be32 addr;
		__be16 port;
	} __attribute__((packed)) key = { raddr, rport };
	u8 hash[FTCS_CHASH_LEN];
	__be32 ret = FTCS_IPADDR_INVALID;
	struct rb_node *next;

	ftcs_calc_chash(&key, sizeof(key), hash);

	read_lock_bh(&conf->neighbors_lock);
	vn_resp = find_vnode(conf, hash);
	if (vn_resp == NULL)
		goto out_unlock;

	next = &vn_resp->rbnode;

	do {
		next = rb_next(next);
		if (next == NULL)
			vn = NULL;
		else
			vn = container_of(next, struct ftcs_kvnode, rbnode);
	} while (vn && vn->addr == vn_resp->addr);

	if (vn != NULL) {
		ret = vn->addr;
		goto out_unlock;
	}

	next = rb_first(&conf->idtree);
	while (next != &vn_resp->rbnode) {
		vn = container_of(next, struct ftcs_kvnode, rbnode);
		if (vn->addr != vn_resp->addr) {
			ret = vn->addr;
			goto out_unlock;
		}
		next = rb_next(next);
	}

out_unlock:
	read_unlock_bh(&conf->neighbors_lock);

	return ret;
}

__be32 ftcs_get_responsible_nodes(struct ftcs_kconfig *conf,
				__be32 raddr, __be16 rport, __be32 *pprev_node)
{
	struct ftcs_kvnode *vn;
	struct {
		__be32 addr;
		__be16 port;
	} __attribute__((packed)) key = { raddr, rport };
	u8 hash[FTCS_CHASH_LEN];
	__be32 ret = FTCS_IPADDR_INVALID;
#ifdef CONFIG_FTCS_DEBUG
	char hashstr[FTCS_CHASH_LEN*2 + 1];
#endif

	ftcs_calc_chash(&key, sizeof(key), hash);
	dprintf(2, "hash=%s\n", ftcs_chash_to_str(hash, hashstr));

	read_lock_bh(&conf->neighbors_lock);
	vn = find_vnode(conf, hash);
	if (vn != NULL) {
		ret = vn->addr;
		if (pprev_node)
			*pprev_node = prev_responsible_node(conf, vn);
	}
	read_unlock_bh(&conf->neighbors_lock);

	return ret;
}

void ftcs_remove_node(struct ftcs_kconfig *conf, __be32 addr, int id)
{
	dprintf(1, "removing %pI4,%d\n", &addr, id);

	write_lock_bh(&conf->neighbors_lock);
	remove_vnode_entries(conf, addr, id);
	if (id == 0) {
		remove_pnode_entries(conf, addr);
		getnstimeofday(&conf->ts_last_died);
	}
	write_unlock_bh(&conf->neighbors_lock);
}

int ftcs_add_vnodes(struct ftcs_kconfig *conf,
		struct ftcs_vnode *vnodes, size_t nr_vnodes)
{
	int ret = 0;
	int i;

	write_lock_bh(&conf->neighbors_lock);
	for (i = 0; i < nr_vnodes; i++) {
		struct ftcs_kvnode *vn_new = NULL;
		char hashstr[FTCS_CHASH_LEN*2 + 1];

		if (vnodes[i].id > FTCS_VNODE_ID_MAX) {
			ret = -EINVAL;
			goto revert;
		}

		ret = add_to_pnode_set(conf, vnodes[i].ipaddr, vnodes[i].id);
		if (ret < 0)
			goto revert;

		/* Special rule: do not add vnode for id == 0 */
		if (vnodes[i].id == 0)
			continue;

		vn_new = kmem_cache_alloc(ftcs_node_cachep, GFP_ATOMIC);
		if (vn_new == NULL)
			ret = -ENOMEM;
		else {
			struct {
				__be32 addr;
				u8 id;
			} __attribute__((packed)) key = { vnodes[i].ipaddr,
							  vnodes[i].id };

		        ftcs_calc_chash(&key, sizeof(key), vn_new->hash);
			vn_new->addr = vnodes[i].ipaddr;
			vn_new->id   = vnodes[i].id;
			ret = add_vnode(conf, vn_new);
		}

	revert:
		if (ret < 0) {
			dprintf(1, "error(%d): reverting\n", -ret);
			revert_nodes(conf, vnodes, i);
			if (vn_new) {
				dprintf(1, "hash=%s\n",
					ftcs_chash_to_str(vn_new->hash, hashstr));
				kmem_cache_free(ftcs_node_cachep, vn_new);
			}
			break;
		}

		dprintf(2, "added vnode: %s -> (%pI4:%u)\n",
			ftcs_chash_to_str(vn_new->hash, hashstr),
			&vn_new->addr, vnodes[i].id);
		dprintf(2, "nr_neighbors=%d\n", conf->nr_neighbors);
	}
	write_unlock_bh(&conf->neighbors_lock);

	return ret;
}

int ftcs_mourning(struct ftcs_kconfig *conf)
{
	int ret = 0;
	struct timespec ts_zero = { 0, }, ts_now, ts;

	if (force_orphan_query) {
		dprintf(1, "force_orphan_query\n");
		return 1;
	}

	getnstimeofday(&ts_now);

	write_lock_bh(&conf->neighbors_lock);
	if (timespec_compare(&conf->ts_last_died, &ts_zero) == 0)
		goto out_unlock;

	ts = timespec_sub(ts_now, conf->ts_last_died);
	if (ts.tv_sec >= DTS_MOURNING_TIMEOUT)
		conf->ts_last_died = ts_zero;
	else
		ret = 1;
out_unlock:
	write_unlock_bh(&conf->neighbors_lock);

	dprintf(2, "ret=%d\n", ret);

	return ret;
}
