/*
  Connection tracking for DTS
*/

#include <net/tcp.h>

#include "ftcs.h"
#include "dts_internal.h"

static int timeout_table[5][5] = {
	/* IN \ EG       OP FI FR TW      CL */
	/* OPEN      */ {0, 0, 0, 0,      0},
	/* FIN_SENT1 */ {0, 0, 0, 0,      0},
	/* FIN_SENT2 */ {0, 0, 0, 0,      0},
	/* TIME_WAIT */ {0, 0, 0, 60*HZ, 15*HZ},
	/* CLOSED    */ {0, 0, 0, 60*HZ, 15*HZ},
};

static inline int one_way_closed(enum dts_half_state hst)
{
	if (hst == DTS_HST_TIME_WAIT || hst == DTS_HST_CLOSED)
		return 1;
	return 0;
}

static void egress_setup(struct dts_conntrack *ct,
			const struct iphdr *iph,
			const struct tcphdr *th)
{
	uint32_t len = tcp_datalen(iph, th);
	int extra = (th->syn || th->fin) ? 1 : 0;

	ct->sent_max = ntohl(th->seq) + len + extra;
	ct->rcv_nxt = ntohl(th->ack_seq);
	ct->rcv_nxt_max = ct->rcv_nxt + ntohs(th->window);
	ct->flags |= DTS_CT_EGRESS_SEEN;
}

/*
  TODO: this can be done at connection track initialization,
  as we only care the server side for now.
  And we can reduce the cost of always testing DTS_CT_INGRESS_SEEN bit.
 */
static void ingress_setup(struct dts_conntrack *ct,
			const struct iphdr *iph,
			const struct tcphdr *th)
{
	uint32_t len = tcp_datalen(iph, th);
	int extra = (th->syn || th->fin) ? 1 : 0;

	ct->rcvd_max = ntohl(th->seq) + len + extra;
	ct->flags |= DTS_CT_INGRESS_SEEN;
}

static inline int fin_sent(enum dts_half_state state)
{
	if (state == DTS_HST_FIN_SENT1 || state == DTS_HST_FIN_SENT2)
		return 1;

	return 0;
}

static inline enum dts_half_state
next_state_on_fin(enum dts_half_state opposite)
{
	switch (opposite) {
	case DTS_HST_OPEN:
		return DTS_HST_FIN_SENT1;
	default:
		return DTS_HST_FIN_SENT2;
	}
}

static enum dts_half_state next_state_on_fin_acked(enum dts_half_state state)
{
	switch (state) {
	case DTS_HST_FIN_SENT1:
		return DTS_HST_TIME_WAIT;

	case DTS_HST_FIN_SENT2:
		return DTS_HST_CLOSED;

	default:
		/* BUG?? */
		return state;
	}
}

static inline int calc_next_timeout(const struct dts_conntrack *ct,
				int fin_rto)
{
	int t = ct->max_timeout;
	if (fin_rto > 0)
		return t > 0 ? min(t, fin_rto) : fin_rto;
	else
		return t;

}

int dts_ct_egress(struct dts_endpoint *ep,
		const struct iphdr *iph, const struct tcphdr *th)
{
	struct dts_conntrack *ct = &ep->ct;
	uint32_t seq = ntohl(th->seq);
	uint32_t ack = ntohl(th->ack_seq);
	uint32_t win = ntohs(th->window);
	int timeout = 0;

#ifdef CONFIG_FTCS_PROFILING
	struct timeval tv0, tv1;

	do_gettimeofday(&tv0);
#endif

	/* TODO: window scaling */

	if ((ct->flags & DTS_CT_EGRESS_SEEN) == 0)
		egress_setup(ct, iph, th);
	else if (before(seq, ct->sent_max)) {
		/* This may be a retransmit segment */
		dprintf(1, "retransmit?\n");
		ct->flags |= DTS_CT_EGRESS_RTX_SEEN;
		dts_endpoint_set_timeout(ep, dts_ctl_conn_retrans_timeout);
	}

	if (th->ack) {
		ct->flags |= DTS_CT_EGRESS_ACK_SEEN;
		ct->rcv_nxt = ack;
		ct->rcv_nxt_max = ack+win;

		if (ct->flags & DTS_CT_INGRESS_RTX_SEEN) {
			if (ct->max_timeout > 0)
				dts_endpoint_set_timeout(ep, ct->max_timeout);
			else
				dts_endpoint_clear_timeout(ep);
		}
	}

	if (!before(seq, ct->sent_max)) {
		int len = tcp_datalen(iph, th);
		int extra = (th->syn || th->fin);
		ct->sent_max = seq + len + extra;
	}

	if (th->rst) {
		ct->egress_state = ct->ingress_state = DTS_HST_CLOSED;
		goto out;
	}

	if (fin_sent(ct->ingress_state) && th->ack &&
		!before(ack, ct->rcvd_max)) {
		ct->ingress_state = next_state_on_fin_acked(ct->ingress_state);
	}

	if (th->fin) {
		if (ct->egress_state == DTS_HST_OPEN)
			ct->egress_state = next_state_on_fin(ct->ingress_state);

		if (ct->flags & DTS_CT_EMULATE_TCP) {
			/* Start timer for FIN-retransmit */
			timeout = ct->max_timeout;
		}
	}
out:

	if (one_way_closed(ct->ingress_state)
		&& one_way_closed(ct->egress_state)) {
		ep->ct.max_timeout = timeout_table[ct->ingress_state][ct->egress_state];
	}

	timeout = calc_next_timeout(ct, timeout);

	if (timeout > 0)
		dts_endpoint_set_timeout(ep, timeout);

#ifdef CONFIG_FTCS_PROFILING
	do_gettimeofday(&tv1);
	atomic_add(dts_prof_tv_diff(&tv1, &tv0),
		&dts_prof_entries[DTS_PROF_OUT_CT].sum);
	atomic_inc(&dts_prof_entries[DTS_PROF_OUT_CT].nr_samples);
#endif

	return timeout;
}

int dts_ct_verify_ingress(struct dts_endpoint *ep,
			const struct sk_buff *skb)
{
	struct dts_conntrack *ct = &ep->ct;
	struct iphdr *iph, _iph;
	struct tcphdr *th, _th;
	uint32_t seq;
	uint16_t len;
	int ret = 0;
	uint32_t rcv_min;

	iph = skb_header_pointer(skb, skb_network_offset(skb),
				sizeof(struct iphdr), &_iph);
	if (iph == NULL)
		return -1;

	th = skb_header_pointer(skb, skb_network_offset(skb) + iph->ihl*4,
				sizeof(struct tcphdr), &_th);
	if (th == NULL)
		return -1;

	/* TODO: checksum */
	/* TODO: window scaling */

	seq = ntohl(th->seq);
	len = tcp_datalen(iph, th);

	if ((ct->flags & DTS_CT_EGRESS_ACK_SEEN) == 0) {
		/* Without egress ACK, we cannot determine
		   the acceptable range of the sequence number */
		ret = 0;
		goto out;
	}

	rcv_min = ct->rcv_nxt - (ct->rcv_nxt_max - ct->rcv_nxt);

	if (before(seq, rcv_min) ||
		before(ct->rcv_nxt_max, (uint32_t) (seq+len))) {
		/* Out-of-window */
		dprintf(2, "out of window segment: "
			"seq=%u len=%u rcv_nxt=%u rcv_nxt_max=%u\n",
			seq, len, ct->rcv_nxt, ct->rcv_nxt_max);
		ret = -1;
		goto out;
	}

out:

	return ret;
}

static int reply_to_fin(struct dts_endpoint *ep, const struct tcphdr *th)
{
	int timeout = 0;

	struct dts_conntrack *ct = &ep->ct;
	uint32_t window = ct->rcv_nxt_max - ct->rcv_nxt;
	/* TODO: are rcv_* really initialized at this moment?
	   ANNOT: might be a good example for verification */

	switch (ct->egress_state) {
	case DTS_HST_FIN_SENT1:
	case DTS_HST_FIN_SENT2:
		/* TODO: window scaling */
		dts_enqueue_reply_fin_ack(ep->dcb,
					ep->raddr, ep->rport,
					ep->laddr, ep->lport,
					htonl(ct->sent_max - 1),
					be32_add(th->seq, 1),
					htons(window));
		timeout = ct->max_timeout;
		break;
	default:
		/* TODO: window scaling */
		dts_enqueue_reply_ack(ep->dcb,
				ep->raddr, ep->rport,
				ep->laddr, ep->lport,
				htonl(ct->sent_max),
				be32_add(th->seq, 1),
				htons(window));
		break;
	}

	return timeout;
}

int dts_ct_ingress(struct dts_endpoint *ep,
		const struct iphdr *iph, const struct tcphdr *th)
{
	struct dts_conntrack *ct = &ep->ct;
	int timeout = 0;
	uint32_t seq = ntohl(th->seq);
	uint32_t ack = ntohl(th->ack_seq);

	if ((ct->flags & DTS_CT_INGRESS_SEEN) == 0)
		ingress_setup(ct, iph, th);
	else if (before(seq, ct->rcvd_max)) {
		dprintf(1, "retransmit?\n");
		ct->flags |= DTS_CT_INGRESS_RTX_SEEN;
		dts_endpoint_set_timeout(ep, dts_ctl_conn_retrans_timeout);
	}

	if (th->rst) {
		ct->ingress_state = ct->egress_state = DTS_HST_CLOSED;
		goto out;
	}

	if (th->ack) {
		if (ct->flags & DTS_CT_INGRESS_ACK_SEEN) {
			if (after(ack, ct->rcvd_ack_max))
				ct->rcvd_ack_max = ack;
		} else {
			ct->rcvd_ack_max = ack;
			ct->flags |= DTS_CT_INGRESS_ACK_SEEN;
		}

		if ((ct->flags & DTS_CT_EGRESS_RTX_SEEN)) {
			if (ct->max_timeout > 0)
				dts_endpoint_set_timeout(ep, ct->max_timeout);
			else
				dts_endpoint_clear_timeout(ep);
		}

		/* FIN is acked by the remote peer */
		if (fin_sent(ct->egress_state)) {
			dprintf(2, "%s l=%d r=%d ack=%u sent_max=%u\n",
				dts_state_str(ep->dcb->state),
				ep->lep, ep->rep, ack, ct->sent_max);
			if (!before(ack, ct->sent_max))
				ct->egress_state
					= next_state_on_fin_acked(ct->egress_state);
		}
	}

	if (!before(seq, ct->rcvd_max)) {
		int len = tcp_datalen(iph, th);
		int extra = (th->syn || th->fin);
		ct->rcvd_max = seq + len + extra;
	}

	if (th->fin) {
		if (ct->ingress_state == DTS_HST_OPEN)
			ct->ingress_state = next_state_on_fin(ct->egress_state);

		if (ct->flags & DTS_CT_EMULATE_TCP) {
			/* Reply ACK or FIN+ACK */
			timeout = reply_to_fin(ep, th);
		}
	}

out:
	if (one_way_closed(ct->ingress_state)
		&& one_way_closed(ct->egress_state)) {
		ep->ct.max_timeout = timeout_table[ct->ingress_state][ct->egress_state];
	}

	timeout = calc_next_timeout(ct, timeout);

	if (timeout > 0)
		dts_endpoint_set_timeout(ep, timeout);

	return timeout;
}

int dts_ct_terminated(struct dts_endpoint *ep)
{
	struct dts_conntrack *ct = &ep->ct;
	int ret;

	ret = one_way_closed(ct->ingress_state)
		&& one_way_closed(ct->egress_state);

	return ret;
}

void dts_ct_retrans_fin(struct dts_endpoint *ep)
{
	struct dts_conntrack *ct = &ep->ct;

	switch (ct->egress_state) {
		uint32_t window;
	case DTS_HST_FIN_SENT1:
	case DTS_HST_FIN_SENT2:
		/* FIN has been sent out but not ACKed until now.
		   Retransmit it. */

		window = ct->rcv_nxt_max - ct->rcv_nxt;
		/* TODO: are rcv_* really initialized at this moment?
		   ANNOT: might be a good example for verification */

		/* Extend timeout period */
		ct->max_timeout = min(ct->max_timeout * 2, TCP_RTO_MAX);

		dts_enqueue_reply_fin_ack(ep->dcb,
					ep->raddr, ep->rport,
					ep->laddr, ep->lport,
					htonl(ct->sent_max - 1),
					htonl(ct->rcv_nxt),
					htons(window));
		break;
	default:
		break;
	}
}

void dts_ct_enable_emulation(struct dts_endpoint *ep)
{
	struct dts_conntrack *ct = &ep->ct;

	ct->flags |= DTS_CT_EMULATE_TCP;
	ct->max_timeout = TCP_RTO_MIN;
}

int dts_ct_emulation_enabled(struct dts_endpoint *ep)
{
	struct dts_conntrack *ct = &ep->ct;
	int ret;

	ret = (ct->flags & DTS_CT_EMULATE_TCP);

	return ret;
}
