#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/version.h>

#include "ftcs.h"
#include "dts_internal.h"

static int endpoint_hashbuckets = 8192;
module_param(endpoint_hashbuckets, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(endpoint_hashbuckets, "Hash table bucket size for DTS endpoints");

static void endpoint_timeout(unsigned long data);
static void orphan_timeout(unsigned long data);
static void backup_timeout(unsigned long data);

static struct proc_dir_entry *proc_net_dts_ep;

static inline unsigned int endpoint_hash(__be32 laddr, __be16 lport,
					__be32 raddr, __be16 rport)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
	return inet_ehashfn(&init_net, laddr, lport, raddr, rport)
#else
	return inet_ehashfn(laddr, lport, raddr, rport)
#endif
		% endpoint_hashbuckets;
}

static struct hlist_head *endpoint_lists;
static rwlock_t *endpoint_locks;

struct dts_ep_iterator {
	int bucket;
};

dts_hashtuple_t dts_get_hashtuple(struct dts_endpoint *ep0,
				struct dts_endpoint *ep1)
{
	dts_hashtuple_t t;

	t.keys[0] = endpoint_hash(ep0->laddr, ep0->lport,
				ep0->raddr, ep0->rport);
	t.keys[1] = endpoint_hash(ep1->laddr, ep1->lport,
				ep1->raddr, ep1->rport);

	return t;
}

unsigned int dts_endpoint_hash(const struct dts_endpoint *ep)
{
	return endpoint_hash(ep->laddr, ep->lport, ep->raddr, ep->rport);
}

void dts_endpoints_write_lock(dts_hashtuple_t t)
{
	/* Lock from a bucket which has smaller index
	   to avoid deadlock */
	if (t.keys[0] < t.keys[1]) {
		write_lock_bh(&endpoint_locks[t.keys[0]]);
		write_lock_bh(&endpoint_locks[t.keys[1]]);
	} else if (t.keys[0] > t.keys[1]) {
		write_lock_bh(&endpoint_locks[t.keys[1]]);
		write_lock_bh(&endpoint_locks[t.keys[0]]);
	} else {
		/* We need only one lock here since two hashkeys
		   are identical */
		write_lock_bh(&endpoint_locks[t.keys[0]]);
	}
}

void dts_endpoints_write_unlock(dts_hashtuple_t t)
{
	if (t.keys[0] < t.keys[1]) {
		write_unlock_bh(&endpoint_locks[t.keys[1]]);
		write_unlock_bh(&endpoint_locks[t.keys[0]]);
	} else if (t.keys[0] > t.keys[1]) {
		write_unlock_bh(&endpoint_locks[t.keys[0]]);
		write_unlock_bh(&endpoint_locks[t.keys[1]]);
	} else
		write_unlock_bh(&endpoint_locks[t.keys[0]]);
}

void dts_endpoint_write_lock(unsigned int hash)
{
	write_lock_bh(&endpoint_locks[hash]);
}

void dts_endpoint_write_unlock(unsigned int hash)
{
	write_unlock_bh(&endpoint_locks[hash]);
}

void dts_init_endpoint(struct dts_endpoint *ep, struct dts_cb *dcb,
		enum dts_local_endpoint lep, __be32 laddr, __be16 lport,
		enum dts_remote_endpoint rep, __be32 raddr, __be16 rport)
{
	atomic_set(&ep->refcnt, 1);
	ep->dcb = dcb;
	ep->lep = lep;
	ep->rep = rep;
	ep->laddr = laddr;
	ep->lport = lport;
	ep->raddr = raddr;
	ep->rport = rport;
	ep->timeout_jiffies = 0;
	dts_cb_hold(dcb);

	setup_timer(&ep->timer, endpoint_timeout, (unsigned long) ep);

	dts_ct_init(ep);
}

struct dts_endpoint *dts_new_endpoint(
	struct dts_cb *dcb,
	enum dts_local_endpoint lep, __be32 laddr, __be16 lport,
	enum dts_remote_endpoint rep, __be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep;

	ep = kmem_cache_alloc(dts_endpoint_cachep, GFP_ATOMIC);
	if (ep == NULL)
		return NULL;

	dts_init_endpoint(ep, dcb, lep, laddr, lport, rep, raddr, rport);

	return ep;
}

struct dts_endpoint *dts_new_orphan_endpoint(
	struct dts_cb *dcb,
	__be32 laddr, __be16 lport, __be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep;

	ep = dts_new_endpoint(dcb, DTS_LEP_INTER, laddr, lport,
			DTS_REP_CLIENT, raddr, rport);
	if (ep == NULL)
		return NULL;

	setup_timer(&ep->timer, orphan_timeout, (unsigned long) ep);

	dts_ct_terminate(ep);

	return ep;
}

struct dts_endpoint *dts_new_backup_endpoint(
	struct dts_cb *dcb,
	__be32 laddr, __be16 lport, __be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep;

	ep = dts_new_endpoint(dcb, DTS_LEP_INTER, laddr, lport,
			DTS_REP_CLIENT, raddr, rport);
	if (ep == NULL)
		return NULL;

	setup_timer(&ep->timer, backup_timeout, (unsigned long) ep);

	dts_ct_terminate(ep);

	return ep;
}

static
void __register_endpoint_common(unsigned int hash, struct dts_endpoint *ep)
{
	dts_endpoint_hold(ep);

	hlist_add_head(&ep->node, &endpoint_lists[hash]);

	dprintf(2, "registered ("
		NIPQUAD_FMT ":%d, " NIPQUAD_FMT ":%d)\n",
		NIPQUAD(ep->laddr), ntohs(ep->lport),
		NIPQUAD(ep->raddr), ntohs(ep->rport));
}

void dts_register_endpoint_nolock(struct dts_endpoint *ep)
{
	unsigned int hash;

	hash = endpoint_hash(ep->laddr, ep->lport, ep->raddr, ep->rport);

	__register_endpoint_common(hash, ep);
}

void dts_register_endpoint(struct dts_endpoint *ep)
{
	unsigned int hash;

	hash = endpoint_hash(ep->laddr, ep->lport, ep->raddr, ep->rport);

	write_lock_bh(&endpoint_locks[hash]);
	__register_endpoint_common(hash, ep);
	write_unlock_bh(&endpoint_locks[hash]);
}

static inline struct dts_endpoint *__find_endpoint_any(unsigned int hash,
						__be32 laddr, __be16 lport,
						__be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep;
	struct hlist_head *head;
	struct hlist_node *node;

	head = &endpoint_lists[hash];

	hlist_for_each_entry(ep, node, head, node) {
		if (ep->rport == rport && ep->raddr == raddr &&
			ep->lport == lport && ep->laddr == laddr) {
			prefetchw(ep->dcb);
			return ep;
		}
	}

	return NULL;
}

static struct dts_endpoint *__find_endpoint(unsigned int hash,
					struct dts_cb *dcb,
					__be32 laddr, __be16 lport,
					__be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep;
	struct hlist_head *head;
	struct hlist_node *node;

	head = &endpoint_lists[hash];

	hlist_for_each_entry(ep, node, head, node) {
		if (ep->rport == rport && ep->raddr == raddr &&
			ep->lport == lport && ep->laddr == laddr) {
			if (dcb == NULL || dcb == ep->dcb) {
				prefetchw(ep->dcb);
				return ep;
			}
		}
	}

	return NULL;
}

/*
  Try to register endpoint,
  fails if there is already an endpoint for the same address/port tuples.
*/
int dts_try_register_endpoint(struct dts_endpoint *ep)
{
	unsigned int hash;
	struct dts_endpoint *ep_exists;
	int ret = -EEXIST;

	hash = endpoint_hash(ep->laddr, ep->lport, ep->raddr, ep->rport);

	write_lock_bh(&endpoint_locks[hash]);

	ep_exists = __find_endpoint_any(hash,
					ep->laddr, ep->lport,
					ep->raddr, ep->rport);
	if (ep_exists == NULL) {
		__register_endpoint_common(hash, ep);
		ret = 0;
	}
	write_unlock_bh(&endpoint_locks[hash]);

	return ret;
}

struct dts_endpoint *dts_find_get_endpoint_nolock(__be32 laddr, __be16 lport,
						__be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep = NULL;
	unsigned int hash = endpoint_hash(laddr, lport, raddr, rport);

	ep = __find_endpoint_any(hash, laddr, lport, raddr, rport);

	if (ep) {
#ifdef CONFIG_FTCS_DEBUG
		BUG_ON(atomic_read(&ep->dcb->refcnt) < 1);
#endif
		dts_endpoint_hold(ep);
	}

	return ep;
}

struct dts_endpoint *dts_find_get_endpoint(struct dts_cb *dcb,
					__be32 laddr, __be16 lport,
					__be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep = NULL;
	unsigned int hash = endpoint_hash(laddr, lport, raddr, rport);

	read_lock_bh(&endpoint_locks[hash]);
	ep = __find_endpoint(hash, dcb, laddr, lport, raddr, rport);

	if (ep) {
#ifdef CONFIG_FTCS_DEBUG
		BUG_ON(atomic_read(&ep->dcb->refcnt) < 1);
#endif
		dts_endpoint_hold(ep);
	}
	read_unlock_bh(&endpoint_locks[hash]);

	return ep;
}

struct dts_endpoint *dts_find_get_endpoint_any(__be32 laddr, __be16 lport,
					__be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep = NULL;
	unsigned int hash = endpoint_hash(laddr, lport, raddr, rport);

	read_lock_bh(&endpoint_locks[hash]);
	ep = __find_endpoint_any(hash, laddr, lport, raddr, rport);

	if (ep) {
#ifdef CONFIG_FTCS_DEBUG
		BUG_ON(atomic_read(&ep->dcb->refcnt) < 1);
#endif
		dts_endpoint_hold(ep);
	}
	read_unlock_bh(&endpoint_locks[hash]);

	return ep;
}

static void __unregister_endpoint_common(struct dts_endpoint *ep)
{
	/*
	 * Note: it is safe to call this function multiple times.
	 * (i.e. call of this function is idempotent)
	 */
	if (!hlist_unhashed(&ep->node)) {
		hlist_del_init(&ep->node);
		dts_release_endpoint(ep);
	}

	if (del_timer(&ep->timer)) {
		/* A pending timer was cancelled */

		/* One more release for timer's reference count */
		dts_release_endpoint(ep);
	}
}

void dts_unregister_endpoint(struct dts_endpoint *ep)
{
	unsigned int hash;

	hash = endpoint_hash(ep->laddr, ep->lport, ep->raddr, ep->rport);

	write_lock_bh(&endpoint_locks[hash]);
	__unregister_endpoint_common(ep);
	write_unlock_bh(&endpoint_locks[hash]);
}

void dts_unregister_endpoint_nolock(struct dts_endpoint *ep)
{
	unsigned int hash;

	hash = endpoint_hash(ep->laddr, ep->lport, ep->raddr, ep->rport);

	__unregister_endpoint_common(ep);
}

static void endpoint_timeout(unsigned long data)
{
	struct dts_endpoint *ep = (struct dts_endpoint *) data;

	dts_cb_lock(ep->dcb);
	if (dts_ct_terminated(ep))
		dts_unregister_endpoint(ep);
	else if (dts_ct_emulation_enabled(ep))
		dts_ct_retrans_fin(ep);
	else
		dts_unregister_endpoint(ep);
	dts_cb_unlock(ep->dcb);

	dts_release_endpoint(ep);
}

/*
  Timeout function for orphan endpoints
  The reason for using separate timeout function for them
  is to avoid locking overhead in normal (non-orphan) path
 */
static void orphan_timeout(unsigned long data)
{
	struct dts_endpoint *ep = (struct dts_endpoint *) data;
	struct dts_cb *dcb = ep->dcb;

	dts_cb_lock(dcb);

	switch (dcb->state) {
	case DTS_ORPHAN_QUERYING:
		if (++dcb->state_priv.orphan.nr_query_retry
			> DTS_ORPHAN_QUERY_RETRY_MAX)
			dts_unregister_endpoint(ep);
		else
			dts_endpoint_set_timeout(ep, DTS_ORPHAN_QUERY_TIMEOUT);
		break;

	case DTS_ORPHAN:
		/* Fall through */
	case DTS_IGNORING:
		dts_unregister_endpoint(ep);
		break;

	default:
		if (net_ratelimit())
			printk(KERN_WARNING "dcb->state=%s\n",
				dts_state_str(dcb->state));
		WARN_ON_ONCE(1);
		break;
	}

	dts_cb_unlock(dcb);

	dts_release_endpoint(ep);
}

static void backup_timeout(unsigned long data)
{
	struct dts_endpoint *ep = (struct dts_endpoint *) data;
	struct dts_cb *dcb = ep->dcb;
	struct sk_buff *skb = NULL;

	dts_cb_lock(dcb);

	dprintf(2, "timeout: (L%pI4:%d-R%pI4:%d) %s\n",
		&ep->laddr, ntohs(ep->lport),
		&ep->raddr, ntohs(ep->rport),
		dts_state_str(dcb->state));

	switch (dcb->state) {
	case DTS_SSPA_BLOCKING:
		dcb->state = DTS_SSPA_PASSING;
		skb = dcb->state_priv.sspa.skb;
		dcb->state_priv.sspa.skb = NULL;
		dts_endpoint_set_timeout(ep, dts_ctl_sspa_timeout);
		break;

	case DTS_SSPA_PASSING:
#ifdef CONFIG_FTCS_DEBUG
		if (dts_ctl_debug_level < 2)
			dprintf(0, "nobody cared?: (L%pI4:%d-R%pI4:%d) %s\n",
				&ep->laddr, ntohs(ep->lport),
				&ep->raddr, ntohs(ep->rport),
				dts_state_str(dcb->state));
#endif

		dts_unregister_endpoint(ep);
		break;

	default:
		if (net_ratelimit())
			printk(KERN_WARNING "dcb->state=%s\n",
				dts_state_str(dcb->state));
		WARN_ON_ONCE(1);
		break;
	}

	dts_cb_unlock(dcb);

	dts_release_endpoint(ep);

	if (skb) {
		if (dts_forward_skb(skb))
			kfree_skb(skb);
	}
}

/*
  Clear endpoints related to conf.
  If conf == NULL, all endpoints are cleared.

  The number of cleared endpoints are returned.
 */
static unsigned long clear_endpoints(struct ftcs_kconfig *conf)
{
	int i;
	unsigned long purged = 0;

	for (i = 0; i < endpoint_hashbuckets; i++) {
		struct dts_endpoint *ep;
		struct hlist_head *head;
		struct hlist_node *node, *ntmp;

		write_lock_bh(&endpoint_locks[i]);
		head = &endpoint_lists[i];

		hlist_for_each_entry_safe(ep, node, ntmp, head, node) {
			if (conf == NULL || ep->dcb->conf == conf) {
				__unregister_endpoint_common(ep);
				purged++;
			}
		}
		write_unlock_bh(&endpoint_locks[i]);
	}

	return purged;
}

void dts_clear_endpoints(struct ftcs_kconfig *conf)
{
	clear_endpoints(conf);
}

/*
  Clear all remaining endpoints so that the module can be unloaded
  without leaking memory objects

  WARNING: This function is only for debug purpose.
  Any unexpected consequences might happen.
*/
static void force_clear_endpoints(void)
{
	unsigned long purged = clear_endpoints(NULL);

	if (purged > 0) {
		printk(KERN_WARNING
			"nf_dts: WARNING: %lu endpoints purged forcedly\n",
			purged);
	}
}

/**
   Sequence file implemetation for endpoints

   Most of the idea comes from net/ipv4/tcp_ipv4.c
   See, for example, tcp_seq_start for details.
*/

static struct dts_endpoint *ep_first_on_bucket(struct dts_ep_iterator *it)
{
	for (; it->bucket < endpoint_hashbuckets; it->bucket++) {
		struct dts_endpoint *ep;
		struct hlist_head *head;
		struct hlist_node *node;
		rwlock_t *lock = &endpoint_locks[it->bucket];

		read_lock_bh(lock);

		head = &endpoint_lists[it->bucket];
		hlist_for_each_entry(ep, node, head, node) {
			return ep;
		}

		read_unlock_bh(lock);
	}

	return NULL;
}

static struct dts_endpoint *ep_first(struct dts_ep_iterator *it)
{
	it->bucket = 0;
	return ep_first_on_bucket(it);
}

static struct dts_endpoint *ep_next(struct dts_ep_iterator *it,
				struct dts_endpoint *cur)
{
	struct dts_endpoint *ep;
	struct hlist_node *node = &cur->node;

	hlist_for_each_entry_continue(ep, node, node)
		return ep;

	read_unlock_bh(&endpoint_locks[it->bucket]);

	it->bucket++;

	return ep_first_on_bucket(it);
}

static struct dts_endpoint *get_ep_from_idx(struct dts_ep_iterator *it,
					loff_t pos)
{
	struct dts_endpoint *ep = ep_first(it);

	while (ep != NULL && pos > 0) {
		ep = ep_next(it, ep);
		--pos;
	}

	return ep;
}

static void *ep_seq_start(struct seq_file *seq, loff_t *pos)
{
	struct dts_ep_iterator *it = seq->private;

	it->bucket = 0;

	return *pos ? get_ep_from_idx(it, *pos - 1) : SEQ_START_TOKEN;
}

static void *ep_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	struct dts_ep_iterator *it = seq->private;
	void *ret;

	if (v == SEQ_START_TOKEN) {
		ret = get_ep_from_idx(it, 0);
		goto out;
	}

	ret = ep_next(it, v);
out:
	++*pos;

	return ret;
}

static void ipv4_ntoa(char *buff, __be32 addr)
{
	snprintf(buff, 16, NIPQUAD_FMT, NIPQUAD(addr));
}

static int ep_seq_show(struct seq_file *seq, void *v)
{
	struct dts_endpoint *ep = v;
	char addr_str[16];

	if (v == SEQ_START_TOKEN) {
		seq_printf(seq,
#ifdef CONFIG_FTCS_DEBUG
			"ID   "
#endif
			"%-15s %5s %-15s %5s %10s Ci Ce Timeout\n",
			"Laddress", "Lport", "Raddress", "Rport", "State");
		return 0;
	}

#ifdef CONFIG_FTCS_DEBUG
	seq_printf(seq, "%04x ", ep->dcb->id);
#endif

	ipv4_ntoa(addr_str, ep->laddr);
	seq_printf(seq, "%-15s %5u ", addr_str, ntohs(ep->lport));

	ipv4_ntoa(addr_str, ep->raddr);
	seq_printf(seq, "%-15s %5u ", addr_str, ntohs(ep->rport));

	seq_printf(seq, "%-10s ", dts_state_str(ep->dcb->state));

	seq_printf(seq, "%02x ", ep->ct.ingress_state);
	seq_printf(seq, "%02x ", ep->ct.egress_state);

	if (timer_pending(&ep->timer)) {
		if (ep->timeout_jiffies > jiffies) {
			unsigned long s = (ep->timeout_jiffies - jiffies)/HZ;
			seq_printf(seq, "%lu", s);
		} else
			seq_printf(seq, "0");
	} else
		seq_printf(seq, "-");

	seq_printf(seq, "\n");

	return 0;
}

static void ep_seq_stop(struct seq_file *seq, void *v)
{
	struct dts_ep_iterator *it = seq->private;

	if (v)
		read_unlock_bh(&endpoint_locks[it->bucket]);
}

static struct seq_operations ep_seq_ops = {
	.start = ep_seq_start,
	.next  = ep_seq_next,
	.show  = ep_seq_show,
	.stop  = ep_seq_stop,
};

static int ep_seq_open(struct inode *inode, struct file *file)
{
	int err;

	err = seq_open_private(file, &ep_seq_ops,
			sizeof(struct dts_ep_iterator));
	if (err < 0)
		return err;

	return 0;
}

static struct file_operations ep_seq_fops = {
	.open    = ep_seq_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release_private,
	.owner   = THIS_MODULE,
};

int dts_endpoint_init(void)
{
	int i, ret;

	endpoint_lists = kmalloc(
		sizeof(struct hlist_head) * endpoint_hashbuckets, GFP_KERNEL);
	endpoint_locks = kmalloc(
		sizeof(rwlock_t) * endpoint_hashbuckets, GFP_KERNEL);

	if (endpoint_lists == NULL || endpoint_locks == NULL) {
		ret = -ENOMEM;
		goto out_free_table;
	}

	for (i = 0; i < endpoint_hashbuckets; i++) {
		INIT_HLIST_HEAD(&endpoint_lists[i]);
		rwlock_init(&endpoint_locks[i]);
	}

	proc_net_dts_ep = proc_create_data("ep", 0444, dts_proc_net_dts,
					&ep_seq_fops, NULL);
	if (proc_net_dts_ep == NULL) {
		ret = -ENOMEM;
		goto out_free_table;
	}

	printk(KERN_INFO "ftcs: endpoint hash bucket size: %d\n",
		endpoint_hashbuckets);

	return 0;

out_free_table:
	kfree(endpoint_locks);
	kfree(endpoint_lists);

	return ret;
}

void dts_endpoint_exit(void)
{
	force_clear_endpoints();

	remove_proc_entry("ep", dts_proc_net_dts);

	kfree(endpoint_locks);
	kfree(endpoint_lists);
}
