#include <linux/kernel.h>
#include <linux/version.h>

#include <linux/module.h>
#include <linux/file.h>

#include <linux/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>

#include <net/ip.h>
#include <net/tcp.h>

#include "ftcs.h"
#include "dts_internal.h"

int dts_ctl_handover_timeout = 0; /* default 0 msec */

static int kernel_send(struct socket *sock,
		const void *buff, size_t len, unsigned flags)
{
	struct msghdr msg = {
		.msg_flags = flags,
	};
	struct kvec kvec = {
		.iov_base = (void *) buff,
		.iov_len  = len,
	};

	return kernel_sendmsg(sock, &msg, &kvec, 1, len);
}

static int kernel_send_skb(struct socket *sock,
			const struct sk_buff *skb,
			unsigned int offset, size_t send_len)
{
	struct sk_buff *copied_skb = NULL;
	int err;

	if (skb->len < send_len || skb->len < offset+send_len)
		return -EINVAL;

	if (skb_is_nonlinear(skb)) {
		copied_skb = skb_copy(skb, GFP_KERNEL);
		if (copied_skb == NULL)
			return -ENOMEM;
		skb = copied_skb;
	}

	err = kernel_send(sock, skb->data+offset, send_len, 0);

	if (copied_skb)
		kfree_skb(copied_skb);

	return err;
}

static int kernel_recv(struct socket *sock,
		const void *buff, size_t len, int flags)
{
	struct msghdr msg = {
		.msg_flags = 0,
	};
	struct kvec kvec = {
		.iov_base = (void *) buff,
		.iov_len  = len,
	};

	return kernel_recvmsg(sock, &msg, &kvec, 1, len, flags);
}


static unsigned int tcp_skb_unread_length(const struct tcp_sock *tp,
					const struct sk_buff *skb)
{
	int offset;
	uint32_t seq = TCP_SKB_CB(skb)->seq;

	/* cf. tcp_recvmsg */

	if (before(tp->copied_seq, seq))
		return skb->len;

	offset = tp->copied_seq - seq;
	if (tcp_hdr(skb)->syn)
		offset--;

	return skb->len - offset;
}

static size_t tcp_receive_queue_len(const struct tcp_sock *tp)
{
	struct sock *sk = (struct sock *) tp;
	struct sk_buff *skb;
	size_t len = 0;

	for (skb = sk->sk_receive_queue.next;
	     skb != (struct sk_buff *) &sk->sk_receive_queue;
	     skb = skb->next) {
		len += tcp_skb_unread_length(tp, skb);
	}

	return len;
}

/*
  Almost do the same thing as inet_autobind does
*/
static int autobind(struct sock *sk, __be16 *port)
{
	struct inet_sock *insk;
	int ret = 0;

	lock_sock(sk);
	if (sk->sk_prot->get_port(sk, 0)) {
		ret = -EAGAIN;
		goto out_release;
	}
	insk = inet_sk(sk);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	insk->inet_sport = *port = htons(insk->inet_num);
#else
	insk->sport = *port = htons(insk->num);
#endif

out_release:
	release_sock(sk);

	return ret;
}

static void schedule_remove_endpoint(struct dts_cb *dcb,
			__be32 laddr, __be16 lport,
			__be32 raddr, __be16 rport)
{
	struct dts_endpoint *ep;

	ep = dts_find_get_endpoint(dcb, laddr, lport, raddr, rport);
	if (ep) {
		dts_endpoint_set_timeout(ep, 30*HZ);
		dts_release_endpoint(ep);
	}
}

static int set_handover_timeout(struct socket *sock)
{
	int err;
	struct timeval tv;
	int timeo = dts_ctl_handover_timeout;

	tv.tv_sec = timeo / 1000;
	tv.tv_usec = (timeo % 1000) * 1000;
	err = kernel_setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO,
			(char*)&tv, sizeof(tv));
	if (err) {
		dprintf(0, "setsockopt(SO_SNDTIMEO) failed: %d\n", err);
		return err;
	}
	err = kernel_setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO,
			(char*)&tv, sizeof(tv));
	if (err) {
		dprintf(0, "setsockopt(SO_RCVTIMEO) failed: %d\n", err);
		return err;
	}
	return 0;
}

/*
  dcb: locked
*/
static int do_push_connection(struct dts_endpoint *proxy_ep,
			struct socket *proxy_sock,
			const struct sockaddr_in *dest_sin,
			const void __user *req_buf, size_t req_len,
			size_t next_req_offset)
{
	struct dts_cb *dcb = proxy_ep->dcb;
	struct dts_endpoint *push_ep;
	struct socket *push_sock;
	struct sock *sk;
	struct tcp_sock *tp;
	uint32_t seq_initial;
	uint32_t rqlen, sent_skblen = 0;
	__be16 lport;
	int err;
	int rewind_flag = 0;
	char c;

	/*
	  TODO:
	  check dcb state after each re-locking as the state might be modified
	  by other context during we are releasing the lock
	*/

	dts_cb_unlock(dcb);

	err = sock_create_kern(AF_INET, SOCK_STREAM, IPPROTO_TCP,
			&push_sock);
	if (err) {
		dprintk(KERN_INFO "%s: sock_create_kern failed: %d\n",
			__func__, err);
		goto out_relock;
	}

	err = set_handover_timeout(push_sock);
	if (err)
		goto out_release;

	err = autobind(push_sock->sk, &lport);
	if (err) {
		dprintk(KERN_INFO "%s: autobind() failed: %d\n", __func__, err);
		goto out_release;
	}

	sk = proxy_sock->sk;

	lock_sock(sk);
	tp = tcp_sk(sk);

	rqlen = tcp_receive_queue_len(tp);
	seq_initial = tp->rcv_nxt - rqlen - (uint32_t)req_len - 1;

	dts_cb_lock(dcb);
	dcb->state_priv.push.local_syn_seq  = htonl(seq_initial);
	dcb->state_priv.push.remote_syn_seq
		= htonl(tp->snd_nxt - 1 - dcb->inspect_seq_offset);
	dcb->state_priv.push.remote_window  = tp->snd_wnd;
	dcb->state_priv.push.handover_len   = req_len + rqlen;
	dts_cb_unlock(dcb);

	push_ep = dts_new_endpoint(dcb,
				DTS_LEP_INTRA,
				dcb->conf->intra_addr, lport,
				DTS_REP_PULL_NEIGHBOR,
				dest_sin->sin_addr.s_addr,
				dest_sin->sin_port);
	if (push_ep == NULL) {
		dprintk(KERN_INFO "%s: dts_new_endpoint failed\n", __func__);
		goto out_release_sock;
	}

	dts_register_endpoint(push_ep);

	dprintf(2, "connecting to " NIPQUAD_FMT ":%d\n",
		NIPQUAD(dest_sin->sin_addr.s_addr),
		ntohs(dest_sin->sin_port));

	err = kernel_connect(push_sock, (struct sockaddr *) dest_sin,
			sizeof(struct sockaddr_in), 0);
	if (err) {
		dprintk(KERN_INFO "%s: kernel_connect failed: %d\n",
			__func__, err);
		/* printk(KERN_INFO
			"dts: kernel_connect(lport:%d-%pI4:%d) failed: %d\n",
			lport, &dest_sin->sin_addr.s_addr,
			ntohs(dest_sin->sin_port), err); */
		rewind_flag = 1;
		goto out_unreg_release;
	}

	/* Hereafter we cannot abort handover safely */

	if (dcb->conf->flags & FTCS_FLAG_DSTSW)
		ftcs_notify_dest_switched(dcb, dest_sin->sin_addr.s_addr);

	schedule_remove_endpoint(dcb, dcb->conf->inter_addr, dcb->conf->port,
			proxy_ep->raddr, proxy_ep->rport);

	dts_cb_lock(dcb);
	dcb->state = DTS_PUSHING;
	dcb->state_priv.push.dest_addr = dest_sin->sin_addr.s_addr;
	dts_cb_unlock(dcb);

	if (req_buf != NULL) {
		/* TODO: XXX! DO NOT pass the user area buffer to
		   kernel_send */
		unsigned flags = skb_queue_empty(&sk->sk_receive_queue)
			? 0 : MSG_MORE;

		err = kernel_send(push_sock, req_buf, req_len, flags);
		if (err < 0) {
			/* TODO: what the new state should be? */
			dprintf(2, "kernel_send failed: %d\n", err);
			goto out_unreg_release;
		}

		if (err != req_len) {
			printk(KERN_ERR "do_push_connection: "
				"err(%d) != req_len(%zu)\n",
				err, req_len);
		}
	}

	while (!skb_queue_empty(&sk->sk_receive_queue)) {
		struct sk_buff *skb = __skb_dequeue(&sk->sk_receive_queue);
		size_t send_len;

		send_len = tcp_skb_unread_length(tp, skb);
		if (send_len == 0) {
			kfree_skb(skb);
			continue;
		}

		err = kernel_send_skb(push_sock, skb,
				skb->len - send_len, send_len);
		kfree_skb(skb);
		if (err < 0) {
			dprintk(KERN_ERR "%s: kernel_send_skb failed: %d\n",
				__func__, err);
			goto out_unreg_release;
		}
		sent_skblen += err;
	}

	dprintf(2, "rqlen=%u sent_skblen=%u\n", rqlen, sent_skblen);

	if (rqlen != sent_skblen) {
		printk(KERN_ERR "do_push_connection: "
			"rqlen(%u) != sent_skblen(%u)\n",
			rqlen, sent_skblen);
	}

	/* Wait for socket to be closed */
	do {
		err = kernel_recv(push_sock, &c, 1, 0);
		if (err < 0) {
			dprintf(1, "kernel_recv() failed: %d\n", err);
			goto out_unreg_release;
		} else if (err > 0) {
			dprintf(1, "Unexpected data received: 0x%02x\n", c);
		}
	} while (err != 0);

	/* On success, keep endpoint */

out_release_ep:
	/* Drop a reference count for this function */
	dts_release_endpoint(push_ep);
out_release_sock:
	release_sock(sk);

out_release:
	sock_release(push_sock);

out_relock:
	dts_cb_lock(dcb);
	if (rewind_flag)
		dcb->state = DTS_INSPECTING;

	return err;

out_unreg_release:
	dts_unregister_endpoint(push_ep);
	goto out_release_ep;
}

static struct dts_endpoint *find_get_endpoint_from_sock(struct socket *sock)
{
	struct sockaddr_in peer_sin, local_sin;
	int sin_len = sizeof(struct sockaddr_in);
	int err;

	err = kernel_getpeername(sock, (struct sockaddr *)&peer_sin, &sin_len);
	if (err) {
		dprintk(KERN_INFO "%s: kernel_getpeername failed: %d\n",
			__func__, err);
		return NULL;
	}

	err = kernel_getsockname(sock, (struct sockaddr *)&local_sin, &sin_len);
	if (err) {
		dprintk(KERN_INFO "%s: kernel_getsockname failed: %d\n",
			__func__, err);
		return NULL;
	}

	return dts_find_get_endpoint(
		NULL,
		local_sin.sin_addr.s_addr, local_sin.sin_port,
		peer_sin.sin_addr.s_addr, peer_sin.sin_port);
}

int dts_request_handover(struct socket *sock,
			const struct sockaddr_in *dest_sin,
			const void __user *req_buf, size_t req_len,
			size_t next_req_offset)
{
	int ret;
	struct dts_endpoint *ep;

	ep = find_get_endpoint_from_sock(sock);
	if (ep == NULL) {
		dprintk(KERN_INFO "%s: find_get_endpoint_from_sock failed\n",
			__func__);
		ret = -EINVAL;
		goto out;
	}

	/* TODO: ep->lep == PROXY ? */

	dts_cb_lock(ep->dcb);

	if (dts_nohandover_mode(ep->dcb->conf)) {
		ret = -EOPNOTSUPP;
		goto out_unlock;
	}

	switch (ep->dcb->state) {
		union dts_state_priv priv;
	case DTS_INSPECTING:
		priv = ep->dcb->state_priv;
		ep->dcb->state = DTS_PUSH_SYN_WAIT;

		ret = do_push_connection(ep, sock, dest_sin,
					req_buf, req_len, next_req_offset);
		if (ep->dcb->state == DTS_PUSH_SYN_WAIT) {
			/* Handover was aborted in early stage
			   - try to resume */
			ep->dcb->state = DTS_INSPECTING;
			ep->dcb->state_priv = priv;
		}

		break;
		/* TODO: IFP */
	default:
		dprintk(KERN_INFO "%s: not in the INSPECTING state\n",
			__func__);
		ret = -EINVAL;
		break;
	}

out_unlock:
	dts_cb_unlock(ep->dcb);

	dts_release_endpoint(ep);

out:
	if (ret == 0)
		sock->ops->shutdown(sock, SHUT_RDWR);

	return ret;
}

/*
  The body part of connection handover request system call
*/
int dts_sys_request_handover(void __user *uptr, socklen_t len)
{
	struct sockaddr_in sin;
	struct socket *sock;
	int err;
	struct ftcs_handover_param param;

	/* TODO: capability check */

	if (len != sizeof(param)) {
		dprintk(KERN_INFO "%s: len=%d sizeof(param)=%zu\n",
			__func__, len, sizeof(param));
		return -EINVAL;
	}

	if (copy_from_user(&param, uptr, len))
		return -EFAULT;

	if (param.addrlen != sizeof(sin)) {
		dprintk(KERN_INFO "%s: param.addrlen=%d sizeof(sin)=%zu\n",
			__func__, param.addrlen, sizeof(sin));
		return -EINVAL; /* TODO: is this code right? */
	}

	if (copy_from_user(&sin, param.dest_addr, sizeof(sin)))
		return -EFAULT;

	if (sin.sin_family != AF_INET)
		return -EOPNOTSUPP;

	if (dts_l4_only_mode)
		return -EOPNOTSUPP;

	sock = sockfd_lookup(param.socket, &err);
	if (sock == NULL) {
		dprintk(KERN_INFO "%s: sockfd_lookup() failed: %d\n",
			__func__, err);
		return err;
	}

	/* TODO: check socket state (eg. is it in the ESTABLISHED state?) */

	err = dts_request_handover(sock,
				&sin,
				param.req_buff,
				param.req_len,
				param.next_req_offset);

	sockfd_put(sock);

	return err;
}
