#include <linux/module.h>
#include <linux/types.h>

#include <linux/skbuff.h>

#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/inetdevice.h>

#include <net/ip.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/protocol.h>
#include <net/sock.h>

#include <linux/netfilter.h>

#include <asm/msr.h>
#include <asm/unaligned.h>

#include "ftcs.h"
#include "dts_internal.h"

extern int tcp_modify_timestamp(struct sk_buff *skb, s64 tsval_offset, s64 tsecr_offset);

int tcp_extract_timestamp(const struct sk_buff *skb,
						  uint32_t *tsval,
						  uint32_t *tsecr)
{
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *th = (void *) iph + iph->ihl * 4;
	int length = th->doff*4 - sizeof (struct tcphdr);
	unsigned char *ptr = (unsigned char *) th + sizeof (struct tcphdr);

	if (unlikely(tsval == NULL || tsecr == NULL))
		return EINVAL;

	while (length > 0) {
		int opcode = *ptr++;
		int opsize;

		switch (opcode) {
		case TCPOPT_NOP:
			length--;
			continue;

		case TCPOPT_EOL:
			return EINVAL;

		default:
			opsize = *ptr++;
			if (opsize < 2) /* "silly options" */
				return EINVAL;
			if (opsize > length)
				return EINVAL;

			switch (opcode) {
			case TCPOPT_TIMESTAMP:
				if (opsize == TCPOLEN_TIMESTAMP) {
					if (tsval != NULL)
						*tsval = get_unaligned_be32(ptr);
					if (tsecr != NULL)
						*tsecr = get_unaligned_be32(ptr + 4);
					return 0;
				}
				return EINVAL;

			default:
				break;
			}

			ptr += opsize - 2;
			length -= opsize;
		}
	}

	return EINVAL;
}

static void parse_inter_syn_opt(struct sk_buff *skb, const struct tcphdr *th,
			struct tcp_options_received *opt_rx)
{
	sk_buff_data_t orig_theader = skb->transport_header;
	u8 *hvp;
#ifdef NET_SKBUFF_DATA_USES_OFFSET
	skb->transport_header = (sk_buff_data_t)(((unsigned char *) th) - skb->head);
#else /* NET_SKBUFF_DATA_USES_OFFSET */
	skb->transport_header = (sk_buff_data_t)th;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
	tcp_parse_options(skb, opt_rx, &hvp, 0);
#else
	tcp_parse_options(skb, opt_rx, 0);
#endif
	skb->transport_header = orig_theader;
}

/* remove ALL data from TCP payload */
static int remove_conn_opt(struct sk_buff *skb)
{
	unsigned char *to, *from;
	struct iphdr *iph;
	struct tcphdr *th;
	size_t ipdatalen, tot_len, len;
	__be16 orig_tot_len;

	len = sizeof(struct dts_conn_opt);

	iph = ip_hdr(skb);
	orig_tot_len = iph->tot_len;
	th = (void *) iph + ip_hdrlen(skb);

	from = skb_network_header(skb);

	to = from  + len;
	tot_len = ntohs(orig_tot_len) - len;

	skb_trim(skb, tot_len);

	/* Adjust length */
	iph->tot_len = htons(tot_len);
	ipdatalen = ntohs(iph->tot_len) - ip_hdrlen(skb);

	/* Adjust checksum */
	csum_replace2(&iph->check, orig_tot_len, iph->tot_len);

	/* yes, you need refactoring these code */
	if (skb->ip_summed == CHECKSUM_PARTIAL) {
		__be16 oldlen, newlen;

		oldlen = htons(ipdatalen + len);
		newlen = htons(ipdatalen);

		skb->csum_start = (unsigned char *)th - skb->head;
		/* Initial checksum value for hardware checksumming
		   should be negative value */
		csum_replace2(&th->check, ~oldlen, ~newlen);
	} else {
		th->check = 0;
		th->check = csum_fold(
			csum_tcpudp_nofold(iph->saddr, iph->daddr,
					ipdatalen, IPPROTO_TCP,
					skb_checksum(skb,
						ip_hdrlen(skb),
						ipdatalen,
						0)));
	}

	return 0;
}

static int parse_conn_opt(struct sk_buff *skb,
			struct dts_conn_opt *opt)
{
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *th = (void *) iph + ip_hdrlen(skb);
	size_t thlen = th->doff*4;
	size_t optlen = thlen - sizeof(struct tcphdr);
	int i = 0;
	unsigned char *ptr = (unsigned char *)th + sizeof(struct tcphdr);

	while (i + DTS_TCPOLEN_CONN_PUSH < optlen) {
		switch (ptr[i]) {
		case TCPOPT_NOP:
			dprintf(2, "i=%d TCPOPT_NOP\n", i);
			i++;
			break;

		case TCPOPT_EOL:
			dprintf(2, "i=%d TCPOPT_EOL\n", i);
			i++;
			goto exit_loop;

		case DTS_TCPOPT_CONN_PUSH:
			dprintf(2, "i=%d TCPOPT_CONN_PUSH\n", i);
			if (ptr[i+2] == sizeof(*opt)) {
				memcpy(opt, (unsigned char *)th + thlen, sizeof(*opt));
				return remove_conn_opt(skb);
			}
			/* fall through */
		default:
			dprintf(2, "i=%d TCPOPT(%u) len=%u\n",
				i, ptr[i], ptr[i+1]);
			i += ptr[i+1];
			break;
		}
	}
exit_loop:

	return -EINVAL;
}

static int is_duplicate_syn(__be32 laddr, __be16 lport,
			__be32 raddr, __be16 rport,
			__be32 seq)
{
	int ret = 0;
	struct dts_endpoint *ep;
	ep = dts_find_get_endpoint_nolock(laddr, lport, raddr, rport);

	if (ep == NULL)
		return 0;

	if (dts_ct_terminated(ep))
		goto out_release;

	dts_cb_lock(ep->dcb);
	switch (ep->dcb->state) {
	case DTS_FIRST_SYN_RECV1:
	case DTS_FIRST_SYN_RECV2:
		ret = ep->dcb->state_priv.first.first_syn_seq == seq;
		break;

	case DTS_PULL_SYN_RECV:
		ret = ep->dcb->state_priv.pull.first_syn_seq == seq;
		break;

	default:
		ret = 1;
		break;
	}
	dts_cb_unlock(ep->dcb);

out_release:
	dts_release_endpoint(ep);

	return ret;
}

struct dts_cb *dts_cb_new(struct ftcs_kconfig *conf,
			enum dts_state state,
			__be32 raddr, __be16 rport)
{
	struct dts_cb *dcb;

	dcb = kmem_cache_alloc(dts_cb_cachep, GFP_ATOMIC);
	if (dcb == NULL)
		return NULL;

	BUG_ON(atomic_read(&dcb->refcnt) != 0);

	atomic_inc(&conf->refcnt);
	dcb->conf = conf;

	dcb->inspect_seq_offset = 0;
	dcb->state = state;
	dcb->raddr = raddr;
	dcb->rport = rport;

	dts_cb_hold(dcb);

	return dcb;
}

static int
handle_intra_syn_input(struct ftcs_kconfig *conf, struct sk_buff *skb,
		const struct iphdr *iph, const struct tcphdr *th,
		struct dts_endpoint **pep)
{
	struct dts_cb *dcb = NULL;
	struct dts_endpoint *ep_intra = NULL, *ep_inter = NULL;
	dts_hashtuple_t hash;
	struct dts_conn_opt opt;
	int ret = 0;

	ret = parse_conn_opt(skb, &opt);
	if (ret) {
		dprintk(KERN_INFO "%s: parse_conn_opt failed: %d\n",
			__func__, ret);
		return ret;
	}

	dcb = dts_cb_new(conf, DTS_PULL_SYN_RECV,
			opt.real_raddr, opt.real_rport);
	if (dcb == NULL) {
		dprintk(KERN_INFO "%s: alloc dcb failed\n", __func__);
		ret = -ENOMEM;
		goto out_free;
	}

	dcb->state_priv.pull.peer_addr = iph->saddr;
	dcb->state_priv.pull.peer_port = th->source;
	dcb->state_priv.pull.egress_syn_seq = opt.initial_egress_seq;
	dcb->state_priv.pull.remote_window  = ntohl(opt.remote_window);
	dcb->state_priv.pull.handover_len   = ntohl(opt.handover_len);
	dcb->state_priv.pull.first_syn_seq = th->seq;
	dcb->remote_last_timestamp = opt.remote_last_timestamp;

	ep_intra = dts_new_endpoint(
		dcb,
		DTS_LEP_INTRA, iph->daddr, th->dest,
		DTS_REP_PUSH_NEIGHBOR, iph->saddr, th->source);

	ep_inter = dts_new_endpoint(
		dcb,
		DTS_LEP_INTER, conf->inter_addr, th->dest,
		DTS_REP_CLIENT, opt.real_raddr, opt.real_rport);

	if (ep_intra == NULL || ep_inter == NULL) {
		ret = -ENOMEM;
		goto out_free;
	}

	hash = dts_get_hashtuple(ep_intra, ep_inter);

	dts_endpoints_write_lock(hash);

	/* Duplicate SYN check */
	if (is_duplicate_syn(iph->daddr, th->dest, iph->saddr, th->source,
				th->seq)) {
		dprintk(KERN_INFO
			"%s: register_endpoint_common(intra): "
			"already exists\n",
			__func__);
		ret = 0;
		goto out_unlock_free;
	}

	dts_ct_set_max_timeout(ep_inter, dts_ctl_initial_conn_timeout);
	dts_ct_set_max_timeout(ep_intra, dts_ctl_initial_conn_timeout);

	/* TODO: recalc hash key is little bit clumsy */
	dts_register_endpoint_nolock(ep_intra);
	dts_register_endpoint_nolock(ep_inter);

	atomic_inc(&conf->accepted_intra);

out_unlock_free:
	dts_endpoints_write_unlock(hash);

out_free:
	/* Even in the normal path we release each pointer
	   because we passed the ownership of them to other list */
	if (ep_inter)
		dts_release_endpoint(ep_inter);
	*pep = ep_intra;
	if (dcb)
		dts_cb_release(dcb);

	return ret;
}

/*
  This function can be called from
  both NF_INET_PREROUTING and NF_INET_LOCAL_IN

  context: softirq
*/
static int
handle_inter_syn_input(struct ftcs_kconfig *conf, struct sk_buff *skb,
		const struct iphdr *iph, const struct tcphdr *th,
		int secondary,
		struct dts_endpoint **pep)
{
	struct dts_cb *dcb = NULL;
	struct dts_endpoint *ep_inter = NULL, *ep_proxy = NULL;
	dts_hashtuple_t hash;
	int ret = 0;

	dcb = dts_cb_new(conf, DTS_FIRST_SYN_RECV1, iph->saddr, th->source);
	if (dcb == NULL) {
		ret = -ENOMEM;
		goto out_free;
	}

	tcp_clear_options(&dcb->tcp_options);
	/* Get tcp options of SYN packet from client */
	parse_inter_syn_opt(skb, th, &dcb->tcp_options);

	dcb->raddr = iph->saddr;
	dcb->rport = th->source;
	dcb->pull_seq_offset = 0;
	dcb->state_priv.first.first_syn_seq = th->seq;
	dcb->state_priv.first.secondary     = secondary;

	ep_inter = dts_new_endpoint(dcb,
				DTS_LEP_INTER, iph->daddr, th->dest,
				DTS_REP_CLIENT, iph->saddr, th->source);
	ep_proxy = dts_new_endpoint(dcb,
				DTS_LEP_PROXY, iph->daddr, conf->proxy_port,
				DTS_REP_CLIENT, iph->saddr, th->source);
	if (ep_inter == NULL || ep_proxy == NULL) {
		ret = -ENOMEM;
		goto out_free;
	}

	hash = dts_get_hashtuple(ep_proxy, ep_inter);

	dts_endpoints_write_lock(hash);

	/* Duplicate SYN check */
	if (is_duplicate_syn(iph->daddr, conf->proxy_port,
				iph->saddr, th->source, th->seq)) {
		dprintk(KERN_INFO
			"%s: register_endpoint_common(intra): "
			"already exists\n",
			__func__);
		ret = 0;
		dts_release_endpoint(ep_inter);
		ep_inter = NULL;
		goto out_unlock_free;
	}

	dts_ct_set_max_timeout(ep_proxy, dts_ctl_initial_conn_timeout);
	dts_ct_set_max_timeout(ep_inter, dts_ctl_initial_conn_timeout);

	dts_register_endpoint_nolock(ep_proxy);
	dts_register_endpoint_nolock(ep_inter);

	atomic_inc(&conf->accepted_inter);

out_unlock_free:
	dts_endpoints_write_unlock(hash);

out_free:
	/* Even in the normal path we release each pointer
	   because we passed the ownership of them to other list */
	if (ep_proxy)
		dts_release_endpoint(ep_proxy);
	*pep = ep_inter;
	if (dcb)
		dts_cb_release(dcb);

	return ret;
}

static int
handle_inter_syn_nohandover(struct ftcs_kconfig *conf, struct sk_buff *skb,
			const struct iphdr *iph, const struct tcphdr *th,
			int secondary,
			struct dts_endpoint **pep)
{
	struct dts_cb *dcb = NULL;
	struct dts_endpoint *ep_inter = NULL;
	int ret = 0;
	unsigned int hash;

	dcb = dts_cb_new(conf, DTS_FIRST_SYN_RECV1, iph->saddr, th->source);
	if (dcb == NULL) {
		ret = -ENOMEM;
		goto out_free;
	}

	dcb->raddr = iph->saddr;
	dcb->rport = th->source;
	dcb->pull_seq_offset = 0;
	dcb->state_priv.first.first_syn_seq = th->seq;
	dcb->state_priv.first.secondary     = secondary;

	ep_inter = &dcb->ep_storage;
	dts_init_endpoint(ep_inter, dcb,
			DTS_LEP_INTER, iph->daddr, th->dest,
			DTS_REP_CLIENT, iph->saddr, th->source);

	hash = dts_endpoint_hash(ep_inter);

	dts_endpoint_write_lock(hash);

	/* Duplicate SYN check */
	if (is_duplicate_syn(iph->daddr, conf->proxy_port,
				iph->saddr, th->source, th->seq)) {
		ret = 0;
		dts_release_endpoint(ep_inter);
		ep_inter = NULL;
		goto out_unlock_free;
	}

	dts_ct_set_max_timeout(ep_inter, dts_ctl_initial_conn_timeout);

	dts_register_endpoint_nolock(ep_inter);

	atomic_inc(&conf->accepted_inter);

out_unlock_free:
	dts_endpoint_write_unlock(hash);

out_free:
	/* Even in the normal path we release each pointer
	   because we passed the ownership of them to other list */
	*pep = ep_inter;
	if (dcb)
		dts_cb_release(dcb);

	return ret;
}

static void terminate_proxy_ep(struct dts_cb *dcb,
			const struct iphdr *iph, const struct tcphdr *th)
{
	struct dts_endpoint *ep_proxy;

	ep_proxy = dts_find_get_endpoint(
		dcb, iph->daddr, dcb->conf->proxy_port,
		iph->saddr, th->source);
	if (ep_proxy == NULL) {
		dprintf(0, "dts_find_get_endpoint failed!\n");
		return;
	}
	dts_ct_terminate(ep_proxy);
	dts_endpoint_set_timeout(ep_proxy, 15*HZ); /* TODO: value */
	dts_release_endpoint(ep_proxy);
}

static enum dts_filter_verdict
inter_inspecting(struct dts_endpoint *ep, struct sk_buff *skb,
		struct iphdr *iph, struct tcphdr *th, int need_to_adjust_ack)
{
	struct dts_cb *dcb = ep->dcb;
	struct dts_endpoint *ep_proxy;

	/* Redirect skb to the proxy port */
	modify_destination(skb, iph, th,
			iph->daddr, dcb->conf->proxy_port);

	if (need_to_adjust_ack) {
		__be32 bef, aft;
		bef = th->ack_seq;
		adjust_ingress_ack(th, dcb->inspect_seq_offset);
		aft = th->ack_seq;
	}

	ep_proxy = dts_find_get_endpoint(dcb, iph->daddr, th->dest,
						iph->saddr, th->source);
	if (ep_proxy == NULL) {
		dprintf(0, "dts_find_get_endpoint failed!\n");
		return DTS_DROP_NOCT;
	}

	dts_ct_ingress(ep, iph, th);
	ep_proxy->ct = ep->ct;

	dts_release_endpoint(ep_proxy);

	return DTS_ACCEPT_NOCT;
}

static inline enum dts_filter_verdict
inter_inspecting_nohandover(struct dts_endpoint *ep, struct sk_buff *skb,
			struct iphdr *iph, struct tcphdr *th,
			int need_to_adjust_ack)
{
	struct dts_cb *dcb = ep->dcb;

	if (need_to_adjust_ack)
		adjust_ingress_ack(th, dcb->inspect_seq_offset);

	dts_ct_ingress(ep, iph, th);

	return DTS_ACCEPT_NOCT;
}

static void do_sspa_lost(struct dts_endpoint *ep,
			const struct iphdr *iph, const struct tcphdr *th)
{
	struct dts_cb *dcb = ep->dcb;

	dcb->state = DTS_FIRST_SYN_LOST;
	atomic_inc(&dcb->conf->sspa_firstsyn_lost);
	dts_ct_terminate(ep);
	dts_endpoint_set_timeout(ep, 15*HZ); /* TODO: value */

	if (!dts_nohandover_mode(dcb->conf))
		terminate_proxy_ep(dcb, iph, th);
}

/*
  dcb: locked
  context: softirq
*/
static enum dts_filter_verdict
process_inter_input(struct dts_endpoint *ep, struct sk_buff *skb)
{
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;
	struct dts_cb *dcb = ep->dcb;
	struct iphdr *iph;
	struct tcphdr *th;
	int need_to_adjust_ack = 1;
	uint32_t tsval, tsecr;

	if (prepare_mangle_tcphdr(skb, &iph, &th))
		goto out;

	switch (dcb->state) {
		int err;
		__be32 expected;

	case DTS_FORWARDING:
		/* Adjust sequence number */
		adjust_ingress_ack(th, dcb->pull_seq_offset);
		tcp_modify_timestamp(skb, -dcb->state_priv.pull.delta1, -dcb->state_priv.pull.delta2);

		verdict = DTS_ACCEPT_CT;
		break;

	case DTS_FIRST_SYN_RECV2:
		if (th->rst)
			goto accept_rst;

		if (th->syn || !th->ack)
			break;

		expected = dcb->state_priv.first.ack_for_synack;
		if (th->ack_seq != expected) {
			dprintf(0, "we lose! expected:%u actual:%u\n",
				ntohl(expected), ntohl(th->ack_seq));
			dts_enqueue_reply_rst(dcb,
					ep->laddr, ep->lport,
					ep->raddr, ep->rport,
					dcb->state_priv.first.third_seq,
					dcb->state_priv.first.ack_for_synack);
			do_sspa_lost(ep, iph, th);
			break;
		}

		adjust_ingress_ack(th, dcb->inspect_seq_offset);
		need_to_adjust_ack = 0;

		dcb->state = DTS_INSPECTING;
		/* Fall through */
	case DTS_FIRST_SYN_RECV1:
		need_to_adjust_ack = 0;

		/* TODO:
		 * - code cleanup
		 * - inject RST
		 */
		if (dcb->state == DTS_FIRST_SYN_RECV1
			&& th->ack && !dts_ctl_sspa_disable) {
			do_sspa_lost(ep, iph, th);
			break;
		}

		/* Fall through */
	accept_rst:
	case DTS_INSPECTING:
		if (!tcp_extract_timestamp(skb, &tsval, &tsecr)) {
			dcb->remote_last_timestamp = tsval;
			dprintf(2, "INSPECTING, TS val %u, ecr %u [SYN %u ACK %u]", tsval, tsecr, th->syn, th->ack);
		}
		if (dts_nohandover_mode(dcb->conf))
			verdict = inter_inspecting_nohandover(ep, skb, iph, th,
							need_to_adjust_ack);
		else
			verdict = inter_inspecting(ep, skb, iph, th,
						need_to_adjust_ack);
		break;

	case DTS_PUSHING:
		if (dcb->conf->flags & FTCS_FLAG_DSTSW) {
			dprintf(2, "notify %pI4:%d-%pI4:%d->%pI4\n",
				&ep->raddr, ntohs(ep->rport),
				&ep->laddr, ntohs(ep->lport),
				&dcb->state_priv.push.dest_addr);

#ifdef CONFIG_FTCS_DEBUG
			atomic_inc(&dcb->conf->renotify);
#endif

			ftcs_notify_dest_switched(dcb,
						dcb->state_priv.push.dest_addr);
		}
		break;

	case DTS_ORPHAN:
		err = dts_enqueue_reply_rst(ep->dcb,
					ep->raddr, ep->rport,
					ep->laddr, ep->lport,
					th->ack_seq,
					ack_seq_for_reply(iph, th));
		if (err)
			dprintf(1, "dts_enqueue_reply_rst failed: %d\n", err);

		break;

	case DTS_SSPA_BLOCKING:
	case DTS_SSPA_PASSING:
		if (th->ack && !dts_ctl_sspa_no_ack_snooping) {
			dprintf(2,
				"ACK arrived for (L%pI4:%d-R%pI4:%d)"
				" -- cleaning up\n",
				&ep->laddr, ntohs(ep->lport),
				&ep->raddr, ntohs(ep->rport));
			atomic_inc(&dcb->conf->sspa_ack_snoop);
			dts_unregister_endpoint(ep);
		}
		break;

	default:
		break;
	}

out:

	return verdict;
}

/*
  dcb: locked
  context: softirq
*/
static enum dts_filter_verdict
process_intra_input(struct dts_endpoint *ep, struct sk_buff *skb)
{
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;
	struct dts_cb *dcb = ep->dcb;
	struct iphdr *iph;
	struct tcphdr *th;
	struct net *net;
	struct net_device *new_dev;
	uint32_t tsval, tsecr;

	if (prepare_mangle_tcphdr(skb, &iph, &th))
		goto out;

	switch (dcb->state) {
		struct dts_endpoint *ep_inter;
	case DTS_PULL_SYN_RECV:
		if (!th->syn || th->ack) {
			dprintk(KERN_INFO "%s:%d: DROPPING PACKET!!\n",
				__func__, __LINE__);
			break;
		}
		/* Fall through */
	case DTS_PULLING:
		/* rewrite destination address to the internet address */
		modify_destination(skb, iph, th,
				dcb->conf->inter_addr, th->dest);
		modify_source(skb, iph, th, dcb->raddr, dcb->rport);
		adjust_ingress_ack(th, dcb->pull_seq_offset);
		modify_window_size(skb, th, 0);

		if ((net = dev_net(skb->dev)) != NULL &&
			(new_dev = ip_dev_find(net, iph->daddr)) != NULL) {
			skb->dev = new_dev;
			dev_put(new_dev);
		}

		ep_inter = dts_find_get_endpoint(dcb, iph->daddr, th->dest,
						iph->saddr, th->source);
		if (ep_inter == NULL) {
			dprintf(0, "dts_find_get_endpoint failed!\n");
			break;
		}

		/* Bypass IPVS */
		if (dcb->conf->flags & FTCS_FLAG_DSTSW)
			skb->ipvs_property = 1;

		dts_ct_ingress(ep, iph, th);
		ep_inter->ct = ep->ct;
		dts_release_endpoint(ep_inter);

		if (!tcp_extract_timestamp(skb, &tsval, &tsecr)) {
			dcb->state_priv.pull.delta1 = (s64) dcb->remote_last_timestamp - (s64) tsecr;
			dcb->state_priv.pull.delta2 = (s64) tsecr - (s64) tsval;
		}

		verdict = DTS_ACCEPT_NOCT;

		break;

	case DTS_CONNECTING:
	case DTS_PUSHING:
		adjust_ingress_ack(th, dcb->push_seq_offset);
		verdict = DTS_ACCEPT_CT;

		break;

	case DTS_FORWARDING:
		adjust_ingress_ack(th, dcb->pull_seq_offset);
		verdict = DTS_DROP_CT;
		break;

	default:
		break;
	}

out:
	return verdict;
}

static enum dts_filter_verdict
process_proxy_input(struct dts_endpoint *ep, struct sk_buff *skb)
{
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;
	struct dts_cb *dcb = ep->dcb;
	struct iphdr *iph;
	struct tcphdr *th;

	if (prepare_mangle_tcphdr(skb, &iph, &th))
		goto out;

	switch (dcb->state) {
	case DTS_PUSHING:
		/* TODO: verify window */
		if (th->rst)
			verdict = DTS_ACCEPT_CT;
		break;
	default:
		break;
	}

out:

	return verdict;
}

static int spray_connstat_query(struct ftcs_kconfig *conf,
	const struct dts_endpoint *ep)
{
	struct ftcs_icc_connstatus cs;
	int err;

	cs.laddr = ep->laddr;
	cs.raddr = ep->raddr;
	cs.lport = ep->lport;
	cs.rport = ep->rport;
	cs.has_conn = 0;

	err = ftcs_icc_bcast(conf,
			FTCS_ICC_CONNSTAT_QUERY, &cs, sizeof(cs));
	if (err)
		dprintf(1, "ftcs_icc_bcast failed: %d\n", -err);

	return err;
}

static void register_orphan(struct ftcs_kconfig *conf,
			__be32 laddr, __be16 lport,
			__be32 raddr, __be16 rport,
			__be32 rst_seq, __be32 rst_ack)
{
	struct dts_cb *dcb = NULL;
	struct dts_endpoint *ep = NULL;
	int nr_neighbors;

	dcb = dts_cb_new(conf, DTS_ORPHAN_QUERYING, raddr, rport);
	if (dcb == NULL)
		goto out_free;

	read_lock_bh(&dcb->conf->neighbors_lock);
	nr_neighbors = dcb->conf->nr_neighbors;
	read_unlock_bh(&dcb->conf->neighbors_lock);

	if (nr_neighbors == 0)
		dcb->state = DTS_ORPHAN;

	nsv4_init(&dcb->state_priv.orphan.ns);
	dcb->state_priv.orphan.rst_seq = rst_seq;
	dcb->state_priv.orphan.rst_ack = rst_ack;
	dcb->state_priv.orphan.nr_query_retry = 0;

	ep = dts_new_orphan_endpoint(dcb, laddr, lport, raddr, rport);
	if (ep == NULL)
		goto out_free;

	if (dts_try_register_endpoint(ep))
		goto out_free;

	if (nr_neighbors > 0) {
		spray_connstat_query(conf, ep);
		dts_endpoint_set_timeout(ep, DTS_ORPHAN_QUERY_TIMEOUT);
	}

out_free:
	if (ep)
		dts_release_endpoint(ep);
	if (dcb)
		dts_cb_release(dcb);

	return;
}

/*
  Handle SYN packet for another host
 */
static unsigned int handle_syn_for_other(
	struct ftcs_kconfig *conf, struct sk_buff *skb,
	const struct iphdr *iph, const struct tcphdr *th,
	struct dts_endpoint **pep)
{
	unsigned int ret = NF_DROP;
	struct dts_cb *dcb;
	struct dts_endpoint *ep = NULL, *ep_exists;
	dts_hashtuple_t hash;

	dcb = dts_cb_new(conf, DTS_SSPA_BLOCKING, iph->saddr, th->source);
	if (dcb == NULL)
		return NF_DROP;

	dcb->state_priv.sspa.skb = skb;

	ep = dts_new_backup_endpoint(dcb,
				iph->daddr, th->dest, iph->saddr, th->source);
	if (ep == NULL)
		goto out_free;

	hash = dts_get_hashtuple(ep, ep);
	dts_endpoints_write_lock(hash);

	ep_exists = dts_find_get_endpoint_nolock(
		iph->daddr, th->dest, iph->saddr, th->source);
	if (ep_exists != NULL) {
		struct dts_cb *dcb_exists = ep_exists->dcb;

		dts_cb_lock(dcb_exists);

		if (dcb_exists->state == DTS_SSPA_PASSING) {
			/* Accept the packet */
			dprintf(2, "accepting SYN (L%pI4:%d-R%pI4:%d)\n",
				&iph->daddr, ntohs(th->dest),
				&iph->saddr, ntohs(th->source));
			ret = NF_ACCEPT;
			atomic_inc(&conf->sspa_syn_accept);
			dts_unregister_endpoint_nolock(ep_exists);
		} else {
			dprintf(1, "already exists (L%pI4:%d-R%pI4:%d)\n",
				&iph->daddr, ntohs(th->dest),
				&iph->saddr, ntohs(th->source));
		}
		/* TODO: if there is a delayed-xmit call above,
		   the following dts_cb_unlock() may cause deadlock */
		dts_cb_unlock(dcb_exists);
		dts_release_endpoint(ep_exists);

		dcb->state_priv.sspa.skb = NULL;
		dts_release_endpoint(ep);
		ep = NULL;
		goto out_unlock_free;
	}

	dts_register_endpoint_nolock(ep);

	dts_endpoint_set_timeout(ep, dts_ctl_sspa_timeout);

	ret = NF_STOLEN;

out_unlock_free:
	dts_endpoints_write_unlock(hash);
out_free:
	*pep = ep;
	if (dcb)
		dts_cb_release(dcb);

	return ret;
}

/*
  Dirty hack for IPVS:

  Unconditionally accept packets from the client-side interface
  so that the LVS director can handle them.

  The assumption here is to have two interfaces, one for
  client-server comm. and the other for server-server comm.,
  and real servers are represented by the "intranet"-side
  addresses.
*/
static int bypass_dts_for_ipvs(struct ftcs_kconfig *conf,
			const struct iphdr *iph, const struct tcphdr *th,
			const struct net_device *in)
{
	int ret = 0;
	struct net_device *dev;

	if (!(conf->flags & FTCS_FLAG_DSTSW))
		return 0;

	if (iph->daddr != conf->inter_addr)
		return 0;

	dev = ip_dev_find(&init_net, iph->daddr);
	if (dev == NULL)
		return 0;

	/*
	  Another dirty hack for loopback packets

	  Loopback packets are usually generated by the DTS filter
	  and they have to be processed by the filter.
	*/
	if (dev == in && !(dev->flags & IFF_LOOPBACK)) {
		dprintf(1, "accepting %pI4:%d->%pI4:%d %c%c%c%c "
			"(dev:%s, in:%s)\n",
			&iph->saddr, ntohs(th->source),
			&iph->daddr, ntohs(th->dest),
			th->syn ? 'S':'-',
			th->ack ? 'A':'-',
			th->fin ? 'F':'-',
			th->rst ? 'R':'-',
			dev->name, in->name);
		ret = 1;
	}

	dev_put(dev);

	return ret;
}

static
unsigned int process_syn(struct sk_buff *skb,
			const struct iphdr *iph, const struct tcphdr *th,
			const struct net_device *in,
			struct dts_endpoint **pep)
{
	int err;
	struct ftcs_kconfig *conf;
	int ret = NF_ACCEPT;
#ifdef CONFIG_FTCS_PROFILING
	struct timeval tv0, tv1;

	do_gettimeofday(&tv0);
#endif

	conf = ftcs_find_get_config(iph->daddr, th->dest);

	if (conf == NULL) {
		ret = NF_STOP;
		goto out;
	}

	if (bypass_dts_for_ipvs(conf, iph, th, in)) {
		ret = NF_STOP;
		goto out_release_config;
	}

	if (iph->daddr == conf->inter_addr) {
		int secondary = 0;
		__be32 resp_node, prev_node;

		resp_node = ftcs_get_responsible_nodes(conf,
						iph->saddr, th->source,
						&prev_node);

		if (resp_node != conf->intra_addr) {
#ifdef CONFIG_FTCS_PROFILING
			struct timeval tv0, tv1;
#endif
			if (prev_node != conf->intra_addr
				|| dts_ctl_sspa_disable) {
				dprintf(2,
					"dropping SYN by sspa_disable / "
					"this is not the previous node "
					"(%pI4:%d->%pI4:%d)\n",
					&iph->saddr, ntohs(th->source),
					&iph->daddr, ntohs(th->dest));
				ret = NF_DROP;
				goto out_release_config;
			}

#ifdef CONFIG_FTCS_PROFILING
			do_gettimeofday(&tv0);
#endif
			ret = handle_syn_for_other(conf, skb, iph, th, pep);
#ifdef CONFIG_FTCS_PROFILING
			do_gettimeofday(&tv1);
			atomic_add(dts_prof_tv_diff(&tv1, &tv0),
				&dts_prof_entries[DTS_PROF_IN_HSFO].sum);
			atomic_inc(&dts_prof_entries[DTS_PROF_IN_HSFO].nr_samples);
#endif
			dprintf(2, "SYN for another host: verdict=%u "
				"(%pI4:%d->%pI4:%d)\n",
				ret,
				&iph->saddr, ntohs(th->source),
				&iph->daddr, ntohs(th->dest));

			if (ret != NF_ACCEPT)
				goto out_release_config;
			secondary = 1;
		}
		dprintf(2, "accepting SYN: "
			"%pI4:%d->%pI4:%d\n",
			&iph->saddr, ntohs(th->source),
			&iph->daddr, ntohs(th->dest));

		if (dts_nohandover_mode(conf)) {
#ifdef CONFIG_FTCS_PROFILING
			struct timeval tv0, tv1;
			do_gettimeofday(&tv0);
#endif
			err = handle_inter_syn_nohandover(conf,
							skb, iph, th,
							secondary, pep);
#ifdef CONFIG_FTCS_PROFILING
			do_gettimeofday(&tv1);
			atomic_add(dts_prof_tv_diff(&tv1, &tv0),
				&dts_prof_entries[DTS_PROF_IN_SYN_NHO].sum);
			atomic_inc(&dts_prof_entries[DTS_PROF_IN_SYN_NHO].nr_samples);
#endif
		} else
			err = handle_inter_syn_input(conf,
						skb, iph, th, secondary, pep);
	} else
		err = handle_intra_syn_input(conf, skb, iph, th, pep);

	if (err < 0)
		ret = NF_DROP;

out_release_config:
	ftcs_release_config(conf);
out:
#ifdef CONFIG_FTCS_PROFILING
	do_gettimeofday(&tv1);
	atomic_add(dts_prof_tv_diff(&tv1, &tv0),
		&dts_prof_entries[DTS_PROF_IN_SYN].sum);
	atomic_inc(&dts_prof_entries[DTS_PROF_IN_SYN].nr_samples);
#endif
	return ret;
}

static unsigned int
process_nonsyn_wo_ep(const struct iphdr *iph, const struct tcphdr *th,
		const struct net_device *in)
{
	struct ftcs_kconfig *conf;
	unsigned int ret = NF_DROP;

	conf = ftcs_find_get_config(iph->daddr, th->dest);

	if (conf == NULL)
		return NF_ACCEPT;

	if (bypass_dts_for_ipvs(conf, iph, th, in)) {
		ret = NF_ACCEPT;
		goto out_release_config;
	}

#ifdef CONFIG_FTCS_DEBUG
	if (conf->flags & FTCS_FLAG_DSTSW) {
		dprintf(2,
			"dropping in DSTSW mode: %pI4:%d-%pI4:%d %c%c%c%c\n",
			&iph->saddr, ntohs(th->source),
			&iph->daddr, ntohs(th->dest),
			th->syn ? 'S':'-',
			th->ack ? 'A':'-',
			th->fin ? 'F':'-',
			th->rst ? 'R':'-');
	}
#endif /* #ifdef CONFIG_FTCS_DEBUG */

	if (ftcs_mourning(conf)
		&& ftcs_responsible_for_packet(conf, iph->saddr, th->source)) {
			register_orphan(conf, iph->daddr, th->dest,
					iph->saddr, th->source,
					th->ack_seq,
					ack_seq_for_reply(iph, th));
	}

out_release_config:
	ftcs_release_config(conf);

	return ret;
}

unsigned int ftcs_preroute_filter(
	unsigned int hooknum, struct sk_buff *skb,
	const struct net_device *in, const struct net_device *out,
	int (*okfn)(struct sk_buff *))
{
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *th, _tcphdr;
	struct dts_endpoint *ep = NULL;
	unsigned int ret = NF_ACCEPT;
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;
#ifdef CONFIG_FTCS_PROFILING
	struct timeval tv0, tv1;
	int do_prof = 0;
#endif

	if (iph->protocol != IPPROTO_TCP)
		return NF_ACCEPT;

	if (unlikely(iph->frag_off & __constant_htons(IP_OFFSET))) {
		/* Pass fragments */
		dprintk(KERN_INFO
			"ftcs_preroute_filter: passing fragment packets\n");
		return NF_ACCEPT;
	}

	th = skb_header_pointer(skb, iph->ihl*4, sizeof(_tcphdr), &_tcphdr);
#ifdef CONFIG_FTCS_DEBUG
	{
		int hlen = skb_headlen(skb);
		int offset = iph->ihl*4;
		int len = sizeof(_tcphdr);

		if (hlen - offset < len) {
			dprintk(KERN_INFO
				"ftcs_preroute_filter: hlen - offset < len\n");
		}
	}
#endif

	if (unlikely(th == NULL)) {
		dprintk(KERN_INFO "ftcs_preroute_filter: th == NULL\n");
		return NF_ACCEPT;
	}

	if (th->syn && !th->ack) {
		ret = process_syn(skb, iph, th, in, &ep);

		switch (ret) {
		case NF_STOP:
			/* Unconditionally accept */
			ret = NF_ACCEPT;
			goto out_release_ep;

		case NF_ACCEPT:
			/* Continue the filtering process */
			break;

		default:
			/* Unconditionally reject */
			goto out_release_ep;
		}

		if (ep)
			goto have_ep;
	}

	/*
	  Incoming packet handling rule for non-SYN segments:

	  ___________________________
	  |	    |______conf_______|
	  |________ |Yes_____|No______|
	  |ep  |Yes |ACCEPT? |ACCEPT? |
	  |___ |No__|DROP____|ACCEPT__|

	  ACCEPT:  accept the packet unconditionally
	  ACCEPT?: pass the packet to the filter
	  DROP:	   drop the packet
	*/

	ep = dts_find_get_endpoint_any(iph->daddr, th->dest,
				iph->saddr, th->source);
	if (ep == NULL) {
		ret = process_nonsyn_wo_ep(iph, th, in);
		goto out;
	}
have_ep:
#ifdef CONFIG_FTCS_PROFILING
	do_gettimeofday(&tv0);
	do_prof = 1;
#endif
	if (bypass_dts_for_ipvs(ep->dcb->conf, iph, th, in)) {
		ret = NF_ACCEPT;
		goto out_release_ep;
	}

	dts_cb_lock(ep->dcb);

	LOG_PACKET(ep->dcb, skb, DTS_LOG_INGRESS);

	if (dts_ct_verify_ingress(ep, skb) < 0
		&& dts_ct_terminated(ep)) {
		/* dts_unregister_endpoint(ep); */
		ret = NF_DROP;
		goto out_release;
	}

	switch (ep->lep) {
	case DTS_LEP_INTER:
		verdict = process_inter_input(ep, skb);
		break;

	case DTS_LEP_INTRA:
		verdict = process_intra_input(ep, skb);
		break;

	case DTS_LEP_PROXY:
		verdict = process_proxy_input(ep, skb);
		break;
	}

	if (ep->dcb->conf->flags & FTCS_FLAG_DSTSW)
		skb->ipvs_property = 1;

	if (verdict & DTS_CT_MASK) {
		struct iphdr _iph;

		iph = skb_header_pointer(skb,
					0,
					sizeof(*iph),
					&_iph);
		if (iph == NULL) {
			ret = NF_DROP;
			goto out_release;
		}

		th = skb_header_pointer(skb,
					iph->ihl*4,
					sizeof(*th),
					&_tcphdr);
		if (th == NULL) {
			ret = NF_DROP;
			goto out_release;
		}

		dts_ct_ingress(ep, iph, th);
	}

	ret = (verdict & DTS_ACCEPT_MASK) ? NF_ACCEPT : NF_DROP;

out_release:

	LOG_PACKET(ep->dcb, skb,
		DTS_LOG_INGRESS|DTS_LOG_VERDICT
		|(ret == NF_ACCEPT?DTS_LOG_ACCEPT:0));

	dts_cb_unlock(ep->dcb);

out_release_ep:
	if (ep)
		dts_release_endpoint(ep);

#ifdef CONFIG_FTCS_PROFILING
	if (do_prof) {
		do_gettimeofday(&tv1);
		atomic_add(dts_prof_tv_diff(&tv1, &tv0),
			&dts_prof_entries[DTS_PROF_IN_BODY].sum);
		atomic_inc(&dts_prof_entries[DTS_PROF_IN_BODY].nr_samples);
	}
#endif

out:

	return ret;
}

static int handle_connstat_query(const struct ftcs_icc_connstatus *cs)
{
	struct dts_endpoint *ep;
	struct ftcs_kconfig *conf;
	struct ftcs_icc_connstatus cs_reply = *cs;
	int err = 0;

	ep = dts_find_get_endpoint_any(cs->laddr, cs->lport,
				cs->raddr, cs->rport);
	cs_reply.has_conn = (ep != NULL);

	conf = ftcs_find_get_config(cs->laddr, cs->lport);
	if (conf != NULL) {
		__be32 dest_node;
		dest_node = ftcs_responsible_node(conf,
						cs->raddr, cs->rport);
		if (dest_node != FTCS_IPADDR_INVALID)
			err = ftcs_icc_xmit(conf, dest_node,
					FTCS_ICC_CONNSTAT_ANSWER,
					&cs_reply, sizeof(cs_reply));

		ftcs_release_config(conf);
	} else
		err = -EINVAL;

	if (ep)
		dts_release_endpoint(ep);

	return err;
}

static void print_neighset(const struct neighsetv4 *ns)
{
	struct nsv4_iter iter;

	if (nsv4_first(ns, &iter) < 0)
		return;
	do {
		__be32 addr = nsv4_iter_address(&iter);
		dprintf(2, "%pI4\n", &addr);
	} while (nsv4_next(&iter) == 0);
}

static int switch_to_orphan(struct dts_endpoint *ep)
{
	int err;

	err = dts_enqueue_reply_rst(ep->dcb,
				ep->raddr, ep->rport,
				ep->laddr, ep->lport,
				ep->dcb->state_priv.orphan.rst_seq,
				ep->dcb->state_priv.orphan.rst_ack);
	if (err)
		dprintf(1, "dts_enqueue_reply_rst failed: %d\n", err);

	ep->dcb->state = DTS_ORPHAN;
	dts_endpoint_set_timeout(ep, DTS_ORPHAN_TIMEOUT);

	return err;
}

static
int handle_connstat_answer(__be32 saddr, const struct ftcs_icc_connstatus *cs)
{
	struct dts_endpoint *ep;
	struct neighsetv4 *ns;
	int ret = 0;

	ep = dts_find_get_endpoint_any(cs->laddr, cs->lport,
				cs->raddr, cs->rport);
	if (ep == NULL)
		return -EINVAL;

	dts_cb_lock(ep->dcb);

	if (ep->dcb->state != DTS_ORPHAN_QUERYING)
		goto out_unlock;

	if (!ftcs_neighbor(ep->dcb->conf, saddr)) {
		dprintf(1, "not from a neighbor: %pI4\n", &saddr);
		goto out_unlock;
	}

	ns = &ep->dcb->state_priv.orphan.ns;

	if (cs->has_conn) {
		nsv4_clear(ns);
		ep->dcb->state = DTS_IGNORING;
		dts_endpoint_set_timeout(ep,  DTS_IGNORING_TIMEOUT);
	} else {
		read_lock(&ep->dcb->conf->neighbors_lock);

		if (nsv4_add(ns, saddr) < 0)
			goto out_unlock2;

		if (nsv4_size(ns) >= ep->dcb->conf->nr_neighbors) {
			dprintf(1, "negative answer from all neighbors,"
				"rport=%d\n", ntohs(cs->rport));
			print_neighset(ns);

			ret = switch_to_orphan(ep);
		}

	out_unlock2:
		read_unlock(&ep->dcb->conf->neighbors_lock);
	}

out_unlock:
	dts_cb_unlock(ep->dcb);

	dts_release_endpoint(ep);

	return ret;
}

static void handle_synack_sent(const struct ftcs_icc_quadruple *q)
{
	struct dts_endpoint *ep;

	ep = dts_find_get_endpoint_any(q->laddr, q->lport, q->raddr, q->rport);
	if (ep == NULL) {
		dprintf(2, "not found: (L%pI4:%d-R%pI4:%d)\n",
		&q->laddr, ntohs(q->lport),
		&q->raddr, ntohs(q->rport));
		return;
	}

	dts_cb_lock(ep->dcb);
	if (ep->dcb->state != DTS_SSPA_BLOCKING
		&& ep->dcb->state != DTS_SSPA_PASSING) {
		dprintf(0, "state == %s\n", dts_state_str(ep->dcb->state));
		goto out_unlock_release;
	}

	dprintf(2, "received SYNACK_SENT for (L%pI4:%d-R%pI4:%d)\n",
		&q->laddr, ntohs(q->lport),
		&q->raddr, ntohs(q->rport));

	atomic_inc(&ep->dcb->conf->sspa_synnotify_recv);

	dts_unregister_endpoint(ep);

out_unlock_release:
	dts_cb_unlock(ep->dcb);
	dts_release_endpoint(ep);
}

static int icc_receive(struct sk_buff *skb)
{
	struct iphdr *iph;
	struct ftcs_icchdr *icchdr;

	if (!pskb_may_pull(skb, sizeof(*icchdr))) {
		dprintf(1, "pskb_may_pull failed\n");
		goto out_free;
	}

	iph = ip_hdr(skb);

	if (skb->dev) {
		if (skb->dev->flags & IFF_LOOPBACK) {
			dprintf(2, "packet from loopback device, dropping\n");
			goto out_free;
		}
	}
	else {
		dprintf(1, "skb->dev == NULL\n");
		goto out_free;
	}

	icchdr = (struct ftcs_icchdr *) skb->data;
	skb_pull(skb, sizeof(*icchdr));

	switch (icchdr->type) {
		struct ftcs_icc_connstatus *cs;
		struct ftcs_icc_quadruple *q;

	case FTCS_ICC_CONNSTAT_QUERY:
		if (!pskb_may_pull(skb, sizeof(*cs))) {
			dprintf(1, "pskb_may_pull failed\n");
			goto out_free;
		}
		cs = (struct ftcs_icc_connstatus *) skb->data;

		handle_connstat_query(cs);
		break;

	case FTCS_ICC_CONNSTAT_ANSWER:
		if (!pskb_may_pull(skb, sizeof(*cs))) {
			dprintf(1, "pskb_may_pull failed\n");
			goto out_free;
		}
		cs = (struct ftcs_icc_connstatus *) skb->data;

		dprintf(2, "CONNSTAT_ANSWER from %pI4: %pI4:%d %d\n",
			&iph->saddr,
			&cs->raddr, ntohs(cs->rport),
			cs->has_conn);

		handle_connstat_answer(iph->saddr, cs);

		break;

	case FTCS_ICC_SYNACK_SENT:
		if (!pskb_may_pull(skb, sizeof(*q))) {
			dprintf(1, "pskb_may_pull failed\n");
			goto out_free;
		}
		q = (struct ftcs_icc_quadruple *) skb->data;
		handle_synack_sent(q);
		break;

	default:
		dprintf(1, "unknown type: %d\n", icchdr->type);
		break;
	}

out_free:
	kfree_skb(skb);

	return 0;
}

struct net_protocol ftcs_icc_protocol = {
	.handler = icc_receive,
	.no_policy = 1,
	.netns_ok  = 1,
};
