#ifndef DTS_INTERNAL_H
#define DTS_INTERNAL_H

#include <linux/kernel.h>
#include <linux/version.h>

#include <linux/ip.h>
#include <linux/tcp.h>

#include <net/ip.h>
#include <net/tcp.h>

#include <linux/netfilter.h>

#include "neighsetv4.h"

/*
 * Debug macros
 */

#ifdef CONFIG_FTCS_DEBUG
#  define LOG_PACKET(dcb, skb, flags)	\
	dts_log(dcb, skb, flags)
#else
#  define LOG_PACKET(dcb, skb, flags) do {} while (0)
#endif

/* NIPQUAD macros -- disappeared from 2.6.36 */

#ifndef NIPQUAD
#define NIPQUAD(addr)			 \
	((unsigned char *)&addr)[0],	 \
		((unsigned char *)&addr)[1],	\
		((unsigned char *)&addr)[2],	\
		((unsigned char *)&addr)[3]
#endif

#ifndef NIPQUAD_FMT
#define NIPQUAD_FMT "%u.%u.%u.%u"
#endif

/* Compatibility macro for struct ctl_path */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,33)
#define CTL_PATH_INITIALIZER(_proc, _ctl)       \
        { .procname = (_proc), }
#else
#define CTL_PATH_INITIALIZER(_proc, _ctl)       \
        { .procname = (_proc), .ctl_name = (_ctl), }
#endif

static inline
void dts_skb_set_checksum_start(struct sk_buff *skb, unsigned short off)
{
	skb->csum_start = skb->data - skb->head + off;
}

static inline
void dts_skb_set_checksum_offset(struct sk_buff *skb, unsigned short csum_offset)
{
	skb->csum_offset = csum_offset;
}

static inline
const void *dts_skb_front(const struct sk_buff *skb, size_t *contig_len)
{
	*contig_len = skb_headlen(skb);

	return skb->data;
}

static inline
void dts_calc_checksum_partial(struct sk_buff *skb, struct iphdr *iph, struct tcphdr *th)
{
	size_t headlen;
	th->check = ~tcp_v4_check(skb->len, iph->saddr, iph->daddr, 0);
	dts_skb_set_checksum_start(skb,
		(unsigned char *)th - (unsigned char *)dts_skb_front(skb, &headlen));

	dts_skb_set_checksum_offset(skb, offsetof(struct tcphdr, check));
}

typedef struct dts_hashtuple {
	unsigned int keys[2];
} dts_hashtuple_t;

extern struct kmem_cache *dts_cb_cachep;
extern struct kmem_cache *dts_endpoint_cachep;

extern struct proc_dir_entry *dts_proc_net_dts;

struct ftcs_kconfig;

struct dts_delayed_work {
	struct list_head list;
	void (*handler)(void *);
	void *data;
};

static inline void dts_delayed_work_init(
	struct dts_delayed_work *work, void (*handler)(void *), void *data)
{
	INIT_LIST_HEAD(&work->list);
	work->handler = handler;
	work->data    = data;
}

enum dts_filter_verdict_masks {
	DTS_ACCEPT_MASK = (1 << 0),
	DTS_CT_MASK     = (1 << 1),
};

enum dts_filter_verdict {
	DTS_DROP_NOCT   = 0,
	DTS_DROP_CT     = DTS_CT_MASK,
	DTS_ACCEPT_NOCT = DTS_ACCEPT_MASK,
	DTS_ACCEPT_CT   = (DTS_ACCEPT_MASK|DTS_CT_MASK),
};

enum dts_local_endpoint {
	DTS_LEP_INTER,
	DTS_LEP_PROXY,
	DTS_LEP_INTRA,
	/* TODO:
	   DTS_LEP_INTRA_PUSH,
	   DTS_LEP_INTRA_PULL,
	*/
};

enum dts_remote_endpoint {
	DTS_REP_CLIENT,
	DTS_REP_PUSH_NEIGHBOR,
	DTS_REP_PULL_NEIGHBOR,
};

/* State for one-way connection */
enum dts_half_state {
	/* Connection is opened */
	DTS_HST_OPEN,

	/* FIN has been sent out first from this side and not ACKed yet */
	DTS_HST_FIN_SENT1,

	/* FIN has been sent out from this side after receiving FIN from
	   the peer. FIN has not been ACKed yet. */
	DTS_HST_FIN_SENT2,

	/* FIN has been sent out first from this side and already ACKed */
	DTS_HST_TIME_WAIT,

	/* FIN has been sent out from this side after receiving FIN from
	   the peer and already ACKed. */
	DTS_HST_CLOSED,
};

struct dts_hst_tuple {
	enum dts_half_state ingress_state;
	enum dts_half_state egress_state;
};

/*
  Endpoint cleanup on iggerular cases:
  There is a case where one side of the connection cannot reply
  because of crash or overload.
  In such a case, packet retransmission from other side will be observed.

  Algorithm:
  If the retransmission is seen, mark that retransmission is sent
  and start the timer.
  If the timer expires, the filter regards that the connection is dead
  and the endpoint will be removed.
  If the ACK comes before the timer expires, the filter regards that
  the connection is still alive and clear the timer.
  (If the endpoint had another timer before the retransmission,
  use that value. ct->max_timeout is for this purpose.)
  Currently we do not check the ACK number for simplicity.
  If the ACK number is wrong (other than expected), the TCP stack will
  retransmit segments again.
 */

enum dts_ct_flag {
	DTS_CT_EGRESS_SEEN = (1 << 0),

	/* ACK bit on the egress segment was seen */
	DTS_CT_EGRESS_ACK_SEEN = (1 << 1),
	DTS_CT_EMULATE_TCP = (1 << 2),
	DTS_CT_INGRESS_SEEN = (1 << 3),
	DTS_CT_EGRESS_RTX_SEEN = (1 << 4),
	DTS_CT_INGRESS_RTX_SEEN = (1 << 5),
	DTS_CT_INGRESS_ACK_SEEN = (1 << 6),
};

struct dts_conntrack {
	uint8_t ingress_state;
	uint8_t egress_state;
	uint16_t flags;
	uint32_t rcv_nxt;
	uint32_t rcv_nxt_max;
	uint32_t sent_max;
	uint32_t rcvd_max;

	/*
	  Maximum ACK number seen on ingress segments

	  Note: sent_ack_max is not needed
	  (it is equal to rcv_nxt!!)
	 */
	uint32_t rcvd_ack_max;

	/*
	  max timeout value in jiffies
	  zero means that no timeout is needed
	*/
	uint32_t max_timeout;
};

struct dts_endpoint {
	struct hlist_node node;
	atomic_t refcnt;

	/* Immutable members -- not modified after initialization */
	struct dts_cb *dcb;
	enum dts_local_endpoint  lep;
	enum dts_remote_endpoint rep;
	__be32 laddr;
	__be32 raddr;
	__be16 lport;
	__be16 rport;

	struct timer_list timer;

	/* Mutable members
	   Locking rule: protected by ep->dcb->lock */
	struct dts_conntrack ct;

	unsigned long timeout_jiffies;
};

struct sk_buff;

union dts_state_priv {
	struct {
		__be32 first_syn_seq;

		/* Expected ACK number in the reply to our SYN+ACK,
		   from the client's point of view */
		__be32 ack_for_synack;

		/* Expected sequenuce number on the third packet in the
		   3-way handshake, which is from the client */
		__be32 third_seq;

		/* SSPA: Are we the secondary node? */
		int secondary;
	} first;

	/* Connection pushing */
	struct {
		__be32 local_syn_seq;
		__be32 remote_syn_seq;
		uint32_t remote_window;
		uint32_t handover_len;
		__be32 dest_addr;
	} push;

	/* Connection pulling */
	struct {
		__be32 first_syn_seq;
		uint32_t remote_window;
		__be32 peer_addr;
		__be16 peer_port;
		__be32 egress_syn_seq;

		/*
		  Last ACK number seen in egress packet
		  This ACK represents the number before mangled by
		  the filter.
		*/
		uint32_t last_tx_ack;

		/* Length of received (and ACKed) data */
		uint32_t rx_bytes;

		uint32_t handover_len;

		/* for TCP option timestamps */
		s64 delta1;
		s64 delta2;
	} pull;

	struct {
		__be32 rst_seq;
		__be32 rst_ack;
		int nr_query_retry;
		/* set of neighbors which do not have
		   an effective endpoint */
		struct neighsetv4 ns;
	} orphan;

	struct {
		struct sk_buff *skb;
	} sspa;
};

struct dts_cb {
	spinlock_t lock;
	atomic_t refcnt;
	struct ftcs_kconfig *conf;

	struct dts_endpoint ep_storage;

	enum dts_state state;

	struct list_head delayed_work_list;

#ifdef CONFIG_FTCS_DEBUG
	int id;
#endif

	/*
	  These addresses and port numbers do not need
	  serialization for reading, because they are set
	  only the initialization time.
	*/

	__be32 raddr;
	__be16 rport;

	/*
	  Sequence number, to be subtracted from SEQ field of egress segments
	*/
	/* TODO: should these be in endpoints? */
	uint32_t inspect_seq_offset;
	uint32_t push_seq_offset;
	uint32_t pull_seq_offset;

	union dts_state_priv state_priv;

	struct tcp_options_received tcp_options;
	uint32_t remote_last_timestamp;
};

static inline void dts_cb_lock(struct dts_cb *dcb)
{
	spin_lock_bh(&dcb->lock);
}

/* TODO: too big to be implemented as an inline function?? */
static inline void dts_cb_unlock(struct dts_cb *dcb)
{
	struct list_head work_list;

	INIT_LIST_HEAD(&work_list);
	list_replace_init(&dcb->delayed_work_list, &work_list);

	spin_unlock_bh(&dcb->lock);

	while (!list_empty(&work_list)) {
		struct dts_delayed_work *work;
		work = list_entry(work_list.next,
				struct dts_delayed_work, list);
		list_del(&work->list);
		work->handler(work->data);
	}
}

static inline void dts_cb_hold(struct dts_cb *dcb)
{
	atomic_inc(&dcb->refcnt);
}

static inline
void dts_cb_register_delayed_work(struct dts_cb *dcb,
				struct dts_delayed_work *work)
{
	list_add_tail(&work->list, &dcb->delayed_work_list);
}

extern void __dts_cb_release(struct dts_cb *dcb);

static inline
void dts_cb_release(struct dts_cb *dcb)
{
	if (atomic_dec_and_test(&dcb->refcnt))
		__dts_cb_release(dcb);
}

#define DTS_TCPOPT_CONN_PUSH 36

/*
  Note: the size of this structure must be multiply of four.
 */
struct dts_conn_tcp_opt {
	u8 type; /* = DTS_TCPOPT_CONN_PUSH */
	u8 len;  /* = DTS_TCPOLEN_CONN_PUSH */
	u8 size; /* sizeof(dts_conn_opt) */
	u8 pad; /* padding */
} __attribute__((packed));

#define DTS_TCPOLEN_CONN_PUSH (sizeof(struct dts_conn_tcp_opt))

struct dts_conn_opt {
	__be32 real_raddr;
	__be16 real_rport;

	/* Sequence number which a server (proxy) first sent to the client
	   (with SYN bit set) */
	__be32 initial_egress_seq;

	/* Window size of the client at the time when handover is issued.
	   We use 32-bit integer for window-scaling */
	__be32 remote_window;

	/* Length of data to be passed via this connection */
	__be32 handover_len;

	/* last timestamp of the client */
	__be32 remote_last_timestamp;
} __attribute__((packed));

#define IPPROTO_FTCSICC2 252
#define IPPROTO_FTCSICC 253

enum ftcs_icc_type {
	FTCS_ICC_CONNSTAT_QUERY,
	FTCS_ICC_CONNSTAT_ANSWER,
	FTCS_ICC_DEST_SWITCH,
	FTCS_ICC_SYNACK_SENT,
};

struct ftcs_icchdr {
	u8 type;
	u8 padding;
	__be16 length; /* length of content */
};

struct ftcs_icc_quadruple {
	__be32 laddr;
	__be32 raddr;
	__be16 lport;
	__be16 rport;
};

struct ftcs_icc_connstatus {
	__be32 laddr;
	__be32 raddr;
	__be16 lport;
	__be16 rport;
	int has_conn;
};

struct ftcs_icc_destsw {
	__be32 laddr;
	__be32 raddr;
	__be16 lport;
	__be16 rport;
	__be32 new_dest;
};

#define DTS_ORPHAN_QUERY_RETRY_MAX 2
#define DTS_ORPHAN_QUERY_TIMEOUT (HZ/10)
#define DTS_ORPHAN_TIMEOUT (120*HZ)
#define DTS_IGNORING_TIMEOUT (30*HZ)
#define DTS_MOURNING_TIMEOUT (120*HZ)

#define DTS_SSPA_TIMEOUT_DEFAULT (HZ/5)
#define DTS_INITIAL_CONN_TIMEOUT_DEFAULT (180*HZ)
#define DTS_CONN_RETRANS_TIMEOUT_DEFAULT (180*HZ)

/*
 * Global configuration variables
 */
extern int dts_ctl_sspa_timeout;
/* TODO: should those be in per-config flags? */
extern int dts_ctl_sspa_disable;
extern int dts_ctl_sspa_no_ack_snooping;
extern int dts_ctl_sspa_no_notification;
extern int dts_ctl_handover_timeout;
extern int dts_ctl_initial_conn_timeout;
extern int dts_ctl_conn_retrans_timeout;

/*
 * dts_endpoints.c
 */

dts_hashtuple_t dts_get_hashtuple(struct dts_endpoint *, struct dts_endpoint *);
unsigned int dts_endpoint_hash(const struct dts_endpoint *ep);
void dts_endpoints_write_lock(dts_hashtuple_t);
void dts_endpoints_write_unlock(dts_hashtuple_t);
void dts_endpoint_write_lock(unsigned int hash);
void dts_endpoint_write_unlock(unsigned int hash);

struct dts_endpoint *dts_new_endpoint(
	struct dts_cb *dcb,
	enum dts_local_endpoint lep, __be32 laddr, __be16 lport,
	enum dts_remote_endpoint rep, __be32 raddr, __be16 rport);
void dts_init_endpoint(struct dts_endpoint *ep, struct dts_cb *dcb,
		enum dts_local_endpoint lep, __be32 laddr, __be16 lport,
		enum dts_remote_endpoint rep, __be32 raddr, __be16 rport);
struct dts_endpoint *dts_new_orphan_endpoint(
	struct dts_cb *dcb,
	__be32 laddr, __be16 lport, __be32 raddr, __be16 rport);
struct dts_endpoint *dts_new_backup_endpoint(
	struct dts_cb *dcb,
	__be32 laddr, __be16 lport, __be32 raddr, __be16 rport);

void dts_register_endpoint(struct dts_endpoint *ep);
void dts_register_endpoint_nolock(struct dts_endpoint *ep);
int dts_try_register_endpoint(struct dts_endpoint *ep);
struct dts_endpoint *dts_find_get_endpoint(struct dts_cb *dcb,
					__be32 laddr, __be16 lport,
					__be32 raddr, __be16 rport);
struct dts_endpoint *dts_find_get_endpoint_any(__be32 laddr, __be16 lport,
					__be32 raddr, __be16 rport);
struct dts_endpoint *dts_find_get_endpoint_nolock(
	__be32 laddr, __be16 lport, __be32 raddr, __be16 rport);
void dts_unregister_endpoint(struct dts_endpoint *ep);
void dts_unregister_endpoint_nolock(struct dts_endpoint *ep);
void dts_clear_endpoints(struct ftcs_kconfig *conf);

int dts_endpoint_init(void);
void dts_endpoint_exit(void);

static inline void dts_release_endpoint(struct dts_endpoint *ep)
{
	if (atomic_dec_and_test(&ep->refcnt)) {
		int need_free = (ep != &ep->dcb->ep_storage);

		dts_cb_release(ep->dcb);
		if (need_free)
			kmem_cache_free(dts_endpoint_cachep, ep);
	}
}

static inline void dts_endpoint_hold(struct dts_endpoint *ep)
{
	atomic_inc(&ep->refcnt);
}

static inline
void dts_endpoint_set_timeout(struct dts_endpoint *ep, unsigned long relative)
{
	ep->timeout_jiffies = jiffies + relative;
	if (!mod_timer(&ep->timer, jiffies + relative)) {
		/* Reference count for timer */
		dts_endpoint_hold(ep);
	}
}

static inline void dts_endpoint_clear_timeout(struct dts_endpoint *ep)
{
	ep->timeout_jiffies = 0;
	if (del_timer(&ep->timer))
		dts_release_endpoint(ep);
}

static inline __be32 be32_add(__be32 base, int d)
{
	return htonl(ntohl(base) + d);
}

static inline void add_to_ack(const struct sk_buff *skb,
			struct tcphdr *th, uint32_t offset)
{
	__be32 new_ack = htonl(ntohl(th->ack_seq) + offset);

	if (skb == NULL || skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace4(&th->check, th->ack_seq, new_ack);

	th->ack_seq = new_ack;
}

static inline void adjust_ingress_ack(struct tcphdr *th, uint32_t offset)
{
	add_to_ack(NULL, th, offset);
}

static inline void adjust_egress_seq(const struct sk_buff *skb,
				struct tcphdr *th, uint32_t offset)
{
	__be32 new_seq = htonl(ntohl(th->seq) - offset);

	/* If the hardware checksumming is available,
	   do not calculate checksum other than that of pseudo-header */
	if (skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace4(&th->check, th->seq, new_seq);

	th->seq = new_seq;
}

/*
  Modify the window size field
  Window size must be fit into 16-bit field.
 */
static inline void modify_window_size(const struct sk_buff *skb,
				struct tcphdr *th, uint16_t size)
{
	__be16 be_window = htons(size);

	if (skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace2(&th->check, th->window, be_window);

	th->window = be_window;
}

static inline void drop_fin_flag(const struct sk_buff *skb, struct tcphdr *th)
{
	__be32 flag_word;

	if (!th->fin)
		return;

	flag_word = tcp_flag_word(th) & ~TCP_FLAG_FIN;

	if (skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace4(&th->check, tcp_flag_word(th), flag_word);

	tcp_flag_word(th) = flag_word;
}

static inline void add_fin_flag(const struct sk_buff *skb, struct tcphdr *th)
{
	__be32 flag_word;

	if (th->fin)
		return;

	flag_word = tcp_flag_word(th) | TCP_FLAG_FIN;

	if (skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace4(&th->check, tcp_flag_word(th), flag_word);

	tcp_flag_word(th) = flag_word;
}

static inline void add_psh_flag(struct tcphdr *th)
{
	__be32 flag_word;

	if (th->psh)
		return;

	flag_word = tcp_flag_word(th) | TCP_FLAG_PSH;

	csum_replace4(&th->check, tcp_flag_word(th), flag_word);

	tcp_flag_word(th) = flag_word;
}

static inline void inc_egress_ack(const struct sk_buff *skb,
				struct tcphdr *th)
{
	__be32 new_ack = be32_add(th->ack_seq, 1);

	/* If the hardware checksumming is available,
	   do not calculate checksum other than that of pseudo-header */
	if (skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace4(&th->check, th->ack_seq, new_ack);

	th->ack_seq = new_ack;
}

/*
  Modify the source address and the port number then adjust checksum
*/
static inline void modify_source(const struct sk_buff *skb,
				struct iphdr *iph, struct tcphdr *th,
				__be32 new_addr, __be16 new_port)
{
	csum_replace4(&iph->check, iph->saddr, new_addr);
	csum_replace4(&th->check, iph->saddr, new_addr);
	iph->saddr = new_addr;

	if (skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace2(&th->check, th->source, new_port);
	th->source = new_port;
}

/*
  Modify the destination address and the port number then adjust checksum
*/
static inline void modify_destination(const struct sk_buff *skb,
				struct iphdr *iph, struct tcphdr *th,
				__be32 new_addr, __be16 new_port)
{
	csum_replace4(&iph->check, iph->daddr, new_addr);
	csum_replace4(&th->check, iph->daddr, new_addr);
	iph->daddr = new_addr;

	if (skb->ip_summed != CHECKSUM_PARTIAL)
		csum_replace2(&th->check, th->dest, new_port);
	th->dest   = new_port;
}

/*
  Packet mangling code comes from tcp_dnat_handler in IPVS
*/
static inline int prepare_mangle_tcphdr(struct sk_buff *skb,
					struct iphdr **iphp,
					struct tcphdr **thp)
{
	size_t iphlen = ip_hdrlen(skb);

	/* Here, we can assume that at least the IP header
	   (excluding IP options) is continuous */

	if (!skb_make_writable(skb, iphlen + sizeof(struct tcphdr))) {
		/* skb_make_writable returns 0 on failure */
		if (net_ratelimit())
			printk(KERN_ERR "dts: %s: skb_make_writable failed\n",
				__func__);
		return -EINVAL;
	}

	*iphp = ip_hdr(skb);

	/* Here skb_transport_header may not be available, so
	   do not use tcp_hdr(skb) */
	*thp  = (void *) skb_network_header(skb) + ip_hdrlen(skb);

	return 0;
}

/*
  Calculate the data length of the TCP segment
*/
static inline int tcp_datalen(const struct iphdr *iph,
			const struct tcphdr *th)
{
	return ntohs(iph->tot_len) - iph->ihl*4 - th->doff*4;
}

static inline __be32 ack_seq_for_reply(const struct iphdr *iph,
				const struct tcphdr *th)
{
	return be32_add(th->seq,
			tcp_datalen(iph, th) + th->fin + th->syn);
}

/*
 * dts_input.c
 */
struct dts_cb *dts_cb_new(struct ftcs_kconfig *conf,
			enum dts_state state,
			__be32 raddr, __be16 rport);
unsigned int ftcs_preroute_filter(
	unsigned int hooknum, struct sk_buff *skb,
	const struct net_device *in, const struct net_device *out,
	int (*okfn)(struct sk_buff *));

extern struct net_protocol ftcs_icc_protocol;

/*
 * dts_output.c
 */
unsigned int ftcs_out_filter(
	unsigned int hooknum, struct sk_buff *skb,
	const struct net_device *in, const struct net_device *out,
	int (*okfn)(struct sk_buff *));
int dts_enqueue_reply_fin_ack(struct dts_cb *dcb,
			__be32 raddr, __be16 rport,
			__be32 laddr, __be16 lport,
			__be32 seq, __be32 ack, __be16 window);
int dts_enqueue_reply_rst(struct dts_cb *dcb,
			__be32 raddr, __be16 rport,
			__be32 laddr, __be16 lport,
			__be32 seq, __be32 ack);
int dts_enqueue_reply_ack(struct dts_cb *dcb,
			__be32 daddr, __be16 dport,
			__be32 saddr, __be16 sport,
			__be32 seq, __be32 ack,
			__be16 window);
int ftcs_icc_xmit(const struct ftcs_kconfig *conf, __be32 dest_addr,
		u8 icc_type, const void *payload, size_t payload_len);
int ftcs_icc_bcast(const struct ftcs_kconfig *conf,
		u8 icc_type, const void *payload, size_t payload_len);
int ftcs_notify_dest_switched(struct dts_cb *dcb, __be32 newdest);
int dts_forward_skb(struct sk_buff *orig_skb);

/*
 * dts_handover.c
 */

int dts_request_handover(struct socket *sock,
			const struct sockaddr_in *dest_sin,
			const void __user *req_buf, size_t req_len,
			size_t next_req_offset);
int dts_sys_request_handover(void __user *uptr, socklen_t len);

/*
 * dts_main.c
 */

int dts_start_proxy(struct ftcs_kconfig *conf);


/*
 * dts_logging.c
 */
void dts_log(const struct dts_cb *fcb,
	const struct sk_buff *skb,
	uint8_t flags);
int __init dts_log_setup(void);
void dts_log_cleanup(void);

/*
 * dts_conntrack.c
 */
int dts_ct_egress(struct dts_endpoint *ep,
		const struct iphdr *iph, const struct tcphdr *th);
int dts_ct_verify_ingress(struct dts_endpoint *ep,
			const struct sk_buff *skb);
int dts_ct_ingress(struct dts_endpoint *ep,
		const struct iphdr *iph, const struct tcphdr *th);
int dts_ct_terminated(struct dts_endpoint *ep);
void dts_ct_retrans_fin(struct dts_endpoint *ep);
void dts_ct_enable_emulation(struct dts_endpoint *ep);
int dts_ct_emulation_enabled(struct dts_endpoint *ep);

static inline void dts_ct_init(struct dts_endpoint *ep)
{
	struct dts_conntrack *ct = &ep->ct;
	ct->flags = 0;
	ct->ingress_state = ct->egress_state = DTS_HST_OPEN;
	ct->max_timeout = 0;
}

static inline void dts_ct_terminate(struct dts_endpoint *ep)
{
	struct dts_conntrack *ct = &ep->ct;

	ct->ingress_state = ct->egress_state = DTS_HST_CLOSED;
}

static inline
void dts_ct_set_max_timeout(struct dts_endpoint *ep, uint32_t timeo)
{
	struct dts_conntrack *ct = &ep->ct;

	ct->max_timeout = timeo;

	if (timeo > 0)
		dts_endpoint_set_timeout(ep, timeo);
	else
		dts_endpoint_clear_timeout(ep);
}


/*
 * consistent hashing
 */

#include <linux/crypto.h>
#include <linux/scatterlist.h>

extern struct kmem_cache *ftcs_node_cachep;
extern struct crypto_hash *ftcs_tfm;
extern spinlock_t ftcs_hash_lock;

#define FTCS_CHASH_LEN 16

struct ftcs_kvnode {
	struct rb_node rbnode;
	u8 hash[FTCS_CHASH_LEN];
	__be32 addr;
	int id;
};

struct ftcs_kpnode {
	struct rb_node rbnode;
	__be32 addr;
	uint64_t idset;
};

static inline
int ftcs_chash_compare(const u8 l[FTCS_CHASH_LEN], const u8 r[FTCS_CHASH_LEN])
{
	int i;

	for (i = 0; i < FTCS_CHASH_LEN; i++) {
		if (l[i] > r[i])
			return 1;
		else if (l[i] < r[i])
			return -1;
	}

	return 0;
}

#define FTCS_CHASH_MAX ((ftcs_chash_t)(-1))

static inline
void ftcs_calc_chash(const void *p, size_t len, u8 hash[FTCS_CHASH_LEN])
{
	struct hash_desc desc;
	struct scatterlist sg;

	desc.tfm = ftcs_tfm;
	desc.flags = 0;

	sg_init_one(&sg, p, len);

	spin_lock_bh(&ftcs_hash_lock);

	if (crypto_hash_digest(&desc, &sg, len, hash))
		BUG();

	spin_unlock_bh(&ftcs_hash_lock);
}

/*
  buff must have at least (FTCS_CHASH_LEN*2 + 1) bytes
 */
static inline
char *ftcs_chash_to_str(u8 hash[FTCS_CHASH_LEN], char *buff)
{
	int i;
	for (i = 0; i < FTCS_CHASH_LEN; i++)
		sprintf(buff+i*2, "%02x", hash[i]);

	return buff;
}

int ftcs_add_vnodes(struct ftcs_kconfig *conf,
		struct ftcs_vnode *vnodes, size_t nr_vnodes);
void ftcs_remove_node(struct ftcs_kconfig *conf, __be32 addr, int id);
__be32 ftcs_responsible_node(struct ftcs_kconfig *conf,
			__be32 raddr, __be16 rport);
__be32 ftcs_prev_responsible_node(struct ftcs_kconfig *conf,
				__be32 raddr, __be16 rport);
__be32 ftcs_next_responsible_node(struct ftcs_kconfig *conf,
				__be32 raddr, __be16 rport);
int ftcs_responsible_for_packet(struct ftcs_kconfig *conf,
				__be32 addr, __be16 port);
__be32 ftcs_get_responsible_nodes(struct ftcs_kconfig *conf,
				__be32 raddr, __be16 rport, __be32 *pprev_node);
int ftcs_mourning(struct ftcs_kconfig *conf);
int ftcs_neighbor(struct ftcs_kconfig *conf, __be32 addr);

/*
  dts_l4.c
*/
unsigned int ftcs_l4_preroute_filter(
	unsigned int hooknum, struct sk_buff *skb,
	const struct net_device *in, const struct net_device *out,
	int (*okfn)(struct sk_buff *));
extern int dts_l4_only_mode;


static inline int dts_nohandover_mode(const struct ftcs_kconfig *conf)
{
	return conf->port == conf->proxy_port;
}

#ifdef CONFIG_FTCS_PROFILING
/*
  Profiling
 */
enum {
	DTS_PROF_OUT_BODY,
	DTS_PROF_OUT_CT,
	DTS_PROF_IN_BODY,
	DTS_PROF_IN_SYN,
	DTS_PROF_IN_HSFO, /* handle_syn_for_other */
	DTS_PROF_IN_SYN_NHO, /* handle_inter_syn_nohandover */
	DTS_PROF_TOTAL,
};

struct dts_prof_entry {
	const char *name;
	atomic_t sum;
	atomic_t nr_samples;
};

extern struct dts_prof_entry dts_prof_entries[DTS_PROF_TOTAL];

unsigned long
dts_prof_tv_diff(const struct timeval *tv1, const struct timeval *tv0);

#endif /* #ifdef CONFIG_FTCS_PROFILING */


#endif /* #ifndef DTS_INTERNAL_H */
