#include <linux/module.h>
#include <linux/types.h>

#include <linux/skbuff.h>

#include <linux/ip.h>
#include <linux/tcp.h>

#include <net/ip.h>
#include <net/route.h>
#include <net/tcp.h>
#include <net/protocol.h>

#include <linux/netfilter.h>

#include "ftcs.h"
#include "dts_internal.h"

int dts_l4_only_mode = 0;
module_param(dts_l4_only_mode, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(dts_l4_only_mode, "Layer-4 load distribution only mode");

static unsigned int handle_syn(struct ftcs_kconfig *conf,
			const struct iphdr *iph, const struct tcphdr *th)
{
	int ret = NF_ACCEPT;

	if (!ftcs_responsible_for_packet(conf, iph->saddr, th->source)) {
		dprintf(2, "dropping " NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
			" : not responsible\n",
			NIPQUAD(iph->saddr), ntohs(th->source),
			NIPQUAD(iph->daddr), ntohs(th->dest));
		ret = NF_DROP;
	}

	dprintf(2, "accepting " NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
		" : responsible\n",
		NIPQUAD(iph->saddr), ntohs(th->source),
		NIPQUAD(iph->daddr), ntohs(th->dest));

	return ret;
}

static unsigned int handle_nonsyn(const struct iphdr *iph,
				const struct tcphdr *th,
				const struct net_device *in)
{
	struct sock* sk;
	struct request_sock **prev;
	unsigned int verdict;

	sk = __inet_lookup(&init_net,
			&tcp_hashinfo, iph->saddr, th->source,
			iph->daddr, th->dest, in->ifindex);

	if (sk == NULL) {
		dprintf(2, "dropping " NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
			" : irrelevant\n",
			NIPQUAD(iph->saddr), ntohs(th->source),
			NIPQUAD(iph->daddr), ntohs(th->dest));
		return NF_DROP;
	}

	if (sk->sk_state == TCP_LISTEN
		&& !inet_csk_search_req(sk, &prev,
					th->source,
					iph->saddr, iph->daddr)) {
		dprintf(2, "dropping " NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
			" : irrelevant\n",
			NIPQUAD(iph->saddr), ntohs(th->source),
			NIPQUAD(iph->daddr), ntohs(th->dest));
		verdict = NF_DROP;
	} else if (sk->sk_state == TCP_TIME_WAIT) {
		struct inet_timewait_sock *tw = inet_twsk(sk);
		if (tw->tw_substate != TCP_FIN_WAIT2) {
			dprintf(2, "dropping "
				NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
				" : TIME-WAIT socket\n",
				NIPQUAD(iph->saddr), ntohs(th->source),
				NIPQUAD(iph->daddr), ntohs(th->dest));
			verdict = NF_DROP;
		} else {
			dprintf(2, "accepting "
				NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
				" : FIN_WAIT_2 socket\n",
				NIPQUAD(iph->saddr), ntohs(th->source),
				NIPQUAD(iph->daddr), ntohs(th->dest));
			verdict = NF_ACCEPT;
		}

		inet_twsk_put(tw);
		return verdict;
	} else {
		dprintf(2, "accepting " NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
			" : relevant\n",
			NIPQUAD(iph->saddr), ntohs(th->source),
			NIPQUAD(iph->daddr), ntohs(th->dest));
		verdict = NF_ACCEPT;
	}

	sock_put(sk);

	return verdict;
}

unsigned int ftcs_l4_preroute_filter(
	unsigned int hooknum, struct sk_buff *skb,
	const struct net_device *in, const struct net_device *out,
	int (*okfn)(struct sk_buff *))
{
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *th, _tcphdr;
	struct ftcs_kconfig *conf;
	unsigned int ret = NF_ACCEPT;

	if (iph->protocol != IPPROTO_TCP)
		goto out;

	if (unlikely(iph->frag_off & __constant_htons(IP_OFFSET))) {
		/* Pass fragments */
		dprintf(0, "passing fragment packets\n");
		goto out;
	}

	th = skb_header_pointer(skb, iph->ihl*4, sizeof(_tcphdr), &_tcphdr);
	if (unlikely(th == NULL)) {
		dprintf(0, "th == NULL\n");
		goto out;
	}

	conf = ftcs_find_get_config(iph->daddr, th->dest);

	if (conf == NULL) {
		dprintf(3, "accepting " NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
			" : config not found\n",
			NIPQUAD(iph->saddr), ntohs(th->source),
			NIPQUAD(iph->daddr), ntohs(th->dest));
		goto out;
	}

	if (iph->daddr != conf->inter_addr) {
		dprintf(2, "accepting " NIPQUAD_FMT ":%d->" NIPQUAD_FMT ":%d"
			" : iph->daddr != conf->inter_addr\n",
			NIPQUAD(iph->saddr), ntohs(th->source),
			NIPQUAD(iph->daddr), ntohs(th->dest));

		goto out_release_config;
	}

	if (th->syn && !th->ack)
		ret = handle_syn(conf, iph, th);
	else
		ret = handle_nonsyn(iph, th, in);

out_release_config:
	ftcs_release_config(conf);
out:
	return ret;
}
