/**
   DTS Packet Logging

   Almost a copy from exported-global-mod.c, distiributed with
   relay-apps.tar.

   Copyright (C) IBM Corporation, 2005
   2005		Tom Zanussi <zanussi@us.ibm.com>

   Modified by:
   2009 Hajime Fujita <hfujita@cc.u-tokyo.ac.jp>
*/

#include <linux/module.h>
#include <linux/relay.h>
#include <linux/debugfs.h>
#include <linux/skbuff.h>
#include <asm/atomic.h>

#include "ftcs.h"
#include "dts_internal.h"

static struct rchan  *rchan;
static struct dentry *debugfs_dir;
static size_t subbuf_size = 512*1024;
static size_t n_subbufs   = 4;
static DEFINE_SPINLOCK(rchan_lock);
static atomic_t id = ATOMIC_INIT(0);

/*
 * file_create() callback.  Creates relay file in debugfs.
 */
static struct dentry *create_buf_file_handler(const char *filename,
					struct dentry *parent,
					int mode,
					struct rchan_buf *buf,
					int *is_global)
{
	struct dentry *buf_file;

	buf_file = debugfs_create_file(filename, mode, parent, buf,
				&relay_file_operations);
	*is_global = 1;

	return buf_file;
}

/*
 * file_remove() default callback.  Removes relay file in debugfs.
 */
static int remove_buf_file_handler(struct dentry *dentry)
{
	debugfs_remove(dentry);

	return 0;
}

/*
 * relayfs callbacks
 */
static struct rchan_callbacks relayfs_callbacks = {
	.create_buf_file = create_buf_file_handler,
	.remove_buf_file = remove_buf_file_handler,
};

void dts_log(const struct dts_cb *dcb,
	const struct sk_buff *skb,
	uint8_t flags)
{
	struct dts_packet_log log;
	unsigned int copy_len = min(DTS_LOG_PACKET_LEN,
				skb->len - skb_network_offset(skb));

	do_gettimeofday(&log.timestamp);
	log.id = atomic_add_return(1, &id);
	log.flags = flags;
	if (dcb != NULL) {
		log.filter_state = dcb->state;
		log.flags |= DTS_LOG_STATE;
	}
	skb_copy_bits(skb, skb_network_offset(skb),
		log.packet_hdr, copy_len);

	spin_lock_bh(&rchan_lock);
	relay_write(rchan, &log, sizeof(log));
	spin_unlock_bh(&rchan_lock);
}

int __init dts_log_setup(void)
{
	debugfs_dir = debugfs_create_dir("dts", NULL);
	if (debugfs_dir == NULL) {
		dprintk(KERN_ERR "%s: debugfs_create_dir failed\n", __func__);
		return -ENOMEM;
	}

	rchan = relay_open("log", debugfs_dir, subbuf_size,
			n_subbufs, &relayfs_callbacks, NULL);
	if (rchan == NULL) {
		dprintk(KERN_ERR "%s: relay_open failed\n", __func__);
		debugfs_remove(debugfs_dir);
		return -ENOMEM;
	}

	return 0;
}

void dts_log_cleanup(void)
{
	relay_close(rchan);
	debugfs_remove(debugfs_dir);
}
