#include <linux/module.h>
#include <linux/version.h>

#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/ip.h>
#include <net/tcp.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>

#include <linux/cryptohash.h>
#include <asm/unaligned.h>

#include "ftcs.h"
#include "dts_internal.h"

extern int tcp_extract_timestamp(struct sk_buff *skb, u32 *tsval, u32 *tsecr);

static void
tcp_calc_checksum(struct sk_buff *skb, struct iphdr *iph, struct tcphdr *th)
{
	size_t ipdatalen = ntohs(iph->tot_len) - iph->ihl*4;
	th->check = 0;
	th->check = csum_fold(
		csum_tcpudp_nofold(iph->saddr, iph->daddr,
				ipdatalen, IPPROTO_TCP,
				skb_checksum(skb,
					iph->ihl * 4,
					ipdatalen,
					0)));
}

int tcp_modify_timestamp(struct sk_buff *skb, s64 tsval_offset, s64 tsecr_offset)
{
	struct iphdr *iph = ip_hdr(skb);
	struct tcphdr *th = (void *) iph + iph->ihl*4;
	int length = th->doff*4 - sizeof (struct tcphdr);
	unsigned char *ptr = (unsigned char *) th + sizeof(struct tcphdr);
	u32 tsval, tsecr;
	s64 sum;
	int is_updated = 0;

	while (length > 0) {
		int opcode = *ptr++;
		int opsize;

		switch (opcode) {
		case TCPOPT_NOP:
			length--;
			continue;

		case TCPOPT_EOL:
			return EINVAL;

		default:
			opsize = *ptr++;
			if (opsize < 2) /* "silly options" */
				return EINVAL;
			if (opsize > length) /* don't parse partial options */
				return EINVAL;

			switch (opcode) {
			case TCPOPT_TIMESTAMP:
				if (opsize == TCPOLEN_TIMESTAMP) {
					if (tsval_offset == 0 && tsecr_offset == 0)
						return 0;
					if (tsval_offset != 0) {
						tsval = get_unaligned_be32(ptr);
						sum = (s64) tsval + tsval_offset;
						put_unaligned_be32((u32)sum, ptr);
					}
					if (tsecr_offset != 0) {
						tsecr = get_unaligned_be32(ptr + 4);
						sum = (s64) tsecr + tsecr_offset;
						put_unaligned_be32((u32)sum, ptr + 4);
					}
					is_updated = 1;
					goto exit;
				} else {
					return EINVAL;
				}
				break;

			default:
				break;
			}

			ptr += opsize - 2;
			length -= opsize;
		}
	}

exit:
	if (is_updated
		&& skb->ip_summed != CHECKSUM_PARTIAL) {
		tcp_calc_checksum(skb, iph, th);
	}
	return 0;
}

static inline struct dst_entry *rt_dst(struct rtable *rt)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
	return &rt->dst;
#else
	return &rt->u.dst;
#endif
}

struct dts_delayed_xmit {
	/* This must be on top of the structure */
	struct dts_delayed_work work;

	struct sk_buff *skb;
};

static struct rtable *get_out_rt(__be32 addr, u32 rtos)
{
	struct rtable *rt;

	/* TODO: caching */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,38)
	struct flowi4 fl;
	flowi4_init_output(&fl,
					   0,
					   0,
					   rtos,
					   0,
					   0,
					   0,
					   addr,
					   0,
					   0,
					   0);
#else
	struct flowi fl = {
		.oif = 0,
		.nl_u = {
			.ip4_u = {
				.daddr = addr,
				.saddr = 0,
				.tos = rtos,
			}
		},
	};
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,38)
	rt = ip_route_output_key(&init_net, &fl);
	if (IS_ERR(rt)) {
#else
	if (ip_route_output_key(&init_net, &rt, &fl)) {
#endif
		dprintk(KERN_ERR "get_out_rt: ip_route_output error,"
			"dest: %u.%u.%u.%u\n",
			NIPQUAD(addr));
		return NULL;
	}

	return rt;
}

static void delayed_xmit(void *data)
{
	struct dts_delayed_xmit *work = data;

	ip_local_out(work->skb);
	kfree(work);
}

static int prepare_ippacket(u8 l4proto, size_t l4len,
			__be32 raddr, __be32 laddr,
			struct sk_buff **skb_p, char **l4hdr_p)
{
	struct sk_buff *skb;
	struct rtable *rt;
	struct iphdr *iph;
	int headroom;
	char *l4hdr;

	*skb_p = NULL;

	rt = get_out_rt(raddr, 0);
	if (rt == NULL)
		return -EHOSTUNREACH;

	headroom = LL_RESERVED_SPACE(rt_dst(rt)->dev)
		+ sizeof(struct iphdr) + l4len;
	skb = alloc_skb(headroom, GFP_ATOMIC|__GFP_ZERO);
	if (!skb)
		return -ENOMEM;

	skb_reserve(skb, headroom);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
	dst_release(skb->dst);
	skb->dst = rt_dst(rt);
#else
	skb_dst_drop(skb);
	skb_dst_set(skb, rt_dst(rt));
#endif

        l4hdr = skb_push(skb, l4len);
	if (l4hdr_p)
		*l4hdr_p = l4hdr;

	skb_reset_transport_header(skb);

	iph = (struct iphdr *) skb_push(skb, sizeof(*iph));
	memset(iph, 0, sizeof(*iph));
	iph->version  = 4;
	iph->ihl      = (sizeof(struct iphdr)>>2);
	iph->frag_off = htons(IP_DF);
	iph->protocol = l4proto;
	iph->tos      = 0;
	iph->daddr    = raddr;
	iph->saddr    = laddr;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
	iph->ttl      = dst_metric(skb->dst, RTAX_HOPLIMIT);
#else
	iph->ttl      = dst_metric(skb_dst(skb), RTAX_HOPLIMIT);
#endif
	iph->tot_len  = htons(skb->len);

	skb_reset_network_header(skb);

	ip_select_ident(iph, rt_dst(rt), NULL);
	ip_send_check(iph);

	skb->ip_summed = CHECKSUM_NONE;

	*skb_p = skb;

	return 0;
}

static inline __be32 bcast_addr(__be32 addr, __be32 netmask)
{
	uint32_t mask_h = ntohl(netmask);
	uint32_t mask_inv_h = ~mask_h;
	uint32_t addr_h = ntohl(addr);

	return htonl(addr_h | mask_inv_h);
}

static int icc_xmit(const struct ftcs_kconfig *conf, __be32 raddr,
		u8 ipproto,
		u8 icc_type, const void *payload, size_t payload_len)
{
	struct sk_buff *skb;
	char *l4p;
	struct ftcs_icchdr *icchdr;
	int err;

	err = prepare_ippacket(ipproto, payload_len + sizeof(*icchdr),
			raddr, conf->intra_addr, &skb, &l4p);
	if (err)
		return err;

	icchdr = (struct ftcs_icchdr *) l4p;
	icchdr->type   = icc_type;
	icchdr->length = htons((u16) payload_len);

	if (payload != NULL)
		memcpy(l4p+sizeof(*icchdr), payload, payload_len);

	return ip_local_out(skb);
}

int ftcs_icc_xmit(const struct ftcs_kconfig *conf, __be32 raddr,
		u8 icc_type, const void *payload, size_t payload_len)
{
	return icc_xmit(conf, raddr, IPPROTO_FTCSICC,
			icc_type, payload, payload_len);
}

int ftcs_icc_bcast(const struct ftcs_kconfig *conf,
		u8 icc_type, const void *payload, size_t payload_len)
{
	__be32 raddr = bcast_addr(conf->intra_addr, conf->intra_mask);
	return ftcs_icc_xmit(conf, raddr,
			icc_type, payload, payload_len);
}

int ftcs_notify_dest_switched(struct dts_cb *dcb, __be32 newdest)
{
	struct ftcs_icc_destsw d;
	__be32 raddr;

	raddr = bcast_addr(dcb->conf->intra_addr, dcb->conf->intra_mask);

	d.raddr = dcb->raddr;
	d.rport = dcb->rport;
	d.laddr = dcb->conf->inter_addr;
	d.lport = dcb->conf->port;
	d.new_dest = newdest;

	return icc_xmit(dcb->conf, raddr, IPPROTO_FTCSICC2,
			FTCS_ICC_DEST_SWITCH, &d, sizeof(d));
}

/*
  Transmit the given skb

  skb->data is assumed to point the head of IP header

  the argument is freed when the function succeeds

  TODO: re-implement to avoid unnecessary copy
  (see ip_vs_dr_xmit as a goal)
*/
int dts_forward_skb(struct sk_buff *orig_skb)
{
	int err;
	struct sk_buff *skb;
	struct iphdr *iph;
	int iphlen;
	int l4len;
	char *l4p;

	iph    = ip_hdr(orig_skb);
	iphlen = iph->ihl*4;
	l4len  = ntohs(iph->tot_len) - iphlen;

	err = prepare_ippacket(iph->protocol, l4len, iph->daddr, iph->saddr,
			&skb, &l4p);
	if (err) {
		dprintf(1, "prepare_ippacket failed: %d\n", err);
		return err;
	}

	err = skb_copy_bits(orig_skb, iphlen, l4p, l4len);
	if (err) {
		dprintf(1, "skb_copy_bits failed: %d\n", err);
		kfree_skb(skb);
		return err;
	}

	err = ip_local_out(skb);
	if (err){
		dprintf(1, "ip_local_out failed: %d\n", err);
		kfree_skb(skb);
		return err;
	}

	kfree_skb(orig_skb);

	return err;
}

/*
  Nearly equivalent to ip_vs_tunnel_xmit

  TODO: see send_reset in net/ipv4/netfilter/ipt_REJECT.c
*/
static int do_enqueue_reply(struct dts_cb *dcb,
			__be32 raddr, __be16 rport,
			__be32 laddr, __be16 lport,
			__be32 seq, __be32 ack,
			__be16 window,
			int fin, int rst)
{
	struct sk_buff *skb;
	struct tcphdr *th;
	struct dts_delayed_xmit *xmit_work;
	int ret;

	ret = prepare_ippacket(IPPROTO_TCP, sizeof(*th),
			raddr, laddr, &skb, (char **) &th);
	if (ret)
		return ret;

	xmit_work = kmalloc(sizeof(*xmit_work), GFP_ATOMIC);
	if (xmit_work == NULL) {
		kfree_skb(skb);
		return -ENOMEM;
	}

	th->source  = lport;
	th->dest    = rport;
	th->seq	    = seq;
	th->ack_seq = ack;
	th->window  = window;
	th->doff    = sizeof(struct tcphdr)/4;
	th->fin	    = fin ? 1 : 0;
	th->rst     = rst ? 1 : 0;
	th->ack	    = 1;

	th->check = tcp_v4_check(th->doff << 2, laddr, raddr,
				csum_partial((char *)th, th->doff << 2, 0));

	xmit_work->skb = skb;
	dts_delayed_work_init(&xmit_work->work, delayed_xmit, xmit_work);
	dts_cb_register_delayed_work(dcb, &xmit_work->work);

	return 0;
}

int dts_enqueue_reply_ack(struct dts_cb *dcb,
			__be32 daddr, __be16 dport,
			__be32 saddr, __be16 sport,
			__be32 seq, __be32 ack,
			__be16 window)
{
	return do_enqueue_reply(dcb,
				daddr, dport, saddr, sport,
				seq, ack, window, 0, 0);
}

int dts_enqueue_reply_fin_ack(struct dts_cb *dcb,
			__be32 daddr, __be16 dport,
			__be32 saddr, __be16 sport,
			__be32 seq, __be32 ack, __be16 window)
{
	return do_enqueue_reply(dcb,
				daddr, dport, saddr, sport,
				seq, ack, window, 1, 0);
}

int dts_enqueue_reply_rst(struct dts_cb *dcb,
			__be32 daddr, __be16 dport,
			__be32 saddr, __be16 sport,
			__be32 seq, __be32 ack)
{
	return do_enqueue_reply(dcb,
				daddr, dport, saddr, sport,
				seq, ack, (__be16) 0, 0, 1);
}

/*
  Insert the connection pushing option data before other options.
*/
static int insert_conn_opt(struct sk_buff *skb,
			const struct dts_conn_opt *opt)
{
	int err;
	unsigned char *to, *from;
	struct iphdr *iph;
	struct tcphdr *th;
	size_t iphdrlen, ipdatalen, thoff, tot_len;
	__be16 orig_tot_len;
	struct dts_conn_tcp_opt tcpopt = {
		.type = DTS_TCPOPT_CONN_PUSH,
		.len  = DTS_TCPOLEN_CONN_PUSH,
		.size = sizeof(*opt),
		.pad = TCPOPT_NOP
	};

	/*
	 *
	 *  Original buffer
	 *
	 *                     +---------+-----------+
	 *                     | header  | options   |
	 *                     +---------+-----------+
	 *  going to be
	 *
	 *   +---------+---+----------+--------------+
	 *   | header  |   | options  | dts_conn_opt |
	 *   +---------+---+----------+--------------+
	 *               |
	 *               +-- TCP option for handover
	 *
	 * This header is IP and TCP header,
	 * and options mean TCP options.
	 * This function insert TCP option for dts handover before
	 * original TCP options, and insert struct dts_conn_opt to
	 * begining of TCP payload.
	 */

	err = skb_cow(skb, sizeof(tcpopt) + sizeof(*opt));
	if (err) {
		dprintk(KERN_INFO "%s: skb_cow failed\n", __func__);
		return err;
	}

	iph = ip_hdr(skb);
	tot_len = ntohs(iph->tot_len);
	th  = tcp_hdr(skb);

	iphdrlen = ip_hdrlen(skb) + sizeof(struct tcphdr);

	from = skb_network_header(skb);
	thoff = skb_transport_header(skb) - skb->data;

	skb_push(skb, (sizeof(tcpopt) + sizeof(*opt)));
	to = from - (sizeof(tcpopt) + sizeof(*opt));

	memmove(to, from, iphdrlen);
	memmove(to+iphdrlen+sizeof(tcpopt), from+iphdrlen, tot_len - iphdrlen);
	memcpy(to+iphdrlen, &tcpopt, sizeof(tcpopt));

	skb_set_network_header(skb, to - skb->data);
	skb_set_transport_header(skb, thoff);

	iph = ip_hdr(skb);
	th  = tcp_hdr(skb);

	/* Adjust length */
	orig_tot_len = iph->tot_len;
	iph->tot_len = htons(ntohs(iph->tot_len) + sizeof(tcpopt) + sizeof(*opt));
	th->doff += (sizeof(tcpopt) >> 2);

	memcpy((unsigned char *) th + th->doff * 4, opt, sizeof(*opt));

	ipdatalen = ntohs(iph->tot_len) - ip_hdrlen(skb);

	/* Adjust checksum */
	csum_replace2(&iph->check, orig_tot_len, iph->tot_len);

	if (skb->ip_summed == CHECKSUM_PARTIAL) {
		__be16 oldlen, newlen;

		oldlen = htons(ipdatalen - (sizeof(tcpopt) + sizeof(*opt)));
		newlen = htons(ipdatalen);

		skb->csum_start = skb_transport_header(skb) - skb->head;
		/* Initial checksum value for hardware checksumming
		   should be negative value */
		csum_replace2(&th->check, ~oldlen, ~newlen);
	} else {
		th->check = 0;
		th->check = csum_fold(
			csum_tcpudp_nofold(iph->saddr, iph->daddr,
					ipdatalen, IPPROTO_TCP,
					skb_checksum(skb,
						ip_hdrlen(skb),
						ipdatalen,
						0)));
	}

	return 0;
}

static int merge_tcp_opt(struct sk_buff *skb, struct tcp_options_received *opts)
{
	struct iphdr* iph = ip_hdr(skb);
	struct tcphdr* th = tcp_hdr(skb);
	size_t thlen = th->doff * 4;
	size_t optlen = thlen - sizeof (struct tcphdr);
	unsigned char *ptr = (unsigned char *)th + sizeof (struct tcphdr);
	int i = 0;
	int len;

	while (i < optlen) {
		switch (ptr[i]) {
		case TCPOPT_NOP:
			i++;
			break;

		case TCPOPT_MSS:
			len = ptr[i+1];
			if (len == TCPOLEN_MSS) {
				u16 mss_clamp;
				put_unaligned_be16(opts->mss_clamp, &mss_clamp);
				memcpy(&ptr[i+2], &mss_clamp, sizeof(u16));

			}
			i += len;
			break;

		case TCPOPT_WINDOW:
			len = ptr[i+1];
			if (len == TCPOLEN_WINDOW) {
				if (opts->wscale_ok)
					ptr[i + 2] = opts->snd_wscale;
				else
					memset(&ptr[i], 1, len);
			}
			i += len;
			break;

		case TCPOPT_SACK_PERM:
			len = ptr[i+1];
			if (len == TCPOLEN_SACK_PERM) {
				if (!opts->sack_ok)
					memset(&ptr[i], 1, len);
			}
			i += len;
			break;

		case TCPOPT_TIMESTAMP:
			len = ptr[i + 1];
			if (len == TCPOLEN_TIMESTAMP) {
				if (!opts->saw_tstamp)
					memset(&ptr[i], 1, len);
			}
			i += len;
			break;

		case TCPOPT_EOL:
			goto exit;

		default:
			i += ptr[i + 1];
			break;
		}
	}

exit:
	if (skb->ip_summed == CHECKSUM_PARTIAL)
		dts_calc_checksum_partial(skb, iph, th);
	return 0;
}

/*
  dcb: locked
*/
static int handle_conn_push_syn(struct dts_cb *dcb,
				struct sk_buff *skb)
{
	struct dts_conn_opt opt;
	struct tcphdr *th;
	__be32 new_seq;
	int err;

	opt.real_raddr = dcb->raddr;
	opt.real_rport = dcb->rport;
	opt.initial_egress_seq = dcb->state_priv.push.remote_syn_seq;
	opt.remote_window = htonl(dcb->state_priv.push.remote_window);
	opt.handover_len  = htonl(dcb->state_priv.push.handover_len);
	opt.remote_last_timestamp = dcb->remote_last_timestamp;

	merge_tcp_opt(skb, &dcb->tcp_options);

	err = insert_conn_opt(skb, &opt);
	if (err) {
		th = tcp_hdr(skb);
		if (net_ratelimit())
			printk(KERN_ERR "dts: insert_conn_opt failed: "
				"%d (lport:%d)\n",
				err, ntohs(th->source));
		return err;
	}

	/* adjust sequence number */
	th = tcp_hdr(skb);

	new_seq = dcb->state_priv.push.local_syn_seq;
	dcb->push_seq_offset = ntohl(th->seq) - ntohl(new_seq);

	csum_replace4(&th->check, th->seq, new_seq);
	th->seq = new_seq;

	return 0;
}

static enum dts_filter_verdict
process_intra_output(struct dts_endpoint *ep, struct sk_buff *skb)
{
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;
	struct dts_cb *dcb = ep->dcb;
	struct iphdr *iph;
	struct tcphdr *th;

	if (prepare_mangle_tcphdr(skb, &iph, &th))
		goto out;

	switch (dcb->state) {
	case DTS_PUSH_SYN_WAIT:
		if (!th->syn || th->ack)
			goto out;
		dcb->state = DTS_CONNECTING;
		/* Fall through */
	case DTS_CONNECTING:
		if (th->syn && !th->ack && handle_conn_push_syn(dcb, skb))
			goto out;
		else
			adjust_egress_seq(skb, th,
					dcb->push_seq_offset);

		modify_window_size(skb, th, 0);
		verdict = DTS_ACCEPT_CT;

		break;

	case DTS_PUSHING:
		adjust_egress_seq(skb, th, dcb->push_seq_offset);

		/* TODO: the exact window size must be informed */
		if (!th->fin)
			modify_window_size(skb, th, 0);

		verdict = DTS_ACCEPT_CT;

		break;

	case DTS_FORWARDING:
		adjust_egress_seq(skb, th, dcb->pull_seq_offset);
		verdict = DTS_ACCEPT_CT;
		break;

	default:
		break;
	}

out:
	return verdict;
}

static enum dts_filter_verdict
inter_output_pulling_common(struct dts_cb *dcb, struct sk_buff *skb)
{
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;
	struct iphdr *iph;
	struct tcphdr *th;

	iph = ip_hdr(skb);
	th  = tcp_hdr(skb);

	modify_source(skb, iph, th,
		dcb->conf->intra_addr, th->source);
	modify_destination(skb, iph, th,
			dcb->state_priv.pull.peer_addr,
			dcb->state_priv.pull.peer_port);

	/* ACK should be inspected before adjusting */
	if (after(ntohl(th->ack_seq), dcb->state_priv.pull.last_tx_ack)) {
		uint32_t last = dcb->state_priv.pull.last_tx_ack;
		uint32_t ack = ntohl(th->ack_seq);

		dcb->state_priv.pull.rx_bytes += (ack - last);
		dcb->state_priv.pull.last_tx_ack = ack;
	}

	adjust_egress_seq(skb, tcp_hdr(skb), dcb->pull_seq_offset);

	/*
	  Rerouting after destination mangling:
	  See ipt_local_hook (net/ipv4/netfilter/iptable_mangle.c)

	  ip_route_me_harder(skb, RTN_UNSPEC) does the
	  exact thing we want to do.
	*/
	if (ip_route_me_harder(skb, RTN_UNSPEC)) {
		dprintk(KERN_INFO "%s: ip_route_me_harder failed\n",
			__func__);
		goto out;
	}

	verdict = DTS_ACCEPT_CT;

out:
	return verdict;
}

static enum dts_filter_verdict
process_inter_output(struct dts_endpoint *ep, struct sk_buff *skb)
{
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;
	struct iphdr *iph, orig_iph;
	struct tcphdr *th, orig_th;
	struct dts_cb *dcb = ep->dcb;
	enum dts_state oldstate = dcb->state;
	uint32_t tsval, tsecr;

	if (prepare_mangle_tcphdr(skb, &iph, &th))
		goto out;

	switch (dcb->state) {
		uint32_t off;
		struct dts_endpoint *ep_intra;

	case DTS_FIRST_SYN_RECV2:
		if (th->rst)
			verdict = DTS_ACCEPT_CT;
		break;

	case DTS_FORWARDING:
		adjust_egress_seq(skb, tcp_hdr(skb), dcb->pull_seq_offset);
		tcp_modify_timestamp(skb, dcb->state_priv.pull.delta2, dcb->state_priv.pull.delta1);

		verdict = DTS_ACCEPT_CT;

		break;

	case DTS_PULL_SYN_RECV:
		/* Expect AYN+ACK, but it may be RST */
		if ((!th->syn || !th->ack) && !th->rst)
			goto out;

		off = ntohl(th->seq)
			- ntohl(dcb->state_priv.pull.egress_syn_seq);
		dcb->pull_seq_offset = off;
		dcb->state_priv.pull.last_tx_ack = ntohl(th->ack_seq);
		dcb->state_priv.pull.rx_bytes    = 0;
		if (th->syn && th->ack && !th->rst) {
			dcb->state = DTS_PULLING;
			dts_ct_set_max_timeout(ep, 0);
		}
		/* Fall through */
	case DTS_PULLING:
		orig_iph = *iph;
		orig_th  = *th;

		verdict = inter_output_pulling_common(dcb, skb);

		if ((verdict & DTS_ACCEPT_MASK) == 0)
			break;

		ep_intra = dts_find_get_endpoint(dcb,
						iph->saddr, th->source,
						iph->daddr, th->dest);
		if (ep_intra == NULL) {
			dprintf(1, "ep_intra == NULL!\n");
			verdict = DTS_DROP_NOCT;
			break;
		}

		if (oldstate == DTS_PULL_SYN_RECV)
			dts_ct_set_max_timeout(ep_intra, 0);

		if (dcb->state == DTS_PULLING
			&& dcb->state_priv.pull.rx_bytes ==
			dcb->state_priv.pull.handover_len) {
			int err;

			/* Reply ACK to inform that window size
			   has been resumed */
			err = dts_enqueue_reply_ack(
				dcb,
				ep->laddr, ep->lport,
				ep->raddr, ep->rport,
				th->ack_seq,
				be32_add(th->seq, tcp_datalen(iph, th)),
				htons(dcb->state_priv.pull.remote_window));

			if (err < 0) {
				dprintf(1,
					"dts_enqueue_reply_ack failed: %d\n",
					err);
				verdict = DTS_DROP_NOCT;
				goto break_release_ep_intra;
			}

			add_fin_flag(skb, th);
			add_fin_flag(skb, &orig_th);

			dts_ct_enable_emulation(ep_intra);
			dcb->state = DTS_FORWARDING;
			dts_ct_set_max_timeout(ep, 0);

			dts_ct_egress(ep, &orig_iph, &orig_th);
			dts_ct_egress(ep_intra, &orig_iph, &orig_th);
		} else {
			dts_ct_egress(ep, &orig_iph, &orig_th);
			ep_intra->ct = ep->ct;
		}

		if (!tcp_extract_timestamp(skb, &tsval, &tsecr)) {
			dcb->state_priv.pull.delta1 = (s64) dcb->remote_last_timestamp - (s64) tsecr;
			dcb->state_priv.pull.delta2 = (s64) tsecr - (s64) tsval;
		}

		verdict &= ~DTS_CT_MASK;

	break_release_ep_intra:
		dts_release_endpoint(ep_intra);
		break;

	case DTS_ORPHAN:
		/* Pass RST to inform a peer that the connection has been
		   broken */
		if (th->rst) {
			dprintf(1, "accepting RST: %pI4:%d->%pI4:%d\n",
				&iph->saddr, ntohs(th->source),
				&iph->daddr, ntohs(th->dest));

			verdict |= DTS_ACCEPT_MASK;
		}
		break;

	default:
		break;
	}

out:
	return verdict;
}

/* dcb: locked */
static void notify_synack(struct dts_endpoint *ep)
{
	/* NOTE: ep is for proxy */

	__be32 next_node;
	struct ftcs_kconfig *conf = ep->dcb->conf;
	struct ftcs_icc_quadruple q = {
		.laddr = ep->laddr, .raddr = ep->raddr,
		.lport = conf->port, .rport = ep->rport,
	};

	next_node = ftcs_next_responsible_node(conf,
					ep->raddr, ep->rport);
	if (next_node != FTCS_IPADDR_INVALID) {
		int err;
		err = ftcs_icc_xmit(conf, next_node,
				FTCS_ICC_SYNACK_SENT, &q, sizeof(q));
		if (err)
			dprintf(1, "ftcs_icc_xmit failed: %d\n", err);
	}
}

static inline int need_to_notify_sinack(struct ftcs_kconfig *conf,
					__be32 raddr, __be16 rport)
{
	return !dts_ctl_sspa_no_notification
		&& ftcs_responsible_for_packet(conf, raddr, rport);
}

/* TODO: dynamically configure these values!
   cf. http://ja.wikipedia.org/wiki/Hexspeak
*/
static uint32_t sspa_isn_keys[] = {
	0xABADBABE,
	0xBAADF00D,
	0xBADDCAFE,
	0xCAFEBABE,
	0xDEADBEEF,
	0xDEFEC8ED,
	0xFACEFEED,
	0xFEE1DEAD,
	0xBADCAB1E,
};

/*
  Calculate an unique initial sequence number

  Code snippet is taken from secure_tcp_sequence_number()
  in drivers/char/random.c
*/
static uint32_t calc_sspa_isn(__be32 laddr, __be16 lport,
			__be32 raddr, __be16 rport, int secondary)
{
	uint32_t offset = secondary ? (1U << 31) : 0;
	__u32 seq;
	__u32 hash[4];

	hash[0] = laddr;
	hash[1] = raddr;
	hash[2] = (lport << 16) + rport;
	hash[3] = sspa_isn_keys[8];

	/* In the original code time is shifted by 6 bits,
	   but we make the clock more slower because of the offset */
	seq = half_md4_transform(hash, sspa_isn_keys)
		+ (ktime_to_ns(ktime_get_real()) >> 7)
		+ offset;

	return seq;
}

static enum dts_filter_verdict
proxy_inspecting(struct dts_endpoint *ep, struct sk_buff *skb,
		struct iphdr *iph, struct tcphdr *th)
{
	struct dts_endpoint *ep_inter;
	struct dts_cb *dcb = ep->dcb;

	modify_source(skb, iph, th, iph->saddr, dcb->conf->port);
	ep_inter = dts_find_get_endpoint(dcb, iph->saddr, th->source,
					iph->daddr, th->dest);
	if (ep_inter == NULL) {
		dprintf(0, "dts_find_get_endpoint failed!\n");
		return DTS_DROP_NOCT;
	}
	dts_ct_egress(ep, iph, th);
	ep_inter->ct = ep->ct;
	dts_release_endpoint(ep_inter);

	adjust_egress_seq(skb, th, dcb->inspect_seq_offset);

	return DTS_ACCEPT_NOCT;
}

static enum dts_filter_verdict
proxy_inspecting_nohandover(struct dts_endpoint *ep, struct sk_buff *skb,
		struct iphdr *iph, struct tcphdr *th)
{
	dts_ct_egress(ep, iph, th);
	adjust_egress_seq(skb, th, ep->dcb->inspect_seq_offset);

	return DTS_ACCEPT_NOCT;
}

static enum dts_filter_verdict
process_proxy_output(struct dts_endpoint *ep, struct sk_buff *skb)
{
	struct iphdr *iph;
	struct tcphdr *th;
	struct dts_cb *dcb = ep->dcb;
	enum dts_filter_verdict verdict = DTS_DROP_NOCT;

	if (prepare_mangle_tcphdr(skb, &iph, &th))
		goto out;

	switch (dcb->state) {

	case DTS_FIRST_SYN_RECV1:
		if (th->syn && th->ack) {
			if (dts_ctl_sspa_disable) {
				dcb->inspect_seq_offset = 0;
			} else {
				uint32_t isn;

				if (need_to_notify_sinack(dcb->conf,
								ep->raddr,
								ep->rport))
					notify_synack(ep);

				isn = calc_sspa_isn(iph->saddr, th->source,
						iph->daddr, th->dest,
						dcb->state_priv.first.secondary);
				/* "virtual(external view) - real" */
				dcb->inspect_seq_offset = isn - ntohl(th->seq);

				dprintf(2, "isn=%u\n", isn);
			}

			dcb->state_priv.first.ack_for_synack
				= be32_add(th->seq,
					tcp_datalen(iph, th) + 1
					- dcb->inspect_seq_offset);
			dcb->state_priv.first.third_seq = th->ack_seq;
			dcb->state = DTS_FIRST_SYN_RECV2;
			/* Fall through */
		} else if (!th->rst)
			break;

		/* Fall through */
	case DTS_FIRST_SYN_RECV2:
	case DTS_INSPECTING:
		if (dts_nohandover_mode(dcb->conf))
			verdict = proxy_inspecting_nohandover(ep, skb, iph, th);
		else
			verdict = proxy_inspecting(ep, skb, iph, th);

		break;

	case DTS_FIRST_SYN_LOST:
		dts_enqueue_reply_rst(dcb,
				ep->laddr, ep->lport,
				ep->raddr, ep->rport,
				dcb->state_priv.first.third_seq,
				dcb->state_priv.first.ack_for_synack);
		break;


	case DTS_PUSHING:
		if (th->fin) {
			dts_enqueue_reply_rst(
				dcb,
				iph->saddr, th->source,
				iph->daddr, th->dest,
				th->ack_seq,
				be32_add(th->seq,
					tcp_datalen(iph, th) + 1
					- dcb->inspect_seq_offset));
		}

		verdict = DTS_DROP_CT;
		break;

	default:
		break;
	}

out:
	return verdict;
}

unsigned int ftcs_out_filter(
	unsigned int hooknum, struct sk_buff *skb,
	const struct net_device *in, const struct net_device *out,
	int (*okfn)(struct sk_buff *))
{
	struct iphdr *iph, _iph, orig_iph;
	struct tcphdr *th, _th, orig_th;
	enum dts_filter_verdict verdict = DTS_ACCEPT_NOCT;
	unsigned int ret = NF_ACCEPT;
	struct dts_endpoint *ep;
#ifdef CONFIG_FTCS_PROFILING
	struct timeval tv0, tv1;
#endif

	iph = skb_header_pointer(skb, skb_network_offset(skb),
				sizeof(struct iphdr),
				&_iph);
	if (iph == NULL) {
		printk(KERN_WARNING "ftcs_out_filter: "
			"cannot extract IP header\n");
		return NF_DROP;
	}

	if (iph->protocol != IPPROTO_TCP)
		return NF_ACCEPT;

	th = skb_header_pointer(skb, iph->ihl*4,
				sizeof(struct tcphdr),
				&_th);
	if (th == NULL) {
		printk(KERN_WARNING "ftcs_out_filter: "
			"cannot extract TCP header\n");
		return NF_DROP;
	}

	ep = dts_find_get_endpoint_any(iph->saddr, th->source,
				iph->daddr, th->dest);
	if (ep == NULL)
		goto out;

#ifdef CONFIG_FTCS_PROFILING
	do_gettimeofday(&tv0);
#endif

	orig_iph = *iph;
	orig_th  = *th;

	dts_cb_lock(ep->dcb);

	LOG_PACKET(ep->dcb, skb, DTS_LOG_EGRESS);

	/* Short cut for performance optimization */
	if (ep->lep == DTS_LEP_INTER && dts_nohandover_mode(ep->dcb->conf)) {
		verdict = process_proxy_output(ep, skb);
		/* We know that process_proxy_output calls ct_ingress */
		goto out_noct;
	}

	switch (ep->lep) {
	case DTS_LEP_INTER:
		verdict = process_inter_output(ep, skb);
		break;

	case DTS_LEP_PROXY:
		verdict = process_proxy_output(ep, skb);
		break;

	case DTS_LEP_INTRA:
		verdict = process_intra_output(ep, skb);
		break;
	}

	if (verdict & DTS_CT_MASK)
		dts_ct_egress(ep, &orig_iph, &orig_th);
out_noct:
	ret = (verdict & DTS_ACCEPT_MASK) ? NF_ACCEPT : NF_DROP;

	LOG_PACKET(ep->dcb, skb,
		DTS_LOG_EGRESS|DTS_LOG_VERDICT
		|(ret == NF_ACCEPT?DTS_LOG_ACCEPT:0));

	dts_cb_unlock(ep->dcb);

	dts_release_endpoint(ep);

#ifdef CONFIG_FTCS_PROFILING
	do_gettimeofday(&tv1);
	atomic_add(dts_prof_tv_diff(&tv1, &tv0),
		&dts_prof_entries[DTS_PROF_OUT_BODY].sum);
	atomic_inc(&dts_prof_entries[DTS_PROF_OUT_BODY].nr_samples);
#endif

out:
	return ret;
}
