#ifndef FTCS_H
#define FTCS_H

#include <linux/types.h>
#include <linux/if.h>

#ifdef __KERNEL__
#include <linux/time.h>
#include <linux/rbtree.h>
#endif /* #ifdef __KERNEL__ */

#define FTCS_BASE_CTL 2048

#define FTCS_SO_ADD_CONFIG    FTCS_BASE_CTL
#define FTCS_SO_FLUSH_CONFIGS (FTCS_BASE_CTL+1)
#define FTCS_SO_REG_FWDCACHE  (FTCS_BASE_CTL+2)
#define FTCS_SO_REQ_HANDOVER  (FTCS_BASE_CTL+3)
#define FTCS_SO_ADD_VNODES    (FTCS_BASE_CTL+4)
#define FTCS_SO_CLEAR_STATS   (FTCS_BASE_CTL+5)
#define FTCS_SO_REMOVE_NODE   (FTCS_BASE_CTL+6)
#define FTCS_SO_SET_MAX       FTCS_SO_REMOVE_NODE

#define FTCS_SO_GET_INFO     FTCS_BASE_CTL
#define FTCS_SO_GET_CONFIGS (FTCS_BASE_CTL+1)
#define FTCS_SO_GET_STATS   (FTCS_BASE_CTL+2)
#define FTCS_SO_GET_MAX      FTCS_SO_GET_STATS

/*
  Configuration flags
*/
enum ftcs_config_flags {
	FTCS_FLAG_DSTSW = 0x1, /* Notify destination switching */
};

struct ftcs_config {
	/* Network addresses and port numbers
	   -- must be in network byte order */
	uint32_t inter_addr;
	uint32_t intra_addr;
	uint16_t port;
	uint16_t proxy_port;
	uint32_t flags;
};

struct ftcs_info {
	unsigned int nr_configs;
	unsigned int totalsize;
};

#define FTCS_ADD_VNODES_MAX 16
#define FTCS_VNODE_ID_MAX 63

struct ftcs_vnode {
	uint8_t id;
	uint32_t ipaddr;
};

struct ftcs_add_vnodes_param {
	const struct ftcs_config __user *config;
	const struct ftcs_vnode __user *vnodes;
	int nr_vnodes;
};

struct ftcs_stats {
	unsigned int nr_accepted_by_proxy;
	unsigned int nr_accepted_by_server;
	unsigned int sspa_synnotify_recv;
	unsigned int sspa_ack_snoop;
	unsigned int sspa_syn_accept;
	unsigned int sspa_firstsyn_lost;
};

struct ftcs_get_stats_param {
	uint32_t inter_addr;
	uint16_t port;
	struct ftcs_stats stats;
};

struct ftcs_remove_node_param {
	const struct ftcs_config __user *config;
	uint32_t addr;
	int id;
};

#ifdef __KERNEL__
typedef int socklen_t;
#endif

#ifndef __KERNEL__
int ftcs_flush_configs(void);
int ftcs_get_info(struct ftcs_info *info);
int ftcs_get_configs(struct ftcs_config *info, socklen_t *len);
int ftcs_add_config(const struct ftcs_config *config);
int ftcs_add_vnodes(const struct ftcs_config *config,
		const struct ftcs_vnode *vnodes, int nr_vnodes);
int ftcs_request_handover(int s, const struct sockaddr *dest_addr,
			socklen_t addrlen,
			const void *req_buff, size_t req_len,
			size_t next_req_offset);
int ftcs_get_stats(const struct ftcs_config *config, struct ftcs_stats *stats);
int ftcs_clear_stats(const struct ftcs_config *config);
int ftcs_remove_node(const struct ftcs_config *config, uint32_t addr, int id);

#endif /*  #ifndef __KERNEL__ */

struct ftcs_handover_param {
	int socket;
	const struct sockaddr __user *dest_addr;
	socklen_t addrlen;
	const void __user *req_buff;
	size_t req_len;
	size_t next_req_offset;
};

#define DTS_LOG_PACKET_LEN 68

/* flags */

/* Direction flag: 0 == ingress  1 == egress */
#define DTS_LOG_EGRESS   0x1
#define DTS_LOG_INGRESS  0x0

#define DTS_LOG_DIRECTION_MASK (1 << 0)
#define DTS_LOG_MODIFIED (1 << 2) /* Packet was modified by the filter */
#define DTS_LOG_STATE (1 << 3) /* State is valid */
#define DTS_LOG_VERDICT (1 << 4)
#define DTS_LOG_ACCEPT  (1 << 5)

struct dts_packet_log {
	unsigned int id;
	struct timeval timestamp;
	uint8_t filter_state;
	uint8_t flags;
	unsigned char packet_hdr[DTS_LOG_PACKET_LEN];
};

enum dts_state {
	DTS_CLOSED,

	/* First SYN from the client has been seen
	   but SYN+ACK from the host has not been seen */
	DTS_FIRST_SYN_RECV1,

	/* First SYN from the client has been seen
	   and SYN+ACK from the host has been seen, too */
	DTS_FIRST_SYN_RECV2,

	/* We accepted SYN, but another node also accepted it and
	   we lost */
	DTS_FIRST_SYN_LOST,

	DTS_INSPECTING,
	DTS_FORWARDING,
	DTS_PUSH_SYN_WAIT,
	DTS_CONNECTING,
	DTS_PUSHING,
	DTS_PULL_SYN_RECV,
	DTS_PULLING,

	/* The connection on the client side might be an orphan.
	   Querying for other nodes to make sure that nobody takes care of
	   the connection. */
	DTS_ORPHAN_QUERYING,

	/* The connection on the client side is an orphan.
	   This node must send RST upon receiving a packet for this
	   connection. */
	DTS_ORPHAN,

	/* The connection on the client side is not orphan,
	   and it is taken care of by some other node in the cluster.
	   This node should just drop the incoming packet on this connection. */
	DTS_IGNORING,

	/* Mark for a endpoint for a backup connection for
	   the Anytime Available feature
	   SYN is now blocking and waiting for the original response node to
	   handle the connection */
	DTS_SSPA_BLOCKING,

	/* There were no vital sign of the original response node,
	   handle SYN at this node */
	DTS_SSPA_PASSING,

	DTS_IFP, /* Inspect */
};

static inline
const char *dts_state_str(enum dts_state state)
{
	switch (state) {
	case DTS_CLOSED:
		return "CLOSED";
	case DTS_FIRST_SYN_RECV1:
		return "FST_SYN_RECV1";
	case DTS_FIRST_SYN_RECV2:
		return "FST_SYN_RECV2";
	case DTS_FIRST_SYN_LOST:
		return "DTS_FIRST_SYN_LOST";
	case DTS_INSPECTING:
		return "INSPECTING";
	case DTS_FORWARDING:
		return "FORWARDING";
	case DTS_PUSH_SYN_WAIT:
		return "PUSH_SYN_WAIT";
	case DTS_CONNECTING:
		return "CONNECTING";
	case DTS_PUSHING:
		return "PUSHING";
	case DTS_PULL_SYN_RECV:
		return "PULL_SYN_RECV";
	case DTS_PULLING:
		return "PULLING";

	case DTS_ORPHAN_QUERYING:
		return "ORPHAN_QUERYING";
	case DTS_ORPHAN:
		return "ORPHAN";
	case DTS_IGNORING:
		return "IGNORING";

	case DTS_SSPA_BLOCKING:
		return "SSPA_BLOCKING";
	case DTS_SSPA_PASSING:
		return "SSPA_PASSING";

	case DTS_IFP: /* Inspect */
		return "IFP";
	}

	/* BUG */
	return "??";
}


#ifdef __KERNEL__

#include <linux/timer.h>
#include <asm/atomic.h>

#ifdef CONFIG_FTCS_DEBUG
extern int dts_ctl_debug_level;
#  define dprintk(args...) printk(args)
#  define dprintf(level, fmt, ...)					\
	do {								\
		if (level <= dts_ctl_debug_level)			\
			printk(KERN_DEBUG "DTS: %s:%d " fmt,		\
				__func__, __LINE__, ##__VA_ARGS__);	\
	} while (0)
#else
#  define dprintk(args...)
#  define dprintf(level, fmt, ...) do {} while (0)
#endif

#define FTCS_IPADDR_SELF 0
#define FTCS_IPADDR_INVALID (0xffffffff)

#define FTCS_INMSG_SYNFWD 0
#define FTCS_INMSG_SNDRST 1
#define FTCS_INMSG_SNDFIN 2

struct ftcshdr {
	uint32_t raddr;
	uint32_t laddr;
	uint16_t msg;
	uint16_t rport;
	uint16_t lport;
};

struct sk_buff;
struct proc_dir_entry;

struct ftcs_kconfig {
	struct list_head list;
	atomic_t refcnt;
	__be32 inter_addr;
	__be32 intra_addr;
	__be32 intra_mask;
	__be16 port;

	__be16 proxy_port;

	uint32_t flags;

	rwlock_t neighbors_lock;
	struct rb_root idtree;
	struct rb_root pnode_tree;
	int nr_neighbors; /* number of neighbor nodes, excluding myself */
	struct timespec ts_last_died;

	/* Statistics */
	atomic_t accepted_inter;
	atomic_t accepted_intra;
	atomic_t sspa_synnotify_recv;
	atomic_t sspa_ack_snoop;
	atomic_t sspa_syn_accept;
	atomic_t sspa_firstsyn_lost;

#ifdef CONFIG_FTCS_DEBUG
	atomic_t renotify;
#endif
	struct proc_dir_entry *proc_file;
};
int ftcs_forward_packet(uint32_t addr, struct sk_buff *skb);
int ftcs_xmit_inmsg(uint32_t dest, const struct ftcshdr *dsh_send);

/*
 * config.c
 */

struct ftcs_kconfig *ftcs_find_get_config(__be32 addr, __be16 port);
struct ftcs_kconfig *ftcs_find_get_config_from_proxy(__be32 addr,
						__be16 proxy_port);
int ftcs_register_config(struct ftcs_kconfig *conf);
void ftcs_release_config(struct ftcs_kconfig *conf);
int ftcs_sys_add_config(void __user *user, unsigned int len);
void ftcs_sys_flush_configs(void);
int ftcs_sys_add_vnodes(void __user *user, unsigned int len);
void ftcs_get_configs_info(struct ftcs_info *info);
int ftcs_sys_get_configs(void __user *uconfs, int *len);
int ftcs_sys_get_stats(void __user *user, int *len);
int ftcs_sys_clear_stats(void __user *user, unsigned int len);
int ftcs_sys_remove_node(void __user *user, unsigned int len);
void ftcs_flush_configs(void);
int ftcs_config_init(void);
void ftcs_config_exit(void);


#endif /* __KERNEL__ */

#endif /* FTCS_H */
