#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/kthread.h>
#include <linux/file.h>
#include <linux/version.h>
#include <linux/sysctl.h>

#include <linux/spinlock.h>

#include <linux/mutex.h>

#include <linux/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>

#include <net/ip.h>
#include <net/tcp.h>
#include <net/protocol.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>

#include <linux/etherdevice.h>

#include <linux/time.h>

#include "ftcs.h"
#include "dts_internal.h"

/* #ifdef CONFIG_FTCS_DEBUG */
/* 192.168.99.1 */
#define FTCS_INTRA_ADDR_FOR_TEST __constant_htonl(0xc0a86301)
/* #endif */

/* Slab caches */
struct kmem_cache *dts_cb_cachep;
struct kmem_cache *dts_endpoint_cachep;
struct kmem_cache *ftcs_node_cachep;

struct proc_dir_entry *dts_proc_net_dts;

struct crypto_hash *ftcs_tfm;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
DEFINE_SPINLOCK(ftcs_hash_lock);
#else
spinlock_t ftcs_hash_lock = SPIN_LOCK_UNLOCKED;
#endif
MODULE_LICENSE("GPL");

#ifdef CONFIG_FTCS_DEBUG
#define DUMP_SKB(skb, len) dump_skb(skb, len)
#else
#define DUMP_SKB(skb, len) do {} while (0)
#endif /* ifdef CONFIG_FTCS_DEBUG */

#ifdef CONFIG_FTCS_DEBUG
static atomic_t dcb_count = ATOMIC_INIT(0);
#endif

/*
 * Global configuration variables
 */
int dts_ctl_sspa_timeout = DTS_SSPA_TIMEOUT_DEFAULT;
int dts_ctl_sspa_disable;
int dts_ctl_sspa_no_ack_snooping;
int dts_ctl_sspa_no_notification = 1;
int dts_ctl_initial_conn_timeout = DTS_INITIAL_CONN_TIMEOUT_DEFAULT;
int dts_ctl_conn_retrans_timeout = DTS_CONN_RETRANS_TIMEOUT_DEFAULT;

#ifdef CONFIG_FTCS_DEBUG
int dts_ctl_debug_level = 0;
#endif

/*
  sysctl tables -- thanks to net/netfilter/ipvs/ip_vs_ctl.c
 */

static struct ctl_table dts_vars[] = {
	{
		.procname     = "handover_timeout",
		.data         = &dts_ctl_handover_timeout,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		/* TODO: what happens if a negative value is passed */
		.proc_handler = proc_dointvec,
	},
	{
		.procname     = "sspa_timeout",
		.data         = &dts_ctl_sspa_timeout,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		.proc_handler = proc_dointvec_ms_jiffies,
	},
	{
		.procname     = "sspa_disable",
		.data         = &dts_ctl_sspa_disable,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname     = "initial_conn_timeout",
		.data         = &dts_ctl_initial_conn_timeout,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		.proc_handler = proc_dointvec_ms_jiffies,
	},
	{
		.procname     = "conn_retrans_timeout",
		.data         = &dts_ctl_conn_retrans_timeout,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		.proc_handler = proc_dointvec_ms_jiffies,
	},
	{
		.procname     = "sspa_no_ack_snooping",
		.data         = &dts_ctl_sspa_no_ack_snooping,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		.proc_handler = proc_dointvec,
	},
	{
		.procname     = "sspa_no_notification",
		.data         = &dts_ctl_sspa_no_notification,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		.proc_handler = proc_dointvec,
	},
#ifdef CONFIG_FTCS_DEBUG
	{
		.procname     = "debug_level",
		.data         = &dts_ctl_debug_level,
		.maxlen       = sizeof(int),
		.mode         = 0644,
		.proc_handler = proc_dointvec,
	},
#endif /* #ifdef CONFIG_FTCS_DEBUG */
	{ }
};

static const struct ctl_path net_dts_ctl_path[] = {
	CTL_PATH_INITIALIZER("net", CTL_NET),
	CTL_PATH_INITIALIZER("dts", 0),
	{ }
};

static struct ctl_table_header *sysctl_header;

#ifdef CONFIG_FTCS_PROFILING

struct dts_prof_entry dts_prof_entries[DTS_PROF_TOTAL] = {
	{ "out_body" },
	{ "out_conntrack" },
	{ "in_body" },
	{ "in_syn" },
	{ "in_syn_for_other" },
	{ "in_syn_noho" },
};

unsigned long
dts_prof_tv_diff(const struct timeval *tv1, const struct timeval *tv0)
{
	unsigned long ret;

	ret = (tv1->tv_sec - tv0->tv_sec) * 1000000;
	ret += tv1->tv_usec;
	ret -= tv0->tv_usec;

	return ret;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
static void filter_ctor(void *ptr)
#else
static void filter_ctor(struct kmem_cache *cache, void *ptr)
#endif
{
	struct dts_cb *dcb = ptr;

	memset(dcb, 0, sizeof(*dcb));

	spin_lock_init(&dcb->lock);
	atomic_set(&dcb->refcnt, 0);
	INIT_LIST_HEAD(&dcb->delayed_work_list);
#ifdef CONFIG_FTCS_DEBUG
	dcb->id = atomic_inc_return(&dcb_count) - 1;
#endif
}

void __dts_cb_release(struct dts_cb *dcb)
{
	ftcs_release_config(dcb->conf);
	switch (dcb->state) {
	case DTS_ORPHAN_QUERYING:
		nsv4_destroy(&dcb->state_priv.orphan.ns);
		break;

	case DTS_SSPA_BLOCKING:
	case DTS_SSPA_PASSING:
		if (dcb->state_priv.sspa.skb != NULL) {
			kfree_skb(dcb->state_priv.sspa.skb);
			dcb->state_priv.sspa.skb = NULL;
		}
		break;
	default:
		break;
	}
	kmem_cache_free(dts_cb_cachep, dcb);
}

static struct nf_hook_ops ftcs_preroute_ops = {
	.hook = ftcs_preroute_filter,
	.owner = THIS_MODULE,
	.pf = PF_INET,
	.hooknum = NF_INET_PRE_ROUTING,
	.priority = NF_IP_PRI_FIRST,
};

static struct nf_hook_ops ftcs_l4_preroute_ops = {
	.hook = ftcs_l4_preroute_filter,
	.owner = THIS_MODULE,
	.pf = PF_INET,
	.hooknum = NF_INET_PRE_ROUTING,
	.priority = NF_IP_PRI_FIRST,
};

static struct nf_hook_ops ftcs_out_ops = {
	.hook = ftcs_out_filter,
	.owner = THIS_MODULE,
	.pf = PF_INET,
	.hooknum = NF_INET_LOCAL_OUT,
	.priority = -1,
};

static int ftcs_so_set(struct sock *sk, int cmd,
		void __user *user, unsigned int len)
{
	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	switch (cmd) {
	case FTCS_SO_ADD_CONFIG:
		return ftcs_sys_add_config(user, len);

	case FTCS_SO_FLUSH_CONFIGS:
		ftcs_flush_configs();
		break;

	case FTCS_SO_REQ_HANDOVER:
		return dts_sys_request_handover(user, len);

	case FTCS_SO_ADD_VNODES:
		return ftcs_sys_add_vnodes(user, len);

	case FTCS_SO_CLEAR_STATS:
		return ftcs_sys_clear_stats(user, len);

	case FTCS_SO_REMOVE_NODE:
		return ftcs_sys_remove_node(user, len);

	default:
		return -EINVAL;
	}

	return 0;
}

static int ftcs_so_get(struct sock *sk, int cmd, void __user *user, int *len)
{
	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	if (len == NULL)
		return -EFAULT;

	switch (cmd) {
		struct ftcs_info info;
	case FTCS_SO_GET_INFO:
		if (*len != sizeof(struct ftcs_info))
			return -EINVAL;

		ftcs_get_configs_info(&info);
		if (copy_to_user(user, &info, sizeof(info)))
			return -EFAULT;

		break;

	case FTCS_SO_GET_CONFIGS:
		return ftcs_sys_get_configs(user, len);

	case FTCS_SO_GET_STATS:
		return ftcs_sys_get_stats(user, len);

	default:
		return -EINVAL;
	}

	return 0;
}

static struct nf_sockopt_ops ftcs_sockopt_ops = {
	.pf = PF_INET,
	.set_optmin = FTCS_BASE_CTL,
	.set_optmax = FTCS_SO_SET_MAX+1,
	.set = ftcs_so_set,
	.get_optmin = FTCS_BASE_CTL,
	.get_optmax = FTCS_SO_GET_MAX+1,
	.get = ftcs_so_get,
	.owner = THIS_MODULE,
};

static int __init ftcs_init(void)
{
	int ret = 0;

	dts_cb_cachep  = NULL;
	dts_endpoint_cachep = NULL;
	ftcs_node_cachep = NULL;

	dts_cb_cachep = kmem_cache_create(
		"dts_cb", sizeof(struct dts_cb),
		0,
		0,
		filter_ctor);
	if (dts_cb_cachep == NULL) {
		printk(KERN_ERR
			"ftcs: kmem_cache_create(dts_cb_cachep) failed\n");
		ret = -ENOMEM;
		goto out_cleancachep;
	}

	dts_endpoint_cachep = kmem_cache_create(
		"dts_endpoint", sizeof(struct dts_endpoint),
		0, 0, NULL);
	if (dts_endpoint_cachep == NULL) {
		printk(KERN_ERR
			"ftcs: kmem_cache_create(endpoint_cachep) failed\n");
		ret = -ENOMEM;
		goto out_cleancachep;
	}

	ftcs_node_cachep = kmem_cache_create(
		"ftcs_node",
		max(sizeof(struct ftcs_kvnode), sizeof(struct ftcs_kpnode)),
		0, 0, NULL);
	if (ftcs_node_cachep == NULL) {
		printk(KERN_ERR
			"ftcs: kmem_cache_create(ftcs_node_cachep) failed\n");
		ret = -ENOMEM;
		goto out_cleancachep;
	}

	ftcs_tfm = crypto_alloc_hash("md5", 0, 0);
	if (IS_ERR(ftcs_tfm)) {
		printk(KERN_ERR "ftcs: crypto_alloc_hash failed\n");
		ret = PTR_ERR(ftcs_tfm);
		goto out_cleancachep;
	}

	ret = nsv4_setup();
	if (ret) {
		printk(KERN_ERR "ftcs: nsv4_setup failed\n");
		goto out_cleantfm;
	}

	ret = inet_add_protocol(&ftcs_icc_protocol, IPPROTO_FTCSICC);
	if (ret) {
		printk(KERN_ERR "ftcs: inet_add_protocol failed\n");
		goto out_cleannsv4;
	}

	ret = dts_log_setup();
	if (ret)
		goto out_cleanproto;

	dts_proc_net_dts = proc_mkdir("dts", init_net.proc_net);
	if (dts_proc_net_dts == NULL) {
		ret = -ENOMEM;
		printk(KERN_ERR "ftcs: proc_mkdir(\"dts\") failed\n");
		goto out_cleanlog;
	}

	ret = dts_endpoint_init();
	if (ret)
		goto out_cleanproc;

	ret = ftcs_config_init();
	if (ret)
		goto out_cleanep;

	sysctl_header = register_sysctl_paths(net_dts_ctl_path, dts_vars);
	if (sysctl_header == NULL) {
		/* TODO: never fails? */
		ret = -ENOMEM;
		printk(KERN_ERR "ftcs: register_sysctl_paths failed\n");
		goto out_cleanconfig;
	}

	nf_register_sockopt(&ftcs_sockopt_ops);

	if (dts_l4_only_mode) {
		nf_register_hook(&ftcs_l4_preroute_ops);
	} else {
		nf_register_hook(&ftcs_out_ops);
		nf_register_hook(&ftcs_preroute_ops);
	}

#ifdef CONFIG_FTCS_DEBUG
	printk(KERN_INFO "ftcs%s: registered (DEBUG)\n",
		dts_l4_only_mode ? "(L4 only mode)" : "");
#else
	printk(KERN_INFO "ftcs%s: registered\n",
		dts_l4_only_mode ? "(L4 only mode)" : "");
#endif
	printk(KERN_INFO "ftcs: sizeof(dts_cb)=%zu\n", sizeof(struct dts_cb));
	printk(KERN_INFO
		"ftcs: sizeof(dts_endpoint)=%zu sizeof(dts_conntrack)=%zu\n",
		sizeof(struct dts_endpoint), sizeof(struct dts_conntrack));

	return 0;

	unregister_sysctl_table(sysctl_header);

out_cleanconfig:
	ftcs_config_exit();

out_cleanep:
	dts_endpoint_exit();

out_cleanproc:
	proc_net_remove(&init_net, "dts");

out_cleanlog:
	dts_log_cleanup();

out_cleanproto:
	inet_del_protocol(&ftcs_icc_protocol, IPPROTO_FTCSICC);

out_cleannsv4:
	nsv4_cleanup();

out_cleantfm:
	crypto_free_hash(ftcs_tfm);

out_cleancachep:
	if (dts_endpoint_cachep)
		kmem_cache_destroy(dts_endpoint_cachep);
	if (dts_cb_cachep)
		kmem_cache_destroy(dts_cb_cachep);
	if (ftcs_node_cachep)
		kmem_cache_destroy(ftcs_node_cachep);

	return ret;
}

static void ftcs_exit(void)
{
	ftcs_flush_configs();

	nf_unregister_sockopt(&ftcs_sockopt_ops);

	if (dts_l4_only_mode) {
		nf_unregister_hook(&ftcs_l4_preroute_ops);
	} else {
		nf_unregister_hook(&ftcs_out_ops);
		nf_unregister_hook(&ftcs_preroute_ops);
	}

	unregister_sysctl_table(sysctl_header);

	inet_del_protocol(&ftcs_icc_protocol, IPPROTO_FTCSICC);

	dts_log_cleanup();

	nsv4_cleanup();

	crypto_free_hash(ftcs_tfm);

	ftcs_config_exit();
	dts_endpoint_exit();
	proc_net_remove(&init_net, "dts");


	kmem_cache_destroy(dts_endpoint_cachep);
	kmem_cache_destroy(dts_cb_cachep);
	kmem_cache_destroy(ftcs_node_cachep);

	printk(KERN_INFO "ftcs: unregistered\n");
}

module_init(ftcs_init);
module_exit(ftcs_exit);
