#include "neighsetv4.h"

#include <linux/slab.h>

struct kmem_cache *node_cachep;

static inline uint32_t addr_prefix(uint32_t addr)
{
	return addr & NSV4_PREFIX_MASK;
}

static inline uint32_t addr_suffix(uint32_t addr)
{
	return addr & NSV4_SUFFIX_MASK;
}

static inline int addr_bitmask(uint32_t addr)
{
	int suffix = addr_suffix(addr);
	int mask = 1 << suffix;

	return mask;
}

static inline int prefix_compare(uint32_t l, uint32_t r)
{
	return (int)(l - r);
}

static void add_to_bitset(struct nsv4_node *n, uint32_t addr)
{
	n->bitset |= addr_bitmask(addr);
}

static int member_of_bitset(struct nsv4_node *n, uint32_t addr)
{
	int mask = addr_bitmask(addr);
	return (n->bitset & mask) != 0;
}

/*
  rbtree insertion code: almost comes from Documentation/rbtree.txt
*/
int nsv4_add(struct neighsetv4 *s, __be32 beaddr)
{
	struct nsv4_node *newnode;
	struct rb_node **new = &(s->rbroot.rb_node), *parent = NULL;
	uint32_t addr   = ntohl(beaddr); 
	uint32_t prefix = addr_prefix(addr);

	while (*new) {
		struct nsv4_node *n;
		int cmp;

		n = container_of(*new, struct nsv4_node, rbnode);
		cmp = prefix_compare(prefix, n->prefix);

		if (cmp < 0)
			new = &((*new)->rb_left);
		else if (cmp > 0)
			new = &((*new)->rb_right);
		else {
			if (!member_of_bitset(n, addr)) {
				add_to_bitset(n, addr);
				s->nr_set++;
			}
			return 0;
		}
	}

	/* Need to allocate a new node */
	newnode = kmem_cache_alloc(node_cachep, GFP_ATOMIC);
	if (newnode == NULL)
		return -ENOMEM;

	newnode->bitset = 0;
	newnode->prefix = addr_prefix(addr);
	add_to_bitset(newnode, addr);
	s->nr_set++;

	rb_link_node(&newnode->rbnode, parent, new);
	rb_insert_color(&newnode->rbnode, &s->rbroot);

	return 0;
}

int nsv4_member(const struct neighsetv4 *s, __be32 beaddr)
{
	uint32_t addr = ntohl(beaddr);
	struct rb_node *node = s->rbroot.rb_node;

	while (node) {
		struct nsv4_node *n;
		int cmp;

		n = container_of(node, struct nsv4_node, rbnode);
		cmp = prefix_compare(addr_prefix(addr), n->prefix);

		if (cmp == 0)
			return member_of_bitset(n, addr);
		else if (cmp < 0)
			node = node->rb_left;
		else
			node = node->rb_right;
	}

	return 0;
}

int nsv4_size(const struct neighsetv4 *s)
{
	return s->nr_set;
}

static int next_index(nsv4_bitset_t bitset, int start_index)
{
	int i = start_index;

	bitset >>= start_index;
	while (i <= NSV4_SUFFIX_MASK) {
		if (bitset & 1)
			return i;
		bitset >>= 1;
		i++;
	}
	return -1;
}

int nsv4_first(const struct neighsetv4 *s, struct nsv4_iter *iter)
{
	struct rb_node *rbnode;
	struct nsv4_node *node;

	rbnode = rb_first(&s->rbroot);
	if (rbnode == NULL)
		return -ENOENT;

	node = container_of(rbnode, struct nsv4_node, rbnode);

	iter->rbnode = rbnode;
	iter->index  = next_index(node->bitset, 0);

	return 0;
}

int nsv4_next(struct nsv4_iter *iter)
{
	int next_idx;
	struct rb_node *rbnode = iter->rbnode;
	struct nsv4_node *node;

	node = container_of(rbnode, struct nsv4_node, rbnode);

	next_idx = next_index(node->bitset, iter->index+1);
	if (next_idx >= 0) {
		iter->index = next_idx;
		return 0;
	}

	rbnode = rb_next(iter->rbnode);
	if (rbnode == NULL)
		return -ENOENT;

	node = container_of(rbnode, struct nsv4_node, rbnode);
	iter->rbnode = rbnode;
	iter->index  = next_index(node->bitset, 0);

	return 0;
}

__be32 nsv4_iter_address(const struct nsv4_iter *iter)
{
	struct nsv4_node *node;

	node = container_of(iter->rbnode, struct nsv4_node, rbnode);
	return htonl(node->prefix | iter->index);
}

void nsv4_clear(struct neighsetv4 *s)
{
	struct rb_node *node;

	while ((node = rb_first(&s->rbroot)) != NULL) {
		struct nsv4_node *n;
		n = container_of(node, struct nsv4_node, rbnode);
		rb_erase(&n->rbnode, &s->rbroot);
		kmem_cache_free(node_cachep, n);
	}

	nsv4_init(s);
}

void nsv4_init(struct neighsetv4 *s)
{
	s->rbroot = RB_ROOT;
	s->nr_set = 0;
}

void nsv4_destroy(struct neighsetv4 *s)
{
	nsv4_clear(s);
}

#define SHOW_ERROR()						\
	do {							\
		printk(KERN_ERR "Test failed: %s:%d\n",		\
			__FILE__, __LINE__);			\
	} while (0)

static int test(void)
{
	struct neighsetv4 s;
	int ret;
	struct nsv4_iter iter;

	nsv4_init(&s);

	if ((ret = nsv4_size(&s)) != 0) {
		SHOW_ERROR();
		goto out;
	}

	/* 172.16.16.1 */
	if ((ret = nsv4_add(&s, htonl(0xAC101001))) != 0) {
		SHOW_ERROR();
		goto out;
	}

	/* 172.16.16.3 */
	if ((ret = nsv4_add(&s, htonl(0xAC101003))) != 0) {
		SHOW_ERROR();
		goto out;
	}

	/* 172.16.16.1 */
	if ((ret = nsv4_member(&s, htonl(0xAC101001))) == 0) {
		SHOW_ERROR();
		ret = -EINVAL;
		goto out;
	}

	/* 172.16.16.2 */
	if ((ret = nsv4_member(&s, htonl(0xAC101002))) != 0) {
		SHOW_ERROR();
		goto out;
	}

	/* 172.16.17.10 */
	if ((ret = nsv4_add(&s, htonl(0xAC10110A))) != 0) {
		SHOW_ERROR();
		goto out;
	}

	/* 172.16.17.10 */
	if ((ret = nsv4_member(&s, htonl(0xAC10110A))) == 0) {
		SHOW_ERROR();
		ret = -EINVAL;
		goto out;
	}

	if ((ret = nsv4_size(&s)) != 3) {
		SHOW_ERROR();
		ret = -EINVAL;
		goto out;
	}

	ret = -EINVAL;

	if (nsv4_first(&s, &iter) < 0) {
		SHOW_ERROR();
		goto out;	
	}

	if (nsv4_iter_address(&iter) != htonl(0xAC101001)) {
		__be32 addr = nsv4_iter_address(&iter);
		printk(KERN_ERR "nsv4_iter_address(&iter)=%pI4\n", &addr);
		SHOW_ERROR();
		goto out;
	}

	if (nsv4_next(&iter) < 0) {
		SHOW_ERROR();
		goto out;	
	}

	if (nsv4_iter_address(&iter) != htonl(0xAC101003)) {
		SHOW_ERROR();
		goto out;
	}

	if (nsv4_next(&iter) < 0) {
		SHOW_ERROR();
		goto out;	
	}

	if (nsv4_iter_address(&iter) != htonl(0xAC10110A)) {
		SHOW_ERROR();
		goto out;
	}

	if (nsv4_next(&iter) == 0) {
		SHOW_ERROR();
		goto out;	
	}

	nsv4_clear(&s);

	if ((ret = nsv4_size(&s)) != 0) {
		SHOW_ERROR();
		goto out;
	}

	ret = 0;
	
out:
	nsv4_destroy(&s);
	return ret;
}

int nsv4_setup(void)
{
	node_cachep = kmem_cache_create("nsv4_node", sizeof(struct nsv4_node),
					0, 0, NULL);
	if (node_cachep == NULL)
		return -ENOMEM;

	if (test()) {
		kmem_cache_destroy(node_cachep);
		return -EINVAL;
	}

	return 0;
}

void nsv4_cleanup(void)
{
	kmem_cache_destroy(node_cachep);
}
