#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <arpa/inet.h>

#include <errno.h>

#include <ftcs.h>

static int invoke_setsockopt(int optname, void *optval, socklen_t optlen)
{
	int s;
	int ret = 0;

	s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (s < 0)
		return -1;

	if (setsockopt(s, SOL_IP, optname, optval, optlen) < 0)
		ret = -1;

	close(s);

	return ret;
}

static int invoke_getsockopt(int optname, void *optval, socklen_t *optlen)
{
	int s;
	int ret = 0;

	s = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (s < 0)
		return -1;

	if (getsockopt(s, SOL_IP, optname, optval, optlen) < 0)
		ret = -1;

	close(s);

	return ret;
}

int ftcs_flush_configs(void)
{
	return invoke_setsockopt(FTCS_SO_FLUSH_CONFIGS, NULL, 0);
}

int ftcs_get_info(struct ftcs_info *info)
{
	socklen_t len = sizeof(*info);
	return invoke_getsockopt(FTCS_SO_GET_INFO, info, &len);
}

int ftcs_get_configs(struct ftcs_config *buff, socklen_t *len)
{
	return invoke_getsockopt(FTCS_SO_GET_CONFIGS, buff, len);
}

int ftcs_add_config(const struct ftcs_config *config)
{
	return invoke_setsockopt(FTCS_SO_ADD_CONFIG, config, sizeof(*config));
}

int ftcs_add_vnodes(const struct ftcs_config *config,
					const struct ftcs_vnode *vnodes, int nr_vnodes)
{
	struct ftcs_add_vnodes_param p = {
		.config = config,
		.vnodes = vnodes,
		.nr_vnodes = nr_vnodes,
	};

	return invoke_setsockopt(FTCS_SO_ADD_VNODES, &p, sizeof(p));
}

int ftcs_get_stats(const struct ftcs_config *config, struct ftcs_stats *stats)
{
	struct ftcs_get_stats_param p = {
		.inter_addr = config->inter_addr,
		.port = config->port,
	};
	socklen_t len = sizeof(p);
	int e;

	e = invoke_getsockopt(FTCS_SO_GET_STATS, &p, &len);
	if (e)
		return e;

	*stats = p.stats;

	return 0;
}

int ftcs_clear_stats(const struct ftcs_config *config)
{
	return invoke_setsockopt(FTCS_SO_CLEAR_STATS, config, sizeof(*config));
}

int ftcs_remove_node(const struct ftcs_config *config, uint32_t addr, int id)
{
	struct ftcs_remove_node_param p;

	p.config = config;
	p.addr   = addr;
	p.id     = id;

	return invoke_setsockopt(FTCS_SO_REMOVE_NODE, &p, sizeof(p));
}

int ftcs_request_handover(int s, const struct sockaddr *dest_addr,
			  socklen_t addrlen,
			  const void *req_buff, size_t req_len,
			  size_t next_req_offset)
{
	struct ftcs_handover_param param;

	param.socket = s;
	param.dest_addr = dest_addr;
	param.addrlen = addrlen;
	param.req_buff = req_buff;
	param.req_len = req_len;
	param.next_req_offset = next_req_offset;

	return invoke_setsockopt(FTCS_SO_REQ_HANDOVER,
				 &param, sizeof(param));
}
