/* 
 * dtsd: DTS node monitoring daemon
 * by Hajime Fujita <dtool1@dependable-os.net>
 *
 * based on api_test, Copyright (C) 2000 Alan Robertson <alanr@unix.sh>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <errno.h>

#include <stdio.h>
#include <string.h>

#include <heartbeat.h>
#include <hb_api.h>

#include "hbc.h"

static void (*on_node_down)(const struct hbc_node *node);
static struct hbc_node *hbc_nodelist;
static int hbc_nr_nodes;
static ll_cluster_t *hb;
static int hbc_debug;

#define dprintf(args...) do {								\
		if (hbc_debug > 0)									\
			fprintf(stderr, "libhbc: " args);				\
	} while (0)

static void clear_globals(void)
{
	on_node_down = NULL;
	hbc_nodelist = NULL;
	hbc_nr_nodes = 0;
	hb = NULL;
	hbc_debug = 0;
}

static struct hbc_node *find_node(const char *name)
{
	int i;

	for (i = 0; i < hbc_nr_nodes; i++) {
		if (strcmp(name, hbc_nodelist[i].name) == 0)
			return &hbc_nodelist[i];
	}

	return NULL;
}

const struct hbc_node *hbc_find_primary_node(void)
{
	const struct hbc_node *node = NULL;
	int i;

	for (i = 0; i < hbc_nr_nodes; i++) {
		dprintf("hbc_find_primary_node: %d\n", i);
		if (hbc_nodelist[i].status != HBC_NODE_UP) {
			dprintf("  != HBC_NODE_UP\n");
			continue;
		}

		if (node == NULL || node->addr > hbc_nodelist[i].addr) {
			dprintf("  node selected\n");
			node = &hbc_nodelist[i];
		}
	}

	return node;
}

static void
NodeStatus(const char * name, const char * status, void * private)
{
	dprintf("Status update: Node %s now has status %s\n", name, status);

	if (strcmp(status, DEADSTATUS) == 0) {
		struct hbc_node *node;

		node = find_node(name);
		if (node != NULL && node->status == HBC_NODE_UP) {
			node->status = HBC_NODE_DOWN;
			on_node_down(node);
		} else 
			dprintf("NodeStatus: find_node returned NULL\n");
	}
}

static void
ClientStatus(const char * name, const char * client, const char * status
,	void * private)
{
	dprintf("Status update: Client %s/%s now has status [%s]\n",
			name, client, status);

	if (strcmp(status, LEAVESTATUS) == 0) {
		struct hbc_node *node;

		node = find_node(name);
		if (node != NULL) {
			node->status = HBC_NODE_DOWN;
			on_node_down(node);
		}
	}
}

void hbc_mainloop(void)
{
	if (hb == NULL)
		return;

	for (;;)
		hb->llc_ops->rcvmsg(hb, 1);
}

static int verify_nodelist(const struct hbc_node *nodelist, int nr_nodes)
{
	int i;

	for (i = 0; i < nr_nodes; i++) {
		if (nodelist[i].status != HBC_NODE_UP)
			return EINVAL;
	}

	return 0;
}

int hbc_init(const char *progname,
			 const char *clname,
			 void (*downevent)(const struct hbc_node *node),
			 struct hbc_node *nodelist,
			 int nr_nodes, int debug)
{
	int ret;

	if (hb != NULL)
		return EINVAL;

	ret = verify_nodelist(nodelist, nr_nodes);
	if (ret)
		return ret;

	cl_log_set_entity(progname);
	cl_log_enable_stderr(debug);
	cl_log_set_facility(LOG_USER);

	hb = ll_cluster_new("heartbeat");
	if (hb == NULL) {
		dprintf("Cannot allocate the cluster object\n");
		return ENOMEM;
	}

	hbc_nodelist = nodelist;
	hbc_nr_nodes = nr_nodes;
	on_node_down = downevent;
	hbc_debug    = debug;

	ret = ENOMEM;

	if (hb->llc_ops->signon(hb, clname)!= HA_OK) {
		dprintf("Cannot sign on with heartbeat\n");
		dprintf("REASON: %s", hb->llc_ops->errmsg(hb));
		goto out_delete;
	}

	if (hb->llc_ops->set_nstatus_callback(hb, NodeStatus, NULL) !=HA_OK){
		dprintf("Cannot set node status callback\n");
		dprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_signoff;
	}

	if (hb->llc_ops->set_cstatus_callback(hb, ClientStatus, NULL)!=HA_OK){
		dprintf("Cannot set client status callback\n");
		dprintf("REASON: %s\n", hb->llc_ops->errmsg(hb));
		goto out_signoff;
	}

	return 0;

out_signoff:
	hb->llc_ops->signoff(hb, TRUE);

out_delete:
	hb->llc_ops->delete(hb);

	clear_globals();

	return ret;
}

void hbc_cleanup(void)
{
	if (hb == NULL)
		return;

	hb->llc_ops->signoff(hb, TRUE);
	hb->llc_ops->delete(hb);

	clear_globals();
}
