#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>

#include <sys/socket.h>
#include <sys/types.h>

#include <arpa/inet.h>

#include <ftcs.h>

static int resolve_host(const char *hostname, uint32_t *addr_p)
{
	struct hostent *ent = gethostbyname(hostname);
	if (ent == NULL) {
		fprintf(stderr, "Host not found: %s\n", hostname);
		return 1;
	}
	memcpy(addr_p, ent->h_addr, ent->h_length);

	return 0;
}

static int fill_config(const char *inter, const char *intra,
					   const char *port, const char *proxy_port,
					   struct ftcs_config *conf)
{
	struct hostent *ent;

	ent = gethostbyname(inter);
	if (ent == NULL) {
		fprintf(stderr, "Host not found: %s\n", inter);
		return 1;
	}
	memcpy(&conf->inter_addr, ent->h_addr, ent->h_length);

	if (intra) {
		ent = gethostbyname(intra);
		if (ent == NULL) {
			fprintf(stderr, "Host not found: %s\n", intra);
			return 1;
		}
		memcpy(&conf->intra_addr, ent->h_addr, ent->h_length);
	}

	conf->port = htons(atoi(port));
	if (proxy_port)
		conf->proxy_port = htons(atoi(proxy_port));

	return 0;
}

static void print_join_usage(void)
{
	fprintf(stderr,
			"Usage: dtsctl join [options] inter_addr intra_addr port proxy_port\n"
			"\n"
			"Options:\n"
			"  -d    enable destination switching feature\n"
		);
}

static int join(int argc, char * const argv[])
{
	struct ftcs_config config;
	const char *inter, *intra, *port, *proxy_port;
	uint32_t flags = 0;

	for (;;) {
		int r = getopt(argc, argv, "d");
		if (r == -1)
			break;

		switch (r) {
		case 'd':
			flags |= FTCS_FLAG_DSTSW;
			break;

		default:
			print_join_usage();
			return 1;
		}
	}
	
	if (argc - optind != 4) {
		print_join_usage();
		return 1;
	}

	inter      = argv[optind];
	intra      = argv[optind+1];
	port       = argv[optind+2];
	proxy_port = argv[optind+3];

	memset(&config, 0, sizeof(config));

	if (fill_config(inter, intra, port, proxy_port, &config))
		return 1;

	config.flags = flags;

	if (ftcs_add_config(&config) < 0) {
		perror("ftcs_add_config() failed");
		return 1;
	}

	return 0;
}

static int list(int argc, char *const argv[])
{
	struct ftcs_info info;
	char *buff, *p;
	socklen_t len;

	if (ftcs_get_info(&info) < 0) {
		perror("ftcs_get_info() failed");
		return 1;
	}

	len = info.totalsize;

	do {
		socklen_t oldlen = len;
		buff = malloc(len);
		if (buff == NULL) {
			fprintf(stderr, "Out of memory\n");
			return 1;
		}

		if (ftcs_get_configs((struct ftcs_config *) buff, &len) < 0) {
			free(buff);
			buff = NULL;
			if (oldlen == len) {
				perror("ftcs_get_configs() failed");
				return 1;
			}
		}
	} while (buff == NULL);
   
	p = buff;
	printf("Installed configurations:\n");
	printf("M/S INTERnet Address INTRAnet Address  Port  Proxyport\n");
    /*      X   XXX.XXX.XXX.XXX  XXX.XXX.XXX.XXX  01234  01234 */

	while (p < buff+info.totalsize) {
		struct ftcs_config *config = (struct ftcs_config *) p;
		struct in_addr in;

		printf("M   ");

		in.s_addr = config->inter_addr;
		printf("%-15s  ", inet_ntoa(in));

		in.s_addr = config->intra_addr;
		printf("%-15s ", inet_ntoa(in));

		printf("%5u  %5u", ntohs(config->port), ntohs(config->proxy_port));
	   
		printf("\n");

		p += sizeof(*config);
	}

	free(buff);

	return 0;
}

static int addnode(int argc, char * const argv[])
{
	const char *inter_s, *port_s, *node_s;
	struct hostent *ent;
	struct ftcs_config config;
	struct ftcs_vnode vnode;

	if (argc < 4) {
		fprintf(stderr,
				"Usage: dtsctl addnode inter_addr port new_node_addr [id]\n");
		return 1;
	}

	memset(&config, 0, sizeof(config));

	inter_s = argv[1];
	port_s  = argv[2];
	node_s  = argv[3];

	if (fill_config(inter_s, NULL, port_s, NULL, &config))
		return 1;

	ent = gethostbyname(node_s);
	if (ent == NULL) {
		fprintf(stderr, "Host not found: %s\n", node_s);
		return 1;
	}

	vnode.id = (argc >= 5) ? atoi(argv[4]) : 0;
	memcpy(&vnode.ipaddr, ent->h_addr, ent->h_length);

	if (ftcs_add_vnodes(&config, &vnode, 1)) {
		perror("ftcs_add_vnodes failed");
		return 1;
	}

	return 0;
}

static int flush(int argc, char * const argv[])
{
	if (ftcs_flush_configs() < 0) {
		perror("ftcs_flush_configs failed");
		return 1;
	}

	return 0;
}

static int get_stats(int argc, char *const argv[])
{
	struct ftcs_config config;
	struct ftcs_stats stats;

	if (argc != 3) {
		fprintf(stderr,
				"Usage: dtsctl getstats inter_addr port\n");
		return 1;
	}

	if (fill_config(argv[1], NULL, argv[2], NULL, &config))
		return 1;

	if (ftcs_get_stats(&config, &stats) < 0) {
		perror("ftcs_get_stats() failed");
		return 1;
	}

	printf("Connections accepted by proxy: %u\n", stats.nr_accepted_by_proxy);
	printf("Connections accepted by server: %u\n", stats.nr_accepted_by_server);
	printf("SSPA: SYN notification received: %u\n", stats.sspa_synnotify_recv);
	printf("SSPA: disposing SYN by ACK snooping: %u\n", stats.sspa_ack_snoop);
	printf("SSPA: Speculative SYN acceptance: %u\n", stats.sspa_syn_accept);
	printf("SSPA: dispose SYN-RECV socket by ACK comparison: %u\n",
		   stats.sspa_firstsyn_lost);

	return 0;
}

static int clear_stats(int argc, char *const argv[])
{
	struct ftcs_config config;

	if (argc != 3) {
		fprintf(stderr,
				"Usage: dtsctl clearstats inter_addr port\n");
		return 1;
	}

	if (fill_config(argv[1], NULL, argv[2], NULL, &config))
		return 1;

	if (ftcs_clear_stats(&config) < 0) {
		perror("ftcs_clear_stats() failed");
		return 1;
	}

	return 0;
}

static int remove_node(int argc, char *const argv[])
{
	struct ftcs_config config;
	uint32_t addr;
	int id = 0;

	if (argc != 4 && argc != 5) {
		fprintf(stderr,
				"Usage: dtsctl removenode inter_addr port addr [id]\n");
		return 1;
	}

	if (fill_config(argv[1], NULL, argv[2], NULL, &config))
		return 1;

	if (resolve_host(argv[3], &addr))
		return 1;

	if (argc == 5)
		id = atoi(argv[4]);

	if (ftcs_remove_node(&config, addr, id) < 0) {
		perror("ftcs_remove_node() failed");
		return 1;
	}

	return 0;
}

struct command {
	const char *command;
	int (*handler)(int argc, char * const argv[]);
};

struct command command_list[] = {
	{
		.command = "join",
		.handler = join,
	},
	{
		.command = "list",
		.handler = list,
	},
	{
		.command = "flush",
		.handler = flush,
	},
	{
		.command = "addnode",
		.handler = addnode,
	},
	{
		.command = "getstats",
		.handler = get_stats,
	},
	{
		.command = "clearstats",
		.handler = clear_stats,
	},
	{
		.command = "removenode",
		.handler = remove_node,
	},
	{

		/* Delimiter */
	},
};

static void print_usage(void)
{
	struct command *cmd;

	fprintf(stderr,
			"Usage: dtsctl <subcommand> [options...]\n"
			"\n"
			"Available subcommands:\n");

	for (cmd = command_list; cmd->command != NULL; cmd++)
		fprintf(stderr, "%s\n", cmd->command);
}

int main(int argc, char *argv[])
{
	const char *cmd_str;
	struct command *cmd;

	if (argc < 2) {
		print_usage();
		return 1;
	}

	cmd_str = argv[1];

	for (cmd = command_list; cmd->command != NULL; cmd++) {
		if (strcmp(cmd->command, cmd_str) == 0)
			return cmd->handler(argc-1, argv+1);
	}

	fprintf(stderr, "Unknown command: %s\n", cmd_str);
	print_usage();

	return 1;
}
