#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#include <arpa/inet.h>
#include <getopt.h>
#include <time.h>
#include <signal.h>
#include <sys/time.h>

#include <linux/ip.h>
#include <linux/tcp.h>

#include <ftcs.h>

struct context {
	int relative;
	int count;
	int report_missing;
	int last_id;
	int missing_max;
	int total_missing;
	struct timeval tv_prev;
};

static void sighandler(int signo)
{
	return;
}

struct timeval tv_sub(const struct timeval *tv0, const struct timeval *tv1)
{
        struct timeval res;

        res.tv_sec = tv0->tv_sec - tv1->tv_sec;
        if (tv0->tv_usec - tv1->tv_usec >= 0)
                res.tv_usec = tv0->tv_usec - tv1->tv_usec;
        else {
                res.tv_sec--;
                res.tv_usec = 1000000 + tv0->tv_usec - tv1->tv_usec;
        }

        return res;
}

static void print_timeval(const struct dts_packet_log *log,
						  const struct context *ctx)
{
	if (ctx->relative) {
		struct timeval tv = {0, 0};

		if (ctx->count > 0)
			tv = tv_sub(&log->timestamp, &ctx->tv_prev);

		printf("%04lu.%06lu ", tv.tv_sec, tv.tv_usec);
	} else {
		struct tm tm;

		localtime_r(&log->timestamp.tv_sec, &tm);

		printf(/* "%04d.%02d.%02d " */
			   "%02d:%02d:%02d.%06lu ",
			   /* 1900+tm.tm_year, 1+tm.tm_mon, tm.tm_mday, */
			   tm.tm_hour, tm.tm_min, tm.tm_sec, log->timestamp.tv_usec);
	}
}

static void print_log(const struct dts_packet_log *log,
					  struct context *ctx)
{
	const struct iphdr *iph;
	const struct tcphdr *th = NULL;
	size_t iphlen;
	struct in_addr saddr, daddr;
	char saddr_c[INET_ADDRSTRLEN], daddr_c[INET_ADDRSTRLEN];

	iph = (const struct iphdr *) log->packet_hdr;

	saddr.s_addr = iph->saddr;
	daddr.s_addr = iph->daddr;
	inet_ntop(AF_INET, &saddr, saddr_c, INET_ADDRSTRLEN);
	inet_ntop(AF_INET, &daddr, daddr_c, INET_ADDRSTRLEN);

	iphlen = iph->ihl*4;

	if (ctx->report_missing && ctx->count > 0) {
		int missing = log->id - ctx->last_id - 1;
		if (missing > 0) {
			ctx->missing_max = ctx->missing_max < missing ? missing : ctx->missing_max;
			ctx->total_missing += missing;
			printf("#### %d PACKET(S) MISSING ####\n", missing);
		}
	}

	if (DTS_LOG_PACKET_LEN - iphlen >= sizeof(*th)) {
		th = (const struct tcphdr *)(log->packet_hdr+iphlen);

		printf("%d ", log->id);
		print_timeval(log, ctx);
		printf("%c%c%c %s:%d->%s:%d"
		       " %s %c%c%c%c%c W=%d (%u%s+%d:%u)\n",
		       (log->flags & DTS_LOG_EGRESS) ? 'E' : 'I',
		       (log->flags & DTS_LOG_MODIFIED) ? 'M' : '-',
			   (log->flags & DTS_LOG_VERDICT) ?
			   ((log->flags & DTS_LOG_ACCEPT) ? 'A' : 'D') : '-',
		       saddr_c, ntohs(th->source),
		       daddr_c, ntohs(th->dest),
		       (log->flags & DTS_LOG_STATE) ?
		       dts_state_str(log->filter_state) : "-",
		       th->syn ? 'S' : '-',
		       th->ack ? 'A' : '-',
		       th->fin ? 'F' : '-',
		       th->rst ? 'R' : '-',
		       th->psh ? 'P' : '-',
			   ntohs(th->window),
		       ntohl(th->seq),
		       th->syn ? "+1" : "",
		       ntohs(iph->tot_len) - iph->ihl*4 - th->doff*4,
		       ntohl(th->ack_seq));
	} else {
		printf("%s->%s (TCPHDR MISSING)\n",
		       saddr_c,
		       daddr_c);
	}
}

static size_t process_log(const char *buff, size_t len, FILE *wfp,
						  struct context *ctx)
{
	const struct dts_packet_log *log;
	size_t processed = 0;

	while (len - processed >= sizeof(struct dts_packet_log)) {
		log = (const struct dts_packet_log *) (buff+processed);
		if (wfp) {
			struct timeval tv0, tv1, tv;

			gettimeofday(&tv0, NULL);
			fwrite(log, sizeof(*log), 1, wfp);
			gettimeofday(&tv1, NULL);

			tv = tv_sub(&tv1, &tv0);

			if (tv.tv_sec > 0 || tv.tv_usec >= 100000) {
				fprintf(stderr, "WARNING: fwrite() took %lu.%06lu sec\n",
						tv.tv_sec, tv.tv_usec);
			}
		}
		else
			print_log(log, ctx);

		ctx->count++;
		ctx->last_id = log->id;
		ctx->tv_prev = log->timestamp;
		processed += sizeof(struct dts_packet_log);
	}

	return processed;
}

static void print_usage(const char *prog_name)
{
	fprintf(stderr, "Usage: %s [options] <filename>\n"
			"\n"
			"Options:\n"
			"  -f           retries when reaching the end of the file\n"
			"  -m           reports missing log\n"
			"  -r           prints relative times since the previous packet\n"
			"  -w filename  writes raw data to the specified file\n"
			,prog_name);
}

static int setup_signal(void)
{
	struct sigaction act = {};

	act.sa_handler = sighandler;

	if (sigaction(SIGINT, &act, NULL) < 0)
		return -1;

	if (sigaction(SIGTERM, &act, NULL) < 0)
		return -1;

	return 0;
}

int main(int argc, char *argv[])
{
	int ret = 0;
	ssize_t s;
	char buff[sizeof(struct dts_packet_log)*100];
	size_t off = 0;
	struct pollfd pfd;
	const char *outfilename = NULL;
	FILE *wfp = NULL;
	int follow = 0;
	struct context ctx = {};

	if (setup_signal()) {
		perror("sigaction() failed");
		return 1;
	}

	for (;;) {
		int r = getopt(argc, argv, "frw:m");

		if (r == -1)
			break;

		switch (r) {
		case 'f':
			follow = 1;
			break;

		case 'm':
			ctx.report_missing = 1;
			break;

		case 'r':
			ctx.relative = 1;
			break;

		case 'w':
			outfilename = optarg;
			break;

		default:
			print_usage(argv[0]);
			return 1;
		}
	}

	if (argc < 1 + optind) {
		print_usage(argv[0]);
		return 1;
	}

	pfd.fd = open(argv[optind], O_RDONLY);
	if (pfd.fd < 0) {
		perror("Cannot open file");
		return 1;
	}

	if (outfilename) {
		wfp = fopen(outfilename, "wb");
		if (wfp == NULL) {
			perror("fopen() failed\n");
			return 1;
		}
	}

	pfd.events = POLLIN | POLLERR;

	while (poll(&pfd, 1, 1) >= 0) {
		size_t total_len;
		size_t processed;

		if (pfd.revents & POLLERR) {
			fprintf(stderr, "POLLERR\n");
			ret = 1;
			break;
		}

		s = read(pfd.fd, buff+off, sizeof(buff)-off);
		if (s < 0) {
			perror("read() failed");
			ret = 1;
			break;
		} else if (s == 0) {
			if (follow)
				continue;
			else
				break;
		}

		total_len = off+s;
		processed = process_log(buff, total_len, wfp, &ctx);
		off = total_len - processed;
		memmove(buff, buff+processed, off);
	}

	if (wfp)
		fclose(wfp);
	else {
		if (ctx.report_missing) {
			printf("MISSING PACKETS: total:%d max-interval:%d\n",
				   ctx.total_missing, ctx.missing_max);
		}
	}

	close(pfd.fd);

	return ret;
}
